const _ = require("lodash");
const fs = require("fs");
const Promise = require('bluebird');
const nodemailer = require('nodemailer');

const { authAxios } = require("./auth-utils");
const { ClientError } = require("./errors");

const verificationEmailTemplatePath = './verification-email-template.html';

const sendVerificationEmail = exports.sendVerificationEmail = userId => {
    console.log(`Attempting to send verification email to user: ${userId}`);
    return authAxios().then(client => client.get(`users/${userId}`).then(({ data }) => {
        const { email, user_metadata } = data; // eslint-disable-line camelcase
        const redirectURL = process.env.portalFrontEndUrl;

        return client.post(
            "tickets/email-verification",
            {
                result_url: redirectURL,
                user_id: userId
            }
        )
            .then(({ data }) => { // eslint-disable-line no-shadow
                const { ticket } = data;
                console.log(`Successfully obtained verification ticket: ${ticket}`);

                const transporter = nodemailer.createTransport({
                    host: 'smtp.hubapi.com',
                    port: 465,
                    secure: true, // secure:true for port 465, secure:false for port 587
                    auth: {
                        user: process.env.hubspotSmtpUserName,
                        pass: process.env.hubspotSmtpPassword
                    }
                });

                // Purposefully not using ES6 template literals to maintain compatibility with Auth0 rules
                const toSend = _.template(fs.readFileSync(verificationEmailTemplatePath, "utf8"), { interpolate: /{{([^}]+)}}/gm })({
                    name: user_metadata.name,
                    url: ticket,
                    service: "NetApp Cloud Central",
                    portalUrl: process.env.portalFrontEndUrl
                });
                const mailOptions = {
                    from: '"NetApp SaaS Support" <saas.support@netapp.com>',
                    to: email,
                    subject: 'Please verify your email for your NetApp Cloud Account',
                    html: toSend
                };

                return new Promise((resolve, reject) => {
                    transporter.sendMail(mailOptions, (err, info) => {
                        if (err) {
                            console.error(`Verification email send failed`, err);
                            reject(err);
                        } else {
                            console.log('Verification email sent successfully. Message %s sent: %s', info.messageId, info.response);
                            resolve();
                        }
                    });
                });
            });
    }, err => {
        console.error(`Verification email send failed`, err);
        throw err;
    }));
};

const getAllClientsRecursive = (acc, lastPage, client, fieldsString) => {
    const getAllClients = page => Promise.delay(1000).then(() => {
        console.log("getting clients");
        return client.get('clients', {
            params: {
                per_page: 100,
                page,
                include_totals: true,
                fields: fieldsString,
                search_engine: 'v2'
            }
        });
    });

    const shouldDoAgain = data => data.start + data.limit < data.total;
    return getAllClients(lastPage).then(({ data }) => {
        const newAcc = acc.concat(data.clients);
        if (shouldDoAgain(data)) {
            return getAllClientsRecursive(newAcc, lastPage + 1, client, fieldsString);
        }
        return newAcc;
    });
};

exports.getAllClients = fieldsString => {
    console.log(`Attempting to get all clients.`);
    return authAxios().then(client => getAllClientsRecursive([], 0, client, fieldsString)
        .then(res => {
            console.log(`Got all ${res.length} clients`);
            return res;
        }), err => {
        console.error(`Error. Getting all clients failed`, err);
        throw err;
    });
};

const getUsers = exports.getUsers = ({ params }) => {
    console.log(`Attempting to get Auth0 users.`);

    const paramsToSend = _.isObject(params) ? params : {};
    paramsToSend.include_totals = true;
    paramsToSend.search_engine = 'v3';

    return authAxios()
        .then(authClient => authClient.get(`users`, { params: paramsToSend })
            .then(({ data }) => {
                if (data.length !== data.total) {
                    console.log(`More than ${data.limit} users exist.`);
                    throw new ClientError(`More than ${data.limit} users exist. Query failed.`);
                }
                console.log(`Successfully retrieved ${data.length} users from Auth0.`);

                return data.users;
            }, err => {
                console.log(`Failed to retrieve Auth0 users.`, err);
                throw err;
            }));
};

const getUsersDetailsByIds = exports.getUsersDetailsByIds = ({ userIds = "" }) => {
    const usersArr = userIds === "" ? [] : userIds.split(",");
    console.log(`Attempting to get user data for ${usersArr.length} users with ids: ${userIds}`);

    if (usersArr.length === 0) {
        return Promise.resolve([]);
    }

    return getUsers({
        params: {
            q: usersArr.map(userId => `user_id: ${userId}`).join(' OR '),
            fields: 'user_id,user_metadata,email'
        }
    })
        .then(users => users.map(user => ({
            userId: user.user_id,
            name: user.user_metadata.name,
            company: user.user_metadata.company,
            email: user.email
        })))
        .then(result => {
            console.log(`Success. Got ${result.length} users' details by ids.`);
            return result;
        }, err => {
            console.error(`Error. Get users' details by ids failed`, err);
            throw err;
        });
};

const getUserByEmail = exports.getUserByEmail = async ({ email }) => {
    console.log(`Attempting to get user data by email ${email}.`);
    const authClient = await authAxios();
    const { data: usersResponse } = await authClient.get("users", {
        params: {
            q: `email:"${email}"`,
            search_engine: 'v3'
        }
    });

    if (usersResponse.length === 0) {
        console.log(`Could not find user with Email ${email}.`);

        return null;
    }
    if (usersResponse.length > 1) {
        throw new ClientError(`Multiple users have the same email address: ${email}`);
    }

    console.log(`Successfully got 1 user.`);

    return usersResponse[0];
};

const yarger = require("yargs") // eslint-disable-line no-unused-vars, import/order
    .command("send-verif-email", "send-verif-email", yargs => {
        yargs.demandOption(["userId"]);
    }, argv => {
        sendVerificationEmail(argv.userId);
    })
    .command("get-user-details-by-ids", "get-user-details-by-ids", yargs => {
        yargs.demandOption(["userIds"]);
    }, argv => {
        getUsersDetailsByIds({ userIds: argv.userIds })
            .then(console.log, console.log);
    })
    .command("get-user-by-email", "get-user-by-email", yargs => {
        yargs.demandOption(["email"]);
    }, argv => {
        getUserByEmail({ email: argv.email })
            .then(console.log, console.log);
    }).argv;
