const mockAWS = require('aws-sdk-mock');
const assert = require('assert');
const sinon = require('sinon');
const nock = require('nock');

const entry = require("./governance-apis");
const { authAccount } = require("./auth-utils");

const nockBase = nock(`https://${authAccount.domain}`);

process.env.isTest = true;
process.env.userInfoTable = "portal-backend-dev-userInfo";
process.env.occmClientInfoTable = "portal-backend-dev-occmClientInfo";

beforeEach(() => {
    nockBase
        .post('/oauth/token')
        .reply(200, { access_token: "mock-access-token" });
});

afterEach(() => {
    mockAWS.restore();
});


describe("deleteUser", () => {
    it("it should just delete the user if no occms are given", () => {
        nockBase
            .delete("/api/v2/users/user-1")
            .reply(204);

        const promise = entry.deleteUser(({ userId: "user-1" }));
        return expect(promise).resolves.toEqual(null);
    });


    it("it should delete occms created by the user, then delete the user", () => {
        // ---batch get all clients---
        const expectedBatchGetParams = {
            RequestItems: {
                [process.env.occmClientInfoTable]: {
                    Keys: [{ clientId: "occm-created-by-user" }, { clientId: "occm-not-created-by-user" }]
                }
            }
        };
        const batchGetSpy = sinon.spy((params, callback) => {
            callback(null, {
                [`Responses.${process.env.occmClientInfoTable}`]: [
                    { clientId: "occm-created-by-user", setupUserId: "user-1" },
                    { clientId: "occm-not-created-by-user", setupUserId: "wrong-user" }
                ]
            });
        });
        mockAWS.mock('DynamoDB.DocumentClient', 'batchGet', batchGetSpy);

        // delete client from dyanamo and auth0
        const expectedDelClientParams = {
            TableName: process.env.occmClientInfoTable,
            Key: { clientId: "occm-created-by-user" }
        };
        const delSpy = sinon.spy((params, callback) => {
            callback(null, 'success');
        });
        mockAWS.mock('DynamoDB.DocumentClient', 'delete', delSpy);

        nockBase
            .delete("/api/v2/clients/occm-created-by-user")
            .reply(204);

        // delete user from auth0
        nockBase
            .delete("/api/v2/users/user-1")
            .reply(204);

        return entry.deleteUser(({ userId: "user-1", occms: ["occm-created-by-user", "occm-not-created-by-user"] }))
            .then(val => {
                assert(batchGetSpy.calledWith(expectedBatchGetParams), 'should pass correct parameters for batch get spy');
                assert(delSpy.calledWith(expectedDelClientParams), 'should pass correct parameters for delete client spy');
                return expect(val).toEqual(null);
            });
    });
});


describe("deleteOccmClientsCreatedBy", () => {
    it("it should do nothing (and not fail) if the user has no associated occms", () => {
        // ---get user info---
        const getUserInfoSpy = sinon.spy((params, callback) => {
            callback(null, { Item: { userId: "user-1" } }); // no associated occms
        });
        mockAWS.mock('DynamoDB.DocumentClient', 'get', getUserInfoSpy);


        const promise = entry.deleteOccmClientsCreatedBy(({ userId: "user-1" }));
        return expect(promise).resolves.toEqual(null);
    });


    it("it should delete only occms created by the user", () => {
        // ---get user info---
        const getUserInfoSpy = sinon.spy((params, callback) => {
            callback(null, { Item: { userId: "user-1", associatedOccms: { values: ["occm-created-by-user", "occm-not-created-by-user"] } } });
        });
        mockAWS.mock('DynamoDB.DocumentClient', 'get', getUserInfoSpy);


        // ---batch get all clients---
        const expectedBatchGetParams = {
            RequestItems: {
                [process.env.occmClientInfoTable]: {
                    Keys: [{ clientId: "occm-created-by-user" }, { clientId: "occm-not-created-by-user" }]
                }
            }
        };
        const batchGetSpy = sinon.spy((params, callback) => {
            callback(null, {
                [`Responses.${process.env.occmClientInfoTable}`]: [
                    { clientId: "occm-created-by-user", setupUserId: "user-1" },
                    { clientId: "occm-not-created-by-user", setupUserId: "wrong-user" }
                ]
            });
        });
        mockAWS.mock('DynamoDB.DocumentClient', 'batchGet', batchGetSpy);

        // delete client from dyanamo and auth0
        const expectedDelClientParams = {
            TableName: process.env.occmClientInfoTable,
            Key: { clientId: "occm-created-by-user" }
        };
        const delSpy = sinon.spy((params, callback) => {
            callback(null, 'success');
        });
        mockAWS.mock('DynamoDB.DocumentClient', 'delete', delSpy);

        nockBase
            .delete("/api/v2/clients/occm-created-by-user")
            .reply(204);


        // ---update user info, remove all associated occms---
        const updateUserInfoParams = {
            TableName: process.env.userInfoTable,
            ExpressionAttributeValues: {
                ":c": { SS: ["occm-created-by-user", "occm-not-created-by-user"] }
            },
            Key: {
                userId: {
                    S: "user-1"
                }
            },
            UpdateExpression: "DELETE associatedOccms :c"
        };
        const updateUserInfoSpy = sinon.spy((params, callback) => {
            callback(null, 'success');
        });

        mockAWS.mock('DynamoDB', 'updateItem', updateUserInfoSpy);


        return entry.deleteOccmClientsCreatedBy(({ userId: "user-1" }))
            .then(val => {
                assert(batchGetSpy.calledWith(expectedBatchGetParams), 'should pass correct parameters for batch get spy');
                assert(delSpy.calledWith(expectedDelClientParams), 'should pass correct parameters for delete client spy');
                assert(updateUserInfoSpy.calledWith(updateUserInfoParams), 'should pass correct parameters for update user info spy');
                return expect(val).toEqual(null);
            });
    });
});
