const axios = require('axios');
const jwt = require('jsonwebtoken');
const _ = require('lodash');
const AWS = require('aws-sdk');

const dispatch = require('./dispatch');
// eslint-disable-line import/no-extraneous-dependencies
AWS.config.update({ region: 'us-east-1' });
AWS.config.setPromisesDependency(require('bluebird'));

const tokenCallback = async ({ queryParams }) => {
    console.log('Attempting to process request to generate refresh token');
    console.log(queryParams);
    try {
        const { data: { refresh_token: refreshToken, access_token: responseAccessToken } } = await axios.post(`https://${process.env.managementDomain}/oauth/token`, {
            grant_type: "authorization_code",
            client_id: process.env.managementRefreshTokenId,
            code: queryParams.code,
            redirect_uri: process.env.portalBackEndUrl
        });
        console.log('Received Auth0 response. Validating requesting user and response token owner are the same');
        console.log('Access token:', responseAccessToken);
        const requestAccessToken = Buffer.from(queryParams.state, 'base64').toString();

        const decodedRequestToken = jwt.decode(requestAccessToken, { complete: true });
        const decodedResponseToken = jwt.decode(responseAccessToken, { complete: true });
        const requestUserId = _.get(decodedRequestToken, 'payload.sub');
        const responseUserId = _.get(decodedResponseToken, 'payload.sub');

        if (requestUserId !== responseUserId) {
            console.log(`Failed Validation. Requesting user (${requestUserId}) != response user (${responseUserId})`);
            return { statusCode: 400, link: `${process.env.portalFrontEndUrl}/refresh-token/?errorMessage=Request%20user%20does%20not%20match` };
        }
        console.log(decodedResponseToken);
        const requestedAudience = _.get(decodedResponseToken, 'payload.aud');
        console.log('Checking for previous tokens in dynamo');
        const dynamoDb = new AWS.DynamoDB.DocumentClient();
        const params = {
            TableName: process.env.refreshTokenTable,
            Key: { userId: responseUserId }
        };
        const dynamoResponse = await dynamoDb.get(params).promise();
        if (dynamoResponse.Item && dynamoResponse.Item[requestedAudience]) {
            console.log(`Failed Validation. Requesting user (${requestUserId}) already has a token in dynamo with the audience: ${requestedAudience}.`);
            return { statusCode: 400, link: `${process.env.portalFrontEndUrl}/refresh-token/?errorMessage=Token%20already%20exists` };
        }
        console.log(`No previous token found for the requested audience (${requestedAudience}). Adding new entry in dynamo`);
        const dynamoEntry = {
            lastCharacters: refreshToken.substr(refreshToken.length - 4),
            scope: _.get(decodedResponseToken, 'payload.scope', 'unknown')
        };
        const putParams = {
            TableName: process.env.refreshTokenTable,
            Key: { userId: responseUserId },
            AttributeUpdates: {
                [requestedAudience]: { Action: "PUT", Value: dynamoEntry }
            }
        };
        await dynamoDb.update(putParams).promise();
        console.log('Dynamo entry added');
        console.log('Refresh token successfully retrieved');
        return { statusCode: 303, link: `${process.env.portalFrontEndUrl}/refresh-token/?refreshToken=${refreshToken}` };
    } catch (err) {
        console.log(err);
        throw new Error('Failed to generate refresh token.');
    }
};

const awsSubscriptionRedirect = ({ serviceName, body }) => {
    console.log(`received aws subscription redirect for ${serviceName}`);
    const token = body['x-amzn-marketplace-token'];
    if (!token) {
        console.log('No amazon marketplace token present in response.');
        return { statusCode: 303, link: `${process.env.subscriptionMappingRedirect}?serviceName=${serviceName}&cloudProvider=aws&error='Invalid token response'` };
    }

    return { statusCode: 303, link: `${process.env.subscriptionMappingRedirect}?serviceName=${serviceName}&cloudProvider=aws&token=${encodeURIComponent(token)}` };
};

const routes = [
    {
        method: "GET",
        pattern: "/misc/token-callback",
        handler: ({ queryParams }) => tokenCallback({
            queryParams
        })
    },
    {
        method: "POST",
        pattern: "/misc/subscription-redirect/aws/:serviceName",
        handler: ({ urlParams, body }) => awsSubscriptionRedirect({ serviceName: urlParams.serviceName, body })
    },
    {
        method: "GET",
        pattern: "/misc/health",
        handler: () => {
            if (process.env.isHealthy === 'true') {
                return Promise.resolve({
                    up: true
                });
            }
            return Promise.reject(new Error("Bad Unhealthy!"));
        }
    }
];

exports.handler = (event, context, callback) => {
    dispatch({
        event, callback, routes, context
    });
};
