const _ = require("lodash");
const Joi = require('joi');
const util = require('util');
const Promise = require('bluebird');
const AWS = require('aws-sdk');

const { authAxios } = require("./auth-utils");
const dispatch = require('./dispatch');
const { ClientError, ForbiddenError } = require("./errors");
const { validateAndExtractToken } = require("./authorizer");
const { sendVerificationEmail, getUsersDetailsByIds, getUserByEmail } = require("./common");
const Notifications = require("./notifications");

const OCCM_NAME = "Cloud Volumes Ontap";

AWS.config.update({ region: 'us-east-1' });
AWS.config.setPromisesDependency(require('bluebird'));

const ignoreBadUris = uris => _.without(uris, "http://", "http:///occmui", "https://", "https:///occmui");

const associateUserInternal = exports.associateUserInternal = (userId, occmClientId, associaterId) => {
    const dynamoDbOld = new AWS.DynamoDB({ region: 'us-east-1' });
    console.log(`Making update call to user with id: ${userId} to associate with occm: ${occmClientId}`);
    const params = {
        TableName: process.env.userInfoTable,
        ExpressionAttributeValues: {
            ":c": { SS: [occmClientId] },
            ":t": { BOOL: true }
        },
        Key: {
            userId: {
                S: userId
            }
        },
        ReturnValues: "UPDATED_OLD",
        UpdateExpression: "ADD associatedOccms :c  SET hadPreviousOccm = :t"
    };

    return dynamoDbOld.updateItem(params).promise().then(updateResponse => {
        const oldAssociatedOccms = _.get(updateResponse, "Attributes.associatedOccms.SS", []);
        if (oldAssociatedOccms.includes(occmClientId)) {
            console.log(`User with id ${userId} already associated with occm with clientId: ${occmClientId}`);
            return Promise.resolve(userId);
        }
        return associaterId ? Notifications.insertAssociateUserNotification(userId, occmClientId, associaterId).then(() => userId) : Promise.resolve(userId);
    });
};

const invalidUriRegex = /allowed_logout_urls\[([0-9]+)\]/g;

const extractUriError = exports.extractUriError = ({ err, manualOverrideUris, allCallbacks }) => {
    const s = _.get(err, "response.data.message", "");
    console.log(`Extracted: ${s}`);
    const matches = s.match(invalidUriRegex);
    console.log(`matches: ${matches}`);

    if (matches && matches.length > 0) {
        let fail = null;
        try {
            const errIndices = matches.map(el => parseInt(el.replace("allowed_logout_urls[", "").replace("]", "")));
            const errStrings = _.filter(allCallbacks, (cb, index) => errIndices.includes(index));
            const manualOverrideUrisValues = _.map(manualOverrideUris, "value");
            const actualErrIndices = errStrings.map(errString => manualOverrideUrisValues.indexOf(errString)).filter(index => index > -1);
            fail = new ClientError(`Invalid URIs: ${actualErrIndices.join(",")}`);
        } catch (e) {
            throw err;
        }

        if (fail) { throw fail; }
    } else { throw err; }
};

const updateOccmUrisInternal = ({
    dynamoDb, occmClientId, manualOverrideUris, automaticCallbackUris, markSetupComplete, mapInvalidUriError
}) => {
    console.log(`Calculating and saving occm uris based on manual-automatic callbacks, ${util.inspect({ manualOverrideUris, automaticCallbackUris }, { showHidden: false, depth: null })}`);
    const primaryCallbackUri = _.get(manualOverrideUris.find(el => el.isDefault), "value", null) || automaticCallbackUris[0];
    const mergedCallbacks = _.uniq(_.map(manualOverrideUris, "value").concat(automaticCallbackUris));
    const allCallbacks = (primaryCallbackUri ? [primaryCallbackUri] : []).concat(_.without(mergedCallbacks, primaryCallbackUri));
    const allCallbacksWithSlashOccm = allCallbacks.concat(_(manualOverrideUris)
        .map("value")
        .map(uri => `${uri + (uri.endsWith('/') ? "" : "/")}occmui/`)
        .value());
    console.log(`PrimaryCallbackUri is: ${primaryCallbackUri}, All callbacks list calculated as: ${allCallbacks}`);

    return authAxios().then(client => client.patch(`clients/${occmClientId}`, {
        callbacks: _.uniq((allCallbacksWithSlashOccm.concat(allCallbacksWithSlashOccm.map(uri => `${uri + (uri.endsWith('/') ? "" : "/")}silent-callback.html`)))),
        allowed_logout_urls: _.uniq(allCallbacks)
    }).then(() => {
        const params = {
            TableName: process.env.occmClientInfoTable,
            Key: { clientId: occmClientId },
            AttributeUpdates: {
                automaticCallbackUris: { Action: "PUT", Value: automaticCallbackUris },
                manualOverrideUris: { Action: "PUT", Value: manualOverrideUris }
            }
        };

        if (markSetupComplete) {
            params.AttributeUpdates.setupComplete = { Action: "PUT", Value: true };
        }

        return dynamoDb.update(params).promise()
            .then(() => {
                console.log(`Update occm uris succeeded.`);
                return null;
            })
            .catch(err => {
                console.log('Possible inconsistent state. Callback URIs updated in Auth0 but not in dynamo');
                throw err;
            });
    }, err => {
        const statusCode = _.get(err, "response.data.statusCode");
        const errMessage = _.get(err, "response.data.message");
        if (statusCode && mapInvalidUriError) {
            console.log(err);
            extractUriError({ err, manualOverrideUris, allCallbacks });
        } else if (statusCode) {
            console.log(err);
            if (statusCode === 404) {
                throw new ClientError(`Could not find Cloud Manager with clientId: ${occmClientId}`);
            } else {
                throw new ClientError(`StatusCode: ${statusCode} Message: ${errMessage}`);
            }
        } else { throw err; }
    })).then(() => {
        console.log("Auth0 and dynamo successfully updated with new OCCM callback URIs");
        return null;
    });
};

const normalizeClientId = clientId => {
    let normalizedClientId = clientId;
    if (clientId.endsWith("@clients")) { normalizedClientId = clientId.substring(0, clientId.length - "@clients".length); }
    console.log(`Normalized client id: ${normalizedClientId}`);
    return normalizedClientId;
};

const associateUser = exports.associateUser = ({ occmClientId, associateeEmail, associaterId }) => {
    console.log(`Attempting to associate user with email: ${associateeEmail} to occm with clientId: ${occmClientId} on behalf of user with id: ${associaterId}`);
    const normalizedClientId = normalizeClientId(occmClientId);
    return getUserByEmail({ email: associateeEmail }).then(associateeUser => {
        if (!associateeUser) {
            const defaultErrorMessage = `No user has the given email address: ${associateeEmail}, the user needs to first sign up to Cloud Central (cloud.netapp.com).`;
            return Promise.reject(new ClientError(defaultErrorMessage));
        }

        const { user_id } = associateeUser; // eslint-disable-line camelcase
        return associateUserInternal(user_id, normalizedClientId, associaterId);
    }).then(userId => {
        console.log(`Association successful`);
        return { userId };
    }, err => {
        console.error(`Association failed`, err);
        throw err;
    });
};

const deassociateUser = exports.deassociateUser = ({
    occmClientId, deassociaterId, deassociateeId, notificationType
}) => {
    console.log(`Attempting to deassociate user with id: ${deassociateeId} from occm with clientId: ${occmClientId} on behalf of user with id: ${deassociaterId}`);
    const dynamoDbOld = new AWS.DynamoDB();
    const normalizedClientId = normalizeClientId(occmClientId);
    const params = {
        TableName: process.env.userInfoTable,
        ExpressionAttributeValues: {
            ":c": { SS: [normalizedClientId] }
        },
        Key: {
            userId: {
                S: deassociateeId
            }
        },
        ReturnValues: "UPDATED_OLD",
        UpdateExpression: "DELETE associatedOccms :c"
    };

    return dynamoDbOld.updateItem(params).promise().then(updateUserResponse => {
        const occms = _.get(updateUserResponse, "Attributes.associatedOccms.SS", []);
        if (!occms.includes(normalizedClientId)) {
            console.log(`User with id ${deassociateeId} is not associated with occm with clientId: ${normalizedClientId}`);
            return Promise.resolve(null);
        }
        console.log(`Made update call to user with id: ${deassociateeId} to deassociate from occm: ${normalizedClientId}`);
        if (notificationType === "deassociate") {
            return Notifications.insertDeassociateUserNotification(deassociateeId, normalizedClientId, deassociaterId).then(() => null); // await notification insertion so that lambda does not terminate before completion
        }
        if (notificationType === "remove") {
            return Notifications.insertRemoveFromListNotification(deassociateeId, normalizedClientId).then(() => null);
        }
        return Promise.resolve(null);
    }).then(() => {
        console.log(`Deassociation successful`);
        return null;
    }, err => {
        console.error(`Deassociation failed`, err);
        throw err;
    });
};

const listOccmsOfUser = exports.listOccmsOfUser = ({ userId }) => {
    const dynamoDb = new AWS.DynamoDB.DocumentClient();
    const userInfoParams = {
        TableName: process.env.userInfoTable,
        Key: { userId: `${userId}` }
    };
    console.log(`Getting user ${userId} info from dynamo`);
    return dynamoDb.get(userInfoParams).promise().then(userInfoResponse => {
        const userInfo = userInfoResponse.Item || {};
        const associatedOccms = _.get(userInfo, "associatedOccms.values", []);
        const hadPreviousOccm = _.get(userInfo, "hadPreviousOccm", false);
        console.log(`associatedOccms found:`);
        console.log(associatedOccms);
        console.log(`hadPreviousOccm: ${hadPreviousOccm}`);

        if (associatedOccms.length === 0) {
            return { occms: [], hadPreviousOccm };
        }

        let batch = 0;
        let results = [];
        const occmsForBatching = _.map(associatedOccms, clientId => ({ clientId }));

        const paginateBatchResults = () => {
            if (batch >= associatedOccms.length || batch >= 1000) {
                console.log(`Collected ${results.length} occm results`);
                return { occms: results, hadPreviousOccm };
            }
            console.log(`Collecting batch ${batch}...${batch + 100}`);
            const requestParams = {
                RequestItems: {
                    [process.env.occmClientInfoTable]: {
                        Keys: _.slice(occmsForBatching, batch, batch + 100)
                    }
                }
            };
            batch += 100;
            return dynamoDb.batchGet(requestParams).promise()
                .then(batchGetResponse => {
                    console.log(batchGetResponse);
                    const responseArr = _.get(batchGetResponse, `Responses.${process.env.occmClientInfoTable}`, []);
                    console.log('Batch get response array');
                    console.log(util.inspect(responseArr, { showHidden: false, depth: null }));
                    return responseArr.map(resp => {
                        if (resp.saas && !resp.setupComplete) {
                            return {
                                occm: resp.clientId,
                                occmName: resp.occmName,
                                status: resp.setupFailed ? "failed" : "initializing",
                                createDate: resp.createDate
                            };
                        }
                        const primaryCallbackUri = _.get((resp.manualOverrideUris || []).find(el => el.isDefault), "value", null) || (resp.automaticCallbackUris || [])[0];
                        if (!primaryCallbackUri || !resp.occmName) {
                            return {
                                occm: resp.clientId,
                                failed: true,
                                message: "Client does not have a primaryCallbackUri or metadata with an occmName"
                            };
                        }
                        return {
                            occm: resp.clientId,
                            status: "ready",
                            occmName: resp.occmName,
                            primaryCallbackUri,
                            manualOverrideUris: resp.manualOverrideUris || [],
                            automaticCallbackUris: resp.automaticCallbackUris,
                            createDate: resp.createDate
                        };
                    });
                })
                .then(occmsBatch => {
                    console.log(`Collected ${occmsBatch.length}`);
                    results = _.concat(results, occmsBatch);
                    return paginateBatchResults();
                })
                .catch(err => {
                    console.error(`List occms failed`, err);
                    throw err;
                });
        };
        return paginateBatchResults();
    });
};

const registerOccm = exports.registerOccm = ({ redirectUris, systemId }) => {
    console.log(`Attempting to register occm: ${JSON.stringify({ redirectUris, systemId })}`);
    const filteredRedirectUris = ignoreBadUris(redirectUris);
    const dynamoDb = new AWS.DynamoDB.DocumentClient();
    const clientName = `Occm-Client-${systemId}`;
    const params = {
        TableName: process.env.occmClientInfoTable,
        IndexName: 'clientName-Index',
        KeyConditionExpression: 'clientName = :hkey',
        ExpressionAttributeValues: {
            ':hkey': `${clientName}`
        }
    };
    return authAxios().then(client => dynamoDb.query(params).promise()
        .then(queryResult => {
            const occm = _.get(queryResult, "Items[0]", null);
            if (occm) {
                return client.get(`clients/${occm.clientId}`)
                    .then(({ data }) => {
                        console.log(data);
                        return Promise.resolve({
                            clientId: occm.clientId,
                            clientSecret: data.client_secret,
                            alreadyExists: true,
                            setupComplete: !!_.get(occm, "setupComplete")
                        });
                    });
            }
            const createDate = `${(new Date()).getTime()}`;
            return client.post("clients", {
                name: clientName,
                description: "OCCM Client",
                callbacks: (filteredRedirectUris.concat(filteredRedirectUris.map(uri => `${uri + (uri.endsWith('/') ? "" : "/")}silent-callback.html`))),
                allowed_logout_urls: filteredRedirectUris,
                grant_types: ["client_credentials", "implicit", "password", "http://auth0.com/oauth/grant-type/password-realm"],
                jwt_configuration: { alg: "RS256" },
                app_type: "spa",
                client_metadata: {
                    displayName: OCCM_NAME,
                    createDate
                }
            }).then(({ data }) => {
                const clientId = data.client_id;
                const clientSecret = data.client_secret;
                const clientDynamoParams = {
                    TableName: process.env.occmClientInfoTable,
                    Item: {
                        clientId,
                        clientName,
                        automaticCallbackUris: filteredRedirectUris,
                        createDate
                    }
                };

                return Promise.all([
                    // since this client include client_credential grants, we need to grant access to the https://api.cloud.netapp.com API
                    // usually SPA client are not required to do so
                    client.post("client-grants", {
                        client_id: clientId,
                        audience: "https://api.cloud.netapp.com",
                        scope: ["occm-access"]
                    }),

                    // change the token_endpoint_auth_method to 'none' to allow usage of the password grant
                    client.patch(`clients/${clientId}`, {
                        token_endpoint_auth_method: 'none'
                    }),

                    // create dynamoDb item for this client
                    dynamoDb.put(clientDynamoParams).promise()

                ])
                    .then(() => ({
                        clientId,
                        clientSecret,
                        alreadyExists: false,
                        setupComplete: false
                    }));
            }, err => { // call to auth0 failed
                const statusCode = _.get(err, 'response.status', null);
                if (statusCode === 400) {
                    err.auth0ClientError = true;
                }
                throw err;
            });
        })).then(occm => {
        if (occm.alreadyExists) { console.log(`Registration not necessary, already exists.`); } else { console.log(`Registration complete.`); }
        const clone = _.clone(occm);
        clone.clientSecret = "****";
        console.log(clone);
        return occm;
    }, err => {
        if (err.auth0ClientError) {
            console.error(`Registration failed. Error returned from auth0.`, err);
            throw new ClientError("Call to Auth0 failed due to client error. Check passed parameters for duplicates or illegal characters.");
        }
        console.error(`Registration failed. Possible inconsistent state.`, err);
        throw new Error('Occm registration failed');
    });
};

const occmSetupComplete = exports.occmSetupComplete = ({ setupUserId, occmClientId, occmName }) => {
    console.log(`Attempting to mark occm setup as completed. ${JSON.stringify({ setupUserId, occmClientId, occmName })}`);
    const dynamoDb = new AWS.DynamoDB.DocumentClient();
    const normalizedClientId = normalizeClientId(occmClientId);
    const params = {
        TableName: process.env.occmClientInfoTable,
        Key: { clientId: normalizedClientId },
        AttributeUpdates: {
            setupComplete: { Action: "PUT", Value: true },
            occmName: { Action: "PUT", Value: occmName },
            setupUserId: { Action: "PUT", Value: setupUserId }
        }
    };
    return dynamoDb.update(params).promise().then(() =>
        associateUserInternal(setupUserId, normalizedClientId).then(() => Notifications.insertOccmSetupCompleteNotification(setupUserId, occmName).then(() => null))) // await notification insertion so that lambda does not terminate before completion
        .then(() => {
            console.log(`Occm setup completion succeeded.`);
            return null;
        }, err => {
            console.error(`Occm setup completion failed`, err);
            throw err;
        });
};

const updateOccmUris = exports.updateOccmUris = ({ redirectUris, occmClientId }) => {
    console.log(`Attempting to update occm uris and marking setup as complete. ${JSON.stringify({ redirectUris, occmClientId })}`);
    const filteredRedirectUris = ignoreBadUris(redirectUris);
    const dynamoDb = new AWS.DynamoDB.DocumentClient();
    const normalizedClientId = normalizeClientId(occmClientId);


    const getParams = {
        TableName: process.env.occmClientInfoTable,
        Key: { clientId: normalizedClientId }
    };

    return dynamoDb.get(getParams).promise().then(response => {
        const occm = response.Item || {};
        return updateOccmUrisInternal({
            dynamoDb, occmClientId: normalizedClientId, manualOverrideUris: occm.manualOverrideUris || [], automaticCallbackUris: filteredRedirectUris, markSetupComplete: true
        });
    }).catch(err => {
        console.error(`Update occm uris failed`, err);
        throw err;
    });
};

const getUserOccms = ({ userId }) => {
    const dynamoDb = new AWS.DynamoDB.DocumentClient();
    const userInfoParams = {
        TableName: process.env.userInfoTable,
        Key: { userId }
    };

    return dynamoDb.get(userInfoParams).promise().then(userInfoResponse => {
        const userInfo = userInfoResponse.Item || {};
        return _.get(userInfo, "associatedOccms.values", []);
    });
};

const verifyUserHasOccm = ({ userId, occmId }) => getUserOccms({ userId })
    .then(userOccms => _.includes(userOccms, occmId));

const manualUpdateOccmUris = exports.manualUpdateOccmUris = async ({ userId, manualOverrideUris, occmId }) => {
    console.log(`Attempting to update manual override uris for OCCM ${JSON.stringify({ manualOverrideUris, occmId })}`);

    const userHasThisOccm = await verifyUserHasOccm({ userId, occmId });

    if (!userHasThisOccm) {
        throw new ForbiddenError('User is not allowed to change OCCM');
    }

    const normalizedClientId = normalizeClientId(occmId);

    const dynamoDb = new AWS.DynamoDB.DocumentClient();
    const getParams = {
        TableName: process.env.occmClientInfoTable,
        Key: { clientId: normalizedClientId }
    };

    return dynamoDb.get(getParams).promise().then(response => {
        const occm = response.Item || {};

        if (occm.setupFailed) {
            return Promise.reject(new ClientError(`Unable to manually set Cloud Manager URIs for a Cloud Manager that failed to setup.`));
        }
        if (occm.saas && !occm.setupComplete) {
            return Promise.reject(new ClientError(`Unable to manually set Cloud Manager URIs for a Cloud Manager that is still initializing.`));
        }

        const automaticCallbackUris = _.get(occm, "automaticCallbackUris", []);
        return updateOccmUrisInternal({
            dynamoDb, occmClientId: normalizedClientId, manualOverrideUris, automaticCallbackUris, markSetupComplete: false, mapInvalidUriError: true
        });
    }).catch(err => {
        console.error("Update manual override uris failed");
        throw err;
    });
};

// eslint-disable-next-line object-curly-newline
const serviceRegisterOccm = exports.serviceRegisterOccm = ({ occmName, systemId, setupUserId, isAgent = false }) => {
    console.log(`Attempting to register occm for OCCM as a service: ${JSON.stringify({ occmName, systemId, setupUserId })}`);
    const dynamoDb = new AWS.DynamoDB.DocumentClient();
    return authAxios().then(client => {
        const clientName = `Occm-Client-${systemId}`;
        const createDate = `${(new Date()).getTime()}`;

        return client.post("clients", {
            name: clientName,
            description: "OCCM Client",
            grant_types: ["client_credentials", "implicit", "password", "http://auth0.com/oauth/grant-type/password-realm"],
            jwt_configuration: { alg: "RS256" },
            app_type: "spa",
            allowed_origins: ["http://localhost:4300", "https://setup.occm.datafabric.io", "https://dev-setup.occm.datafabric.io", "https://setup.cloudmanager.netapp.com", "https://dev-setup.cloudmanager.netapp.com"],
            client_metadata: {
                displayName: OCCM_NAME,
                createDate
            }
        }).then(({ data }) => {
            const clientId = data.client_id;
            const clientSecret = data.client_secret;
            const clientDynamoParams = {
                TableName: process.env.occmClientInfoTable,
                Item: {
                    clientId,
                    clientName,
                    setupUserId,
                    occmName,
                    saas: true,
                    agent: isAgent,
                    createDate
                }
            };

            return Promise.all([
                // since this client include client_credential grants, we need to grant access to the https://api.cloud.netapp.com API
                // usually SPA client are not required to do so
                client.post("client-grants", {
                    client_id: clientId,
                    audience: "https://api.cloud.netapp.com",
                    scope: ["occm-access"].concat(isAgent ? ["agents:agent"] : [])
                }),

                // change the token_endpoint_auth_method to 'none' to allow usage of the password grant
                client.patch(`clients/${clientId}`, {
                    token_endpoint_auth_method: 'none'
                }),

                // create client item in dynamo
                dynamoDb.put(clientDynamoParams).promise(),

                // associate user (not for agents)
                isAgent ? Promise.resolve() : associateUserInternal(setupUserId, clientId)

            ]).then(() => ({
                clientId,
                clientSecret
            }));
        }, err => { // call to auth0 failed
            const statusCode = _.get(err, 'response.status', null);
            if (statusCode === 400) {
                err.auth0ClientError = true;
            }
            throw err;
        });
    }).then(occm => {
        console.log(`Registration complete.`);
        const clone = _.clone(occm);
        clone.clientSecret = "****";
        console.log(clone);
        return occm;
    }, err => {
        if (err.auth0ClientError) {
            console.error(`Registration failed. Error returned from auth0.`, err);
            throw new ClientError("Call to Auth0 failed due to client error. Check passed parameters for duplicates or illegal characters.");
        }
        console.error(`Registration failed. Possible inconsistent state.`, err);
        throw new Error('Occm registration failed');
    });
};

const serviceSetupFailed = exports.serviceSetupFailed = ({ occmClientId, setupUserId }) => {
    console.log(`Attempting to mark occm setup (as a service) as failed. ${JSON.stringify({ occmClientId, setupUserId })}`);
    const dynamoDb = new AWS.DynamoDB.DocumentClient();
    const normalizedClientId = normalizeClientId(occmClientId);
    let occmName;
    const clientInfoParams = {
        TableName: process.env.occmClientInfoTable,
        Key: { clientId: normalizedClientId }
    };
    return dynamoDb.get(clientInfoParams).promise().then(clientInfoResponse => {
        const clientInfo = clientInfoResponse.Item || {};
        occmName = _.get(clientInfo, "occmName");
        if (clientInfo.setupUserId !== setupUserId) {
            return Promise.reject(new ForbiddenError(`User ${setupUserId} does not have permission to affect ${normalizedClientId}`));
        }
        const params = {
            TableName: process.env.occmClientInfoTable,
            Key: { clientId: normalizedClientId },
            AttributeUpdates: {
                setupFailed: { Action: "PUT", Value: true }
            }
        };
        return dynamoDb.update(params).promise();
    }).then(() => {
        console.log(`Marking occm setup as failed - succeeded.`);
        return Notifications.insertOccmSetupFailedNotification(setupUserId, occmName).then(() => null); // await notification insertion so that lambda does not terminate before completion
    }, err => {
        console.error(`Marking occm setup as failed - failed.`, err);
        throw err;
    });
};

const getOccmUsers = exports.getOccmUsers = ({ occmClientId, usersString = "" }) => {
    const normalizedClientId = normalizeClientId(occmClientId);
    console.log(`Attempting to get user data for all users associated with this occm client: ${JSON.stringify({ normalizedClientId })}.`);

    return getUsersDetailsByIds({ userIds: usersString });
};

const getOccmUsersRecursive = (acc, lastPage, client, occmClientId) => {
    const getOccmUsersPage = page => client.get('users', {
        params: {
            per_page: 100,
            page,
            include_totals: true,
            fields: 'email,user_id,user_metadata',
            include_fields: true,
            q: `app_metadata.occms:${occmClientId}`,
            search_engine: 'v3'
        }
    });

    const shouldDoAgain = data => (data.start + data.length < data.total);

    return getOccmUsersPage(lastPage).then(({ data }) => {
        const newAcc = acc.concat(data.users);
        if (shouldDoAgain(data)) {
            return getOccmUsersRecursive(newAcc, lastPage + 1, client, occmClientId);
        }
        return newAcc;
    });
};

const getAssociatedUsersOfOccm = exports.getAssociatedUsersOfOccm = ({ occmClientId }) => {
    console.log(`Attempting to get user data for all users associated with this occm client. ${JSON.stringify({ occmClientId })}`);
    const normalizedClientId = normalizeClientId(occmClientId);
    return authAxios().then(client => getOccmUsersRecursive([], 0, client, normalizedClientId)
        .then(res => {
            console.log("Success. Got list of users associated with given occm");
            return res.map(obj => ({
                userId: obj.user_id,
                name: obj.user_metadata.name,
                email: obj.email,
                company: obj.user_metadata.company
            }));
        }), err => {
        console.error(`Error. Get list of associated users for occm failed`, err);
        throw err;
    });
};

const getOccmUris = exports.getOccmUris = ({ occmClientId }) => {
    console.log(`Attempting to get all Uris for this occm client. ${JSON.stringify({ occmClientId })}`);
    const normalizedClientId = normalizeClientId(occmClientId);

    const params = {
        TableName: process.env.occmClientInfoTable,
        Key: { clientId: normalizedClientId }
    };

    const dynamoDb = new AWS.DynamoDB.DocumentClient();
    return dynamoDb.get(params).promise().then(response => {
        const occm = response.Item || null;
        if (occm) {
            console.log("Successfully returned OCCM info");
            return occm.automaticCallbackUris || [];
        }
        console.log("No OCCM found, returning empty list");
        return [];
    }, err => {
        console.error(`Error. Failed to return list of all Occm Uris`, err);
        throw err;
    });
};

const emailDomainInUse = emailDomain => { // case insensitive comparison
    const normalizedEmailDomain = emailDomain.toLowerCase();
    let result = false;
    return authAxios().then(client => client.get(`connections`)
        .then(({ data }) => {
            data.forEach(conn => {
                if (conn.strategy === 'ad') {
                    conn.options.domain_aliases.forEach(domain => {
                        if (domain.toLowerCase() === normalizedEmailDomain) {
                            result = true;
                        }
                    });
                }
            });
            return result;
        }));
};

const resendVerificationEmail = exports.resendVerificationEmail = ({ userToken }) => {
    console.log(`Attempting to resend a verification email for the user: ${userToken}`);
    return validateAndExtractToken(userToken, "https://cloudmanager.cloud.netapp.com")
        .then(
            payload => sendVerificationEmail(payload.sub),
            err => {
                console.error(`Error. Failed to validate token.`);
                return Promise.reject(new ClientError(`Payload token validation error, failed to authorize. Did not send verification email. ${err}`));
            }
        );
};

const migrateUser = exports.migrateUser = async ({
    occmClientId, email, fullName, company
}) => {
    console.log(`Attempting to migrate user with email: ${email} for occm with clientId: ${occmClientId}`);
    const normalizedClientId = normalizeClientId(occmClientId);
    const generator = require('generate-password'); /* eslint-disable-line global-require */

    try {
        console.log("Determining if user exists with given email");
        const authClient = await authAxios();
        // Note that this API (GET /users-by-email) is case-sensitive.
        // In auth0, the emails of User-Password users are always kept in lowercase, but emails of federated users keep
        // the original casing, so this API call might miss them.
        // Consider to refactor to use GET /users with query which is case-insensitive.
        const { data: users } = await authClient.get("users-by-email", { params: { email: email.toLowerCase() } });

        let userId = null;
        let userCreated = false;

        if (users.length === 0) {
            console.log("No user exists with email, creating new user");

            const { data: user } = await authClient.post(`users`, {
                connection: "Username-Password-Authentication",
                email,
                password: generator.generate({
                    length: 15, numbers: true, uppercase: true, strict: true
                }),
                verify_email: false,
                user_metadata: {
                    name: fullName,
                    company
                }
            });

            userId = user.user_id;
            userCreated = true;
            console.log(`Successfully created user, new userId ${userId}`);
        } else {
            userId = users[0].user_id;
            console.log(`${users.length} users exit with the given email, will use first user ${userId}`);
        }

        console.log(`Marking user as migrated in dynamo`);
        const dynamoDb = new AWS.DynamoDB.DocumentClient();
        const AttributeUpdates = {
            occmMigration: { Action: "PUT", Value: true }
        };

        if (userCreated) { AttributeUpdates.occmMigrationCreated = { Action: "PUT", Value: true }; }

        const userInfoParams = {
            TableName: process.env.userInfoTable,
            Key: { userId },
            AttributeUpdates
        };

        await dynamoDb.update(userInfoParams).promise();
        try {
            await associateUserInternal(userId, normalizedClientId);
        } catch (e) {
            console.log("WARNING: Association failed, but user creation successful.. API will return success.");
            console.log(e);
        }

        return {
            userId,
            userCreated
        };
    } catch (err) {
        console.log(`Error. Failed to migrate occm user`);
        console.log(err);
        throw err;
    }
};

const registerAgent = exports.registerAgent = ({ agentName, systemId, setupUserId }) => {
    console.log(`Attempting to register agent: ${JSON.stringify({ agentName, systemId, setupUserId })}`);
    const dynamoDb = new AWS.DynamoDB.DocumentClient();
    return authAxios().then(client => {
        const clientName = `Agent-Client-${systemId}`;
        const createDate = `${(new Date()).getTime()}`;

        return client.post("clients", {
            name: clientName,
            description: "Agent Client",
            grant_types: ["client_credentials", "password"],
            jwt_configuration: { alg: "RS256" },
            app_type: "non_interactive",
            allowed_origins: ["http://localhost:4300", "https://setup.occm.datafabric.io", "https://dev-setup.occm.datafabric.io", "https://setup.cloudmanager.netapp.com", "https://dev-setup.cloudmanager.netapp.com"],
            client_metadata: { createDate }
        }).then(({ data }) => {
            const clientId = data.client_id;
            const clientSecret = data.client_secret;
            const clientDynamoParams = {
                TableName: process.env.occmClientInfoTable,
                Item: {
                    clientId,
                    clientName,
                    setupUserId,
                    occmName: agentName,
                    saas: true,
                    agent: true,
                    createDate
                }
            };

            return Promise.all([
                // since this client include client_credential grants, we need to grant access to the https://api.cloud.netapp.com API
                // usually SPA client are not required to do so
                client.post("client-grants", {
                    client_id: clientId,
                    audience: "https://api.cloud.netapp.com",
                    scope: ["agents:agent"]
                }),

                // create client item in dynamo
                dynamoDb.put(clientDynamoParams).promise()
            ]).then(() => ({
                clientId,
                clientSecret
            }));
        }, err => { // call to auth0 post:clients failed
            const statusCode = _.get(err, 'response.status', null);
            if (statusCode === 400) {
                err.auth0ClientError = true;
            }
            throw err;
        });
    }).then(agent => {
        console.log(`Registration complete.`);
        const clone = _.clone(agent);
        clone.clientSecret = "****";
        console.log(clone);
        return agent;
    }, err => {
        if (err.auth0ClientError) {
            console.error(`Registration failed. Error returned from auth0.`, err);
            throw new ClientError("Call to Auth0 failed due to client error. Check passed parameters for duplicates or illegal characters.");
        }
        console.error(`Registration failed. Possible inconsistent state.`, err);
        throw new Error('Agent registration failed');
    });
};

const agentSetupComplete = exports.agentSetupComplete = ({ agentClientId }) => {
    console.log(`Attempting to mark agent setup as completed. agentClientId: ${agentClientId}`);
    const dynamoDb = new AWS.DynamoDB.DocumentClient();
    const normalizedClientId = normalizeClientId(agentClientId);
    const params = {
        TableName: process.env.occmClientInfoTable,
        Key: { clientId: normalizedClientId },
        ReturnValues: "ALL_NEW",
        AttributeUpdates: {
            setupComplete: { Action: "PUT", Value: true }
        }
    };

    return dynamoDb.update(params).promise()
        .then(({ Attributes: itemAttributes }) => Notifications.insertAgentSetupCompleteNotification(itemAttributes.setupUserId, itemAttributes.occmName))
        .then(() => {
            console.log(`Agent setup completion succeeded.`);
            return null;
        }, err => {
            console.error(`Agent setup completion failed`, err);
            throw err;
        });
};

const agentSetupFailed = exports.agentSetupFailed = ({ agentClientId, setupUserId }) => {
    console.log(`Attempting to mark agent setup as failed. ${JSON.stringify({ agentClientId, setupUserId })}`);
    const dynamoDb = new AWS.DynamoDB.DocumentClient();
    const normalizedClientId = normalizeClientId(agentClientId);
    let occmName;
    const clientInfoParams = {
        TableName: process.env.occmClientInfoTable,
        Key: { clientId: normalizedClientId }
    };
    return dynamoDb.get(clientInfoParams).promise()
        .then(clientInfoResponse => {
            const clientInfo = clientInfoResponse.Item || {};
            occmName = _.get(clientInfo, "occmName");
            if (!clientInfo.agent) {
                return Promise.reject(new ForbiddenError(`Client ${agentClientId} is not an agent`));
            }
            if (clientInfo.setupUserId !== setupUserId) {
                return Promise.reject(new ForbiddenError(`User ${setupUserId} does not have permission to affect ${normalizedClientId}`));
            }
            const params = {
                TableName: process.env.occmClientInfoTable,
                Key: { clientId: normalizedClientId },
                AttributeUpdates: {
                    setupFailed: { Action: "PUT", Value: true }
                }
            };
            return dynamoDb.update(params).promise();
        }).then(() => {
            console.log(`Marking agent setup as failed - succeeded.`);
            return Notifications.insertOccmSetupFailedNotification(setupUserId, occmName);
        }, err => {
            console.error(`Marking agent setup as failed - failed.`, err);
            throw err;
        });
};

const routes = [
    {
        method: "POST",
        scopes: ["occm-access"],
        pattern: "/occm/associate-user",
        bodySchema: Joi.object().keys({
            associateeEmail: Joi.string().required(),
            associaterId: Joi.string().required()
        }),
        handler: ({ principalId, body }) => associateUser({
            occmClientId: principalId,
            associateeEmail: body.associateeEmail,
            associaterId: body.associaterId
        })
    },
    {
        method: "POST",
        scopes: ["occm-access"],
        pattern: "/occm/deassociate-user",
        bodySchema: Joi.object().keys({
            deassociaterId: Joi.string().required(),
            deassociateeId: Joi.string().required()
        }),
        handler: ({ principalId, body }) => deassociateUser({
            occmClientId: principalId,
            deassociaterId: body.deassociaterId,
            deassociateeId: body.deassociateeId,
            notificationType: "deassociate"
        })
    },
    {
        method: "POST",
        pattern: "/occm/deassociate-self",
        bodySchema: Joi.object().keys({
            occmId: Joi.string().required()
        }),
        handler: ({ principalId, body }) => deassociateUser({
            occmClientId: body.occmId,
            deassociaterId: principalId,
            deassociateeId: principalId,
            notificationType: "remove"
        }),
        doc: {
            summary: "Remove specified Cloud Manager from user's list of Cloud Managers.",
            responses: { 204: "#ref" }
        }
    },
    {
        method: "GET",
        pattern: "/occm/list-occms",
        handler: ({ principalId }) => listOccmsOfUser({ userId: principalId }),
        doc: {
            summary: "Lists all Cloud Managers to which the user has access",
            responses: {
                200: "occms-response.yaml"
            }
        }
    },
    {
        method: "POST",
        scopes: ["occm-create"],
        pattern: "/occm/register-occm",
        bodySchema: Joi.object().keys({
            redirectUris: Joi.array().required(),
            systemId: Joi.string().required()
        }),
        handler: ({ body }) => registerOccm({ redirectUris: body.redirectUris, systemId: body.systemId })
    },
    {
        method: "POST",
        scopes: ["occm-access"],
        pattern: "/occm/setup-complete",
        bodySchema: Joi.object().keys({
            occmName: Joi.string().required(),
            setupUserId: Joi.string().required()
        }),
        handler: ({ principalId, body }) => occmSetupComplete({
            setupUserId: body.setupUserId,
            occmClientId: principalId,
            occmName: body.occmName
        })
    },
    {
        method: "POST",
        scopes: ["occm-access"],
        pattern: "/occm/update-uris",
        bodySchema: Joi.object().keys({
            redirectUris: Joi.array().required()
        }),
        handler: ({ principalId, body }) => updateOccmUris({ redirectUris: body.redirectUris, occmClientId: principalId })
    },
    {
        method: "POST",
        pattern: "/occm/manual-update-uris",
        bodySchema: Joi.object().keys({
            redirectUris: Joi.array().items(Joi.object().keys({
                isDefault: Joi.boolean(),
                value: Joi.string().required()
            })).required(),
            occmId: Joi.string().required()
        }),
        handler: ({ principalId, body }) => manualUpdateOccmUris({ userId: principalId, manualOverrideUris: body.redirectUris, occmId: body.occmId }),
        doc: {
            summary: "Provide a list of redirect URIs valid for log in with a specific Cloud Manager instance.",
            responses: { 204: "#ref" }
        }
    },
    {
        method: "POST",
        pattern: "/occm/service-register-occm",
        bodySchema: Joi.object().keys({
            occmName: Joi.string().required(),
            systemId: Joi.string().required(),
            isAgent: Joi.boolean()
        }),
        handler: ({ principalId, body }) => serviceRegisterOccm({
            occmName: body.occmName,
            systemId: body.systemId,
            isAgent: body.isAgent,
            setupUserId: principalId
        }),
        doc: { hidden: true }
    },
    {
        method: "POST",
        pattern: "/occm/service-setup-failed",
        bodySchema: Joi.object().keys({
            clientId: Joi.string().required()
        }),
        handler: ({ principalId, body }) => serviceSetupFailed({ setupUserId: principalId, occmClientId: body.clientId }),
        doc: { hidden: true }
    },
    {
        // this API is deprecated but still in use in OCCM version 3.4 and should not be deleted
        method: "GET",
        pattern: "/occm/list-users",
        scopes: ["occm-access"],
        handler: ({ principalId }) => getAssociatedUsersOfOccm({ occmClientId: principalId })
    },
    {
        method: "GET",
        pattern: "/occm/users",
        scopes: ["occm-access"],
        querySchema: Joi.object().keys({
            usersString: Joi.string().required()
        }),
        handler: ({ principalId, queryParams }) => getOccmUsers({ occmClientId: principalId, usersString: queryParams.usersString })
    },
    {
        method: "GET",
        pattern: "/occm/list-uris",
        scopes: ["occm-access"],
        handler: ({ principalId }) => getOccmUris({ occmClientId: principalId })
    },
    {
        method: "POST",
        pattern: "/occm/send-verification-email",
        scopes: ["occm-access"],
        bodySchema: Joi.object().keys({
            userToken: Joi.string().required()
        }),
        handler: ({ body }) => resendVerificationEmail({ userToken: body.userToken })
    },
    {
        method: "POST",
        pattern: "/occm/migrate-user",
        scopes: ["occm-access"],
        bodySchema: Joi.object().keys({
            email: Joi.string().required(),
            fullName: Joi.string().required(),
            company: Joi.string().required()
        }),
        handler: ({ principalId, body }) => migrateUser({
            occmClientId: principalId, email: body.email, fullName: body.fullName, company: body.company
        })
    },
    {
        method: "POST",
        pattern: "/occm/register-agent",
        bodySchema: Joi.object().keys({
            agentName: Joi.string().required(),
            systemId: Joi.string().required()
        }),
        handler: ({ principalId, body }) => registerAgent({
            agentName: body.agentName,
            systemId: body.systemId,
            setupUserId: principalId
        }),
        doc: { hidden: true }
    },
    {
        method: "POST",
        pattern: "/occm/agent-setup-complete",
        // scopes: ["agents:agent"],
        handler: ({ principalId }) => agentSetupComplete({ agentClientId: principalId }),
        doc: { hidden: true }
    },
    {
        method: "POST",
        pattern: "/occm/agent-setup-failed",
        bodySchema: Joi.object().keys({
            clientId: Joi.string().required()
        }),
        handler: ({ principalId, body }) => agentSetupFailed({ setupUserId: principalId, agentClientId: body.clientId }),
        doc: { hidden: true }
    }
];

exports.handler = (event, context, callback) => {
    dispatch({
        event, callback, routes, context
    });
};

exports.handler.routes = routes;
exports.Notifications = Notifications;

const yarger = require("yargs") // eslint-disable-line no-unused-vars, import/order
    .command("associate-user", "associate-user", yargs => {
        yargs.demandOption(['associateeEmail', 'associaterId', 'occmClientId']);
    }, argv => {
        associateUser({
            associateeEmail: argv.associateeEmail,
            associaterId: argv.associaterId,
            occmClientId: argv.occmClientId
        });
    })
    .command("deassociate-user", "deassociate-user", yargs => {
        yargs.demandOption(['deassociateeId', 'deassociaterId', 'occmClientId']);
    }, argv => {
        deassociateUser({
            deassociateeId: argv.deassociateeId,
            deassociaterId: argv.deassociaterId,
            occmClientId: argv.occmClientId
        });
    })
    .command("list-occms", "list-occms", yargs => {
        yargs.demandOption(['userId']);
    }, argv => {
        listOccmsOfUser({ userId: argv.userId });
    })
    .command("register-occm", "register-occm", yargs => {
        yargs.demandOption(["redirectUris", "systemId"]);
    }, argv => {
        registerOccm({ redirectUris: argv.redirectUris.split(","), systemId: argv.systemId });
    })
    .command("service-setup-failed", "service-setup-failed", yargs => {
        yargs.demandOption(["occmClientId", "setupUserId"]);
    }, argv => {
        process.env.userInfoTable = "portal-backend-dev-userInfo";
        process.env.occmClientInfoTable = "portal-backend-dev-occmClientInfo";
        serviceSetupFailed({ occmClientId: argv.occmClientId, setupUserId: argv.setupUserId });
    })
    .command("occm-setup-complete", "occm-setup-complete", yargs => {
        yargs.demandOption(['setupUserId', "occmClientId", "occmName"]);
    }, argv => {
        occmSetupComplete({ setupUserId: argv.setupUserId, occmClientId: argv.occmClientId, occmName: argv.occmName });
    })
    .command("update-occm-uris", "update-occm-uris", yargs => {
        yargs.demandOption(['redirectUris', "occmClientId"]);
    }, argv => {
        updateOccmUris({ redirectUris: argv.redirectUris.split(","), occmClientId: argv.occmClientId });
    })
    .command("list-users", "list-users", yargs => {
        yargs.demandOption(["occmClientId"]);
    }, argv => {
        getAssociatedUsersOfOccm({ occmClientId: argv.occmClientId });
    })
    .command("migrate-user", "migrate-user", yargs => {
        yargs.demandOption(["occmClientId", "email", "fullName", "company"]);
    }, ({
        occmClientId, email, fullName, company
    }) => {
        migrateUser({
            occmClientId, email, fullName, company
        }).then(result => console.log(result));
    })
    .command("check-email", "check-email", yargs => {
        yargs.demandOption(["emailDomain"]);
    }, argv => {
        emailDomainInUse(argv.emailDomain);
    }).argv;
