const _ = require("lodash");
const Joi = require('joi');
const AWS = require('aws-sdk');
const axios = require('axios');
const csvToJson = require('csvtojson');

const { authAxios } = require("./auth-utils");
const dispatch = require('./dispatch');
const { getUserNotifications, setNotificationVisibility } = require("./notifications");
const { ClientError } = require("./errors");
const assessmentHandler = require("./assessments/assessmentHandler");
const { sendInternalEmail } = require('./utils/internal-email');
const { occmDemoMailList } = require('./consts');
const { getUsersDetailsByIds, getUserByEmail } = require("./common");
// eslint-disable-line import/no-extraneous-dependencies
AWS.config.update({ region: 'us-east-1' });
AWS.config.setPromisesDependency(require('bluebird'));


const updateUserInfo = exports.updateUserInfo = ({ userId, userName }) => {
    console.log(`Attempting to update user info for user: ${userId} with name: ${userName}`);

    return authAxios().then(client => client.patch(`users/${userId}`, {
        user_metadata: { name: userName }
    })).then(() => {
        console.log(`Update user info succeeded.`);
        return null;
    }, err => {
        console.error(`Update user info failed`, err);
        throw err;
    });
};

const getNotifications = exports.getNotifications = ({ userId }) => {
    console.log(`Attempting to get all visible ('new' and 'viewed') notifications for user ${userId}`);
    return getUserNotifications(userId);
};

const notificationVisibility = exports.notificationVisibility = ({ userId, notificationsList, visibility }) => {
    console.log(`Attempting to set notification(s) for user ${userId} to ${visibility}`);

    if (visibility !== 'viewed' && visibility !== 'deleted') {
        console.log(`Unexpected visibility: ${visibility}. Notifications not updated.`);
        return Promise.reject(new ClientError(`Unexpected visibility: "${visibility}". Must be either 'viewed' or 'deleted'`));
    }
    const notificationsPromises = notificationsList.map(notification => setNotificationVisibility(notification.userId, notification.createDate, visibility).then(() => ({ createDate: notification.createDate, updatedStatus: true }), err => ({ createDate: notification.createDate, updatedStatus: false, message: _.get(err, "response.data.message") || err.message || "Unknown" })));
    return Promise.all(notificationsPromises);
};

const createQstackUser = exports.createQstackUser = ({
    email, password, name, company
}) => {
    console.log(`Attempting to a new qstack Azure NFSaaS user for user: ${email}`);
    return authAxios().then(client => client.post(`users`, {
        connection: "Username-Password-Authentication",
        email,
        password,
        verify_email: false,
        user_metadata: {
            name: name || "Azure NFSaaS User",
            company: company || "Qstack"
        },
        app_metadata: {
            qstack: "Azure NFSaaS"
        }
    }).then(({ data }) => {
        console.log(`Successfully created qstack Azure NFSaaS user`);
        return Promise.resolve(data);
    }), err => {
        console.error(`Error. Failed to create qstack Azure NFSaaS user`, err);
        throw err;
    });
};

const refreshToken = exports.refreshToken = async ({ userId }) => {
    console.log(`Attempting return user ${userId}'s refresh tokens`);
    const dynamoDb = new AWS.DynamoDB.DocumentClient();
    const params = {
        TableName: process.env.refreshTokenTable,
        Key: { userId }
    };
    try {
        const dynamoResponse = await dynamoDb.get(params).promise();
        console.log(dynamoResponse.Item);
        return dynamoResponse.Item || null;
    } catch (err) {
        throw err;
    }
};

const revokeRefreshToken = exports.revokeRefreshToken = async ({ userId }) => {
    console.log(`Attempting to revoke refresh token for user ${userId}`);
    try {
        const dynamoDb = new AWS.DynamoDB.DocumentClient();
        const client = await authAxios();
        const { data: userGrants } = await client.get('grants', {
            params: { user_id: userId, client_id: process.env.managementRefreshTokenId }
        });
        console.log(userGrants);
        if (!userGrants) {
            console.log('User does not have any refresh tokens. Nothing to delete.');
        } else {
            return Promise.all(_.map(userGrants, grant => client.delete(`grants/${grant.id}`, {
                params: { user_id: userId }
            })
                .then(res => console.log(`Auth0 delete response: ${res.status}`))))
                .then(async () => {
                    const deleteParams = {
                        TableName: process.env.refreshTokenTable,
                        Key: { userId }
                    };
                    await dynamoDb.delete(deleteParams).promise();
                    console.log('Refresh token(s) deleted from dynamo');
                    return null;
                });
        }
        console.log('Refresh token(s) revoked');
        return null;
    } catch (err) {
        console.log(err);
        throw new Error('Failed to revoke refresh token(s)');
    }
};

const occmDemoEnvSignUp = exports.occmDemoEnvSignUp = async ({ userId }) => {
    console.log(`Attempting to sign up user to occm demo environment. Getting user info from Auth0...`);
    const email = process.env.chaimsohnEmail;
    const password = process.env.chaimsohnPassword;
    const roleId = process.env.chaimsohnRoleId;
    const accessKey = process.env.chaimsohnAccessKey;
    const secretKey = process.env.chaimsohnSecretKey;
    const baseURL1 = process.env.chaimsohnBaseUrl1;
    const baseURL2 = process.env.chaimsohnBaseUrl2;
    let userEmail = '';
    const userPassword = process.env.chaimsohnUserPassword;
    try {
        const client = await authAxios();
        const { data: user } = await client.get(`users/${userId}`);
        const nameArr = _.get(user, 'user_metadata.name', 'unknown').trim().split(" ");
        const firstName = user.given_name || nameArr[0];
        const lastName = user.family_name || nameArr[nameArr.length - 1];
        userEmail = user.email;
        const tenantName = _.get(user, 'nameIdAttributes.value', userEmail.substring(0, userEmail.indexOf('@'))).replace(/\W/g, '');
        console.log(`Requesting user information: id: ${userId}, firstName: ${firstName}, lastName: ${lastName}, email: ${userEmail}, tenantName: ${tenantName}`);
        console.log(`Authenticating in origus...`);
        const loginResp = await axios.post(`${baseURL1}/auth/login`, {
            email, password
        });
        const cookie = _.get(loginResp, 'headers.set-cookie', null);
        const demoEnvAPI = axios.create({
            baseURL: baseURL1,
            withCredentials: true,
            headers: {
                'Content-Type': 'application/json',
                Cookie: cookie.join('; ')
            }
        });
        console.log(`Successfully logged in. Creating user tenant...`);
        const tenantsResp = await demoEnvAPI.post('/tenants', {
            name: tenantName, description: `${firstName} ${lastName}`
        });
        const tenantId = _.get(tenantsResp, 'data.publicId');
        console.log(`Successfully created tenant with tenantId: ${tenantId}. Creating user...`);

        await demoEnvAPI.post('/users', {
            firstName, lastName, email: userEmail, roleId, tenantId, password: userPassword, accessKey, secretKey
        });

        console.log('Completed origus flow. Authenticating in demous...');
        const loginResp2 = await axios.post(`${baseURL2}/auth/login`, {
            email, password
        });
        const cookie2 = _.get(loginResp2, 'headers.set-cookie', null);
        const demoEnvAPI2 = axios.create({
            baseURL: baseURL2,
            withCredentials: true,
            headers: {
                'Content-Type': 'application/json',
                Cookie: cookie2.join('; ')
            }
        });
        console.log(`Successfully logged in. Creating user tenant...`);
        const tenantsResp2 = await demoEnvAPI2.post('/tenants', {
            name: tenantName, description: `${firstName} ${lastName}`
        });
        const tenantId2 = _.get(tenantsResp2, 'data.publicId');
        console.log(`Successfully created tenant with tenantId: ${tenantId2}. Creating user...`);

        await demoEnvAPI2.post('/users', {
            firstName,
            lastName,
            email: userEmail,
            roleId,
            tenantId: tenantId2,
            password: userPassword,
            azureCredentials: {
                subscriptionId: '1', tenantId: '1', applicationId: '1', applicationKey: '1'
            },
            accessKey: '1',
            secretKey: '1'
        });
        console.log(`Success. User created in demo environment`);
        if (process.env.stage === 'prod') {
            const body = `An occm demo account has been created for user: ${userEmail}.`;
            await sendInternalEmail(`Added ${userEmail} to demo OCCM`, body, occmDemoMailList);
        }
        return { username: userEmail, password: userPassword, created: true };
    } catch (err) {
        const causeMessage = _.get(err, 'response.data.causeMessage', null);
        const errDescription = _.get(err, 'response.data', null);
        console.log(causeMessage || errDescription || err);
        if (causeMessage && (causeMessage.startsWith('TenantNameAlreadyExistsException') || causeMessage.startsWith('UserFacade.EmailAlreadyExistsException'))) {
            return {
                username: userEmail, password: userPassword, created: false, message: causeMessage
            };
        }
        throw err;
    }
};

const readFromS3Bucket = ({ bucket, key, encoding = 'ascii' }) => {
    console.log(`Reading file content from bucket: ${bucket}, key: ${key}.`);

    const s3 = new AWS.S3();
    const params = { Bucket: bucket, Key: key };

    return s3.getObject(params).promise()
        .then(data => {
            console.log(`Successfully read file content from S3 bucket. ContentLength: ${data.ContentLength}. LastModified: ${data.LastModified}.`);

            return data.Body.toString(encoding);
        }, err => {
            console.log('Error reading file from S3 bucket:');
            console.log(err);

            throw err;
        });
};

const getAmericasGlobalSpiffData = async () => {
    console.log("Returning Americas Global Spiff Data - fetching CSV from s3");

    const accounts = await readFromS3Bucket({ bucket: process.env.privateBucket, key: process.env.americasGlobalSpiffAccounts, encoding: 'utf-8' }).then(contents => csvToJson().fromString(contents));

    const teamChallenge = await readFromS3Bucket({ bucket: process.env.privateBucket, key: process.env.americasGlobalSpiffTeamChallenge, encoding: 'utf-8' }).then(contents => csvToJson().fromString(contents));

    const partnersReport = await readFromS3Bucket({ bucket: process.env.privateBucket, key: process.env.americasGlobalSpiffPartnersReport, encoding: 'utf-8' }).then(contents => csvToJson().fromString(contents));

    const individualChallengeRules = await readFromS3Bucket({ bucket: process.env.privateBucket, key: process.env.americasGlobalSpiffIndividualChallengeRules, encoding: 'utf-8' }).then(contents => csvToJson().fromString(contents));

    const individualChallengeRunners = await readFromS3Bucket({ bucket: process.env.privateBucket, key: process.env.americasGlobalSpiffIndividualChallengeRunners, encoding: 'utf-8' }).then(contents => csvToJson().fromString(contents));

    return {
        accounts, teamChallenge, individualChallengeRules, individualChallengeRunners, partnersReport
    };
};

const getCustomerSuccessStories = () => {
    console.log('Returning Customer Success Stories as CSV.');

    return readFromS3Bucket({ bucket: process.env.privateBucket, key: process.env.customerSuccessStoriesKey });
};

const updateFileContentInS3 = ({ bucket, key, body }) => {
    console.log(`Updating file content in s3 bucket: ${bucket}, key: ${key}.`);

    const s3 = new AWS.S3();
    const params = { Bucket: bucket, Key: key, Body: body };

    return s3.putObject(params).promise()
        .then(() => {
            console.log('Successfully updated file content in S3');
        }, err => {
            console.log('Error updating file to S3:');
            console.log(err);

            throw err;
        });
};

const reportNonPrintableChars = string => {
    const justPrintableCharsRegex = /[^ -~]/;

    return _(string)
        .split('\r\n')
        .map((rowText, rowNumber) => {
            const locationOfIllegalChar = rowText.search(justPrintableCharsRegex);
            return (locationOfIllegalChar !== -1) ? `{line: ${rowNumber + 1}, col: ${locationOfIllegalChar + 1}}` : null;
        })
        .compact()
        .join();
};

const updateCustomerSuccessStories = ({ body }) => {
    console.log('Updating Customer Success Stories CSV.');

    const nonPrintableCharsMessage = reportNonPrintableChars(body.content);
    if (nonPrintableCharsMessage) {
        return Promise.reject(new ClientError(`CSV file contains illegal whitespace at ${nonPrintableCharsMessage}. Only spaces and newlines allowed.`));
    }

    if (!body.content.startsWith(process.env.customerSuccessStoriesHeaders)) {
        return Promise.reject(new ClientError('CSV file does not have expected column headers.'));
    }

    return updateFileContentInS3({
        bucket: process.env.privateBucket,
        key: process.env.customerSuccessStoriesKey,
        body: body.content
    });
};

const getUserIdByEmail = exports.getUserIdByEmail = async ({ email }) => {
    console.log(`Attempting to get userId for user with email ${email}`);
    const user = await getUserByEmail({ email });

    if (!user) {
        throw new ClientError(`No user has the given email address: ${email}`);
    }
    return { userId: user.user_id };
};

const doesUserExist = exports.doesUserExist = async ({ email }) => {
    console.log(`Attempting to check if user with email ${email} exists.`);
    const user = await getUserByEmail({ email });

    return { doesExist: !!user };
};

const accountRoutes = [
    {
        method: "POST",
        pattern: "/portal/update-user-info",
        bodySchema: Joi.object().keys({
            userName: Joi.string()
        }),
        handler: ({ principalId, body }) => updateUserInfo({ userId: principalId, userName: body.userName }),
        doc: {
            summary: "Update a user's name",
            responses: { 204: "#ref" }
        }
    },
    {
        method: "GET",
        pattern: "/portal/refresh-token",
        handler: ({ principalId }) => refreshToken({ userId: principalId }),
        doc: {
            summary: "Retrieve a user's refresh tokens"
        }
    },
    {
        method: "POST",
        pattern: "/portal/revoke-refresh-token",
        handler: ({ principalId }) => revokeRefreshToken({ userId: principalId }),
        doc: {
            summary: "Revoke all of a user's refresh tokens"
        }
    },
    {
        method: "GET",
        scopes: ["cc:tenancy"],
        pattern: "/portal/tenancy/users",
        querySchema: Joi.object().keys({
            userIds: Joi.string()
        }),
        handler: ({ queryParams }) => getUsersDetailsByIds({ userIds: queryParams.userIds }),
        doc: { hidden: true }
    },
    {
        method: "GET",
        scopes: ["cc:tenancy"],
        pattern: "/portal/tenancy/user-id-by-email",
        querySchema: Joi.object().keys({
            email: Joi.string()
        }),
        handler: ({ queryParams }) => getUserIdByEmail({ email: queryParams.email }),
        doc: { hidden: true }
    }
];

const notificationsRoutes = [{
    method: "GET",
    pattern: "/portal/notifications",
    handler: ({ principalId }) => getNotifications({ userId: principalId }),
    doc: {
        summary: "List visible notifications (all those which have not been marked as deleted)",
        responses: {
            200: "notifications-response.yaml"
        }
    }
},
{
    method: "POST",
    pattern: "/portal/set-notification-visibility",
    bodySchema: Joi.object().keys({
        notificationsList: Joi.array().items(Joi.object()).description("array containing notification objects whose visibility is to be changed").required(),
        visibility: Joi.string().description("new visibility to set for the notifications").required()
    }),
    handler: ({ principalId, body }) => notificationVisibility({ userId: principalId, notificationsList: body.notificationsList, visibility: body.visibility }),
    doc: {
        summary: "Marks specified notifications as per the visibility property passed in the body.",
        responses: {
            204: "#ref",
            400: { description: "An error message indicating the user passed in an unexpected visibility" }
        }
    },
    audit: {
        exclude: true
    }
}];

const miscRoutes = [
    {
        method: "POST",
        scopes: ["qstack-create"],
        pattern: "/portal/create-qstack-user",
        bodySchema: Joi.object().keys({
            email: Joi.string().required(),
            password: Joi.string().required(),
            name: Joi.string(),
            company: Joi.string()
        }),
        handler: ({ body }) => createQstackUser({
            email: body.email, password: body.password, name: body.name, company: body.company
        })
    },
    {
        method: "POST",
        pattern: "/portal/occm-demo-signup",
        handler: ({ principalId }) => occmDemoEnvSignUp({ userId: principalId }),
        doc: { hidden: true }
    },
    {
        method: "GET",
        pattern: "/portal/does-user-exist",
        querySchema: Joi.object().keys({
            email: Joi.string()
        }),
        handler: ({ queryParams }) => doesUserExist({ email: queryParams.email }),
        doc: { hidden: true }
    }
];

const assessmentRoutes = [
    {
        method: "POST",
        pattern: "/portal/assessment/assessment-finished",
        bodySchema: Joi.object().keys({}),
        handler: ({ principalId, body }) => assessmentHandler.assessmentFinished({ userId: principalId, body })
    },
    {
        method: "POST",
        pattern: "/portal/assessment/assessment-failed",
        bodySchema: Joi.object().keys({}),
        handler: ({ principalId, body }) => assessmentHandler.assessmentFailed({ userId: principalId, body })
    }];

const customerSuccessRoutes = [
    {
        method: "GET",
        pattern: "/portal/customer-success-stories",
        scopes: ['cc:netapp-internal'],
        handler: getCustomerSuccessStories
    },
    {
        method: "PUT",
        pattern: "/portal/customer-success-stories",
        bodySchema: Joi.object().keys({
            content: Joi.string().max(50000)
        }),
        scopes: ['cc:netapp-internal', 'cc:customer-stories'],
        handler: updateCustomerSuccessStories
    }
];

const americasGlobalSpiffRoutes = [
    {
        method: "GET",
        pattern: "/portal/americas-global-spiff",
        scopes: ['cc:netapp-internal'],
        handler: getAmericasGlobalSpiffData
    }
];

exports.handler = (event, context, callback) => {
    dispatch({
        event,
        callback,
        context,
        routes: accountRoutes
            .concat(notificationsRoutes)
            .concat(miscRoutes)
            .concat(assessmentRoutes)
            .concat(customerSuccessRoutes)
            .concat(americasGlobalSpiffRoutes)
    });
};

exports.handler.notificationsRoutes = notificationsRoutes;
exports.handler.accountRoutes = accountRoutes;

const yarger = require("yargs") // eslint-disable-line no-unused-vars, import/order
    .command("set-notification-visibility", "set-notification-visibility", yargs => {
        yargs.demandOption(["userId", "notificationsList", "visibility"]);
    }, argv => {
        notificationVisibility(argv.userId, argv.notificationsList.split(","), argv.visibility);
    })
    .command("update-customer-success", "update-customer-success", yargs => {
        yargs.demandOption(["body"]);
    }, argv => updateCustomerSuccessStories({ body: argv.body }))
    .command("get-customer-success", "get-customer-success", _.noop, () => getCustomerSuccessStories())
    .command("occm-demo-signup", "occm-demo-signup", yargs => {
        yargs.demandOption(["userId"]);
    }, argv => {
        occmDemoEnvSignUp({ userId: argv.userId });
    }).argv;
