const _ = require('lodash');
const elasticsearch = require('elasticsearch');
const uuidv4 = require('uuid/v4');
const moment = require('moment');
const util = require('util');
const Promise = require('bluebird');
const zlib = require('zlib');

const retryLimit = 3;
const elasticSearchIndexPrefix = `${process.env.stage}-`;

const constructDateStamp = date => `${date.getUTCFullYear()}.${date.getUTCMonth() + 1}.${date.getUTCDate()}`;

const elasticClient = new elasticsearch.Client({
    host: process.env.elasticSearchIp,
    apiVersion: '6.0'
});

const auditNormalizer = data => {
    data.date = moment(data.timestamp).format();

    if (_.get(data, "auditContext.pattern") === "/analytics/promo/:promoId" && data.type === "AUDIT-CLOSE") {
        const userTrack = _.get(data, "data.queryParams.track");
        if (userTrack) {
            const userId = Buffer.from(decodeURIComponent(userTrack), 'base64').toString();

            data.normalized = {
                type: "promo-clicked",
                promoId: _.get(data, "data.urlParams.promoId"),
                userId
            };
        }
    } else if (_.get(data, "auditContext.pattern") === "/promo/track-view/:promoId" && data.type === "AUDIT-CLOSE") {
        const userId = _.get(data, "auditContext.requestData.principalId");
        const promoId = _.get(data, "auditContext.requestData.urlParams.promoId");

        data.normalized = {
            type: "promo-viewed",
            promoId,
            userId
        };
    }
};

const processObject = data => {
    if (data === null) { return null; }

    const keys = Object.keys(data);
    keys.forEach(key => {
        const val = data[key];
        let newKey = key.replace(/\./g, '_').replace(/^\$/, '_');

        if (_.isString(val)) {
            newKey = `s_${newKey}`;
        } else if (_.isPlainObject(val)) {
            newKey = `o_${newKey}`;
            processObject(val);
        } else if (_.isBoolean(val)) {
            newKey = `b_${newKey}`;
        } else if (_.isNumber(val)) {
            newKey = `n_${newKey}`;
        } else if (_.isArray(val)) {
            const prefix = processArray(val); // eslint-disable-line no-use-before-define
            newKey = `${prefix}_${newKey}`;
        }

        if (newKey !== key) {
            data[newKey] = val;
            delete data[key];
        }
    });

    return data;
};

const processArray = data => {
    if (data.length === 0) { return data; }

    const firstItem = data[0];
    let functionCheck = null;
    let prefix = null;
    if (_.isString(firstItem)) {
        functionCheck = _.isString;
        prefix = "a_s";
    } else if (_.isPlainObject(firstItem)) {
        functionCheck = _.isPlainObject;
        prefix = "a_o";
    } else if (_.isBoolean(firstItem)) {
        functionCheck = _.isBoolean;
        prefix = "a_b";
    } else if (_.isNumber(firstItem)) {
        functionCheck = _.isNumber;
        prefix = "a_n";
    } else if (_.isArray(firstItem)) {
        functionCheck = _.isArray;
        prefix = "a_a";
    } else {
        prefix = "a_unknown";
    }

    const allSame = _.every(data, functionCheck);

    if (!allSame) {
        prefix = "a_o";
        _.forEach(data, (val, index, array) => {
            array[index] = { _wrapped_s: util.inspect(val, { showHidden: true, depth: null }) };
        });
    }

    return prefix;
};

const auditCleaner = data => {
    processObject(_.get(data, "auditContext.requestData.queryParams", null));
    processObject(_.get(data, "auditContext.requestData.urlParams", null));
    processObject(_.get(data, "auditContext.requestData.body", null));
    processObject(_.get(data, "data.queryParams", null));
    processObject(_.get(data, "data.urlParams", null));
    const body = _.get(data, "data.body", null);
    if (body && !_.isObject(body)) {
        data.data.body = { body };
    }
    processObject(_.get(data, "data.body", null));
};

const handleAudit = exports.handleAudit = data => {
    const date = new Date(data.timestamp);
    auditNormalizer(data);
    auditCleaner(data);
    const uuid = uuidv4();
    const index = `audit-${elasticSearchIndexPrefix}${constructDateStamp(date)}`;
    if (data.body && !_.isPlainObject(data.body)) {
        data.body = {
            _wrapped: data.body
        };
    }
    return [{ create: { _id: uuid, _index: index, _type: "audit" } }, data];
};

const handleAuth0Rate = data => {
    data.date = moment(data.timestamp).format();

    return [{ create: { _id: uuidv4(), _index: `rate-auth-${elasticSearchIndexPrefix}`, _type: "rate" } }, data];
};

const messageRegex = /([^\t]+)\t([^\t]+)\t([\s\S]*)/;
const wrapperRegex = /(?:START|END|REPORT)\sRequestId: ([^\s]+)/;

const handleLogMessages = exports.handleLogMessages = data => _.flatten(data.logEvents.map(logEvent => {
    const date = new Date(logEvent.timestamp);
    const index = `log-${elasticSearchIndexPrefix}${constructDateStamp(date)}`;

    let { message } = logEvent;
    let requestId = "unknown";

    const wrapperMatch = logEvent.message.match(wrapperRegex);
    const messageMatch = logEvent.message.match(messageRegex);

    if (wrapperMatch) {
        [, requestId] = wrapperMatch;
        if (logEvent.message.startsWith("START")) {
            message = "START";
        } else if (logEvent.message.startsWith("END")) {
            message = "END";
        } else {
            message = "REPORT";
        }
    } else if (messageMatch) {
        [, , requestId, message] = messageMatch;
    }

    const entry = {
        timestamp: logEvent.timestamp,
        date: moment(logEvent.timestamp).format(),
        logGroup: data.logGroup.substring(27),
        message: message.endsWith("\n") ? message.substring(0, message.length - 1) : message,
        requestId
    };

    return [{ create: { _id: logEvent.id, _index: index, _type: "log" } }, entry];
}));

const formatRecords = exports.formatRecords = records => Promise.all(records.map(record => {
    if (record.eventSource === 'aws:kinesis') {
        const inputStream = Buffer.from(record.kinesis.data, 'base64');

        return Promise
            .try(() => JSON.parse(inputStream.toString()))
            .catch(() => JSON.parse(zlib.gunzipSync(inputStream).toString()))
            .catch(() => null)
            .then(data => {
                if (!data) {
                    console.log("Failed to parse input as JSON with and without gunzip");
                    console.log(record.kinesis);
                    return [];
                }
                if (data.type === "AUDIT-OPEN" || data.type === "AUDIT-CLOSE") {
                    return handleAudit(data);
                }
                if (data.type === "AUTH-RATE") {
                    return handleAuth0Rate(data);
                }
                if (data.messageType === "DATA_MESSAGE" && data.logEvents) {
                    return handleLogMessages(data);
                }
                console.log("Unhandled data type");
                console.log(data);
                return [];
            })
            .catch(err => {
                console.log("Formatting record failed");
                console.log(record.kinesis.data);
                console.log(err);
                return [];
            });
    }

    return Promise.resolve([]);
})).then(_.flatten);

const sendRecordsToElasticSearch = records => {
    const retrier = times => {
        const promise = elasticClient.bulk({ body: records });

        return promise.then(response => {
            if (response.errors) {
                const failedElements = response.items;
                console.log("Failed to push specific events to elastic search");
                console.log(util.inspect(failedElements, { showHidden: true, depth: null }));
                console.log(util.inspect(records, { showHidden: true, depth: null }));
            }
            return null;
        }, err => {
            console.log("Failed to push all events to elastic search");
            console.log(util.inspect(records, { showHidden: true, depth: null }));
            console.log(err);

            if (times < retryLimit) {
                console.log("Retrying in 3 seconds");
                return Promise.delay(3000).then(() => retrier(times + 1));
            }
            console.log("No more retries..");
            return null;
        });
    };

    return retrier(0);
};

exports.handler = (event, context, callback) => {
    if (!event.Records || event.Records.length === 0) {
        callback(null, {});
    } else {
        formatRecords(event.Records)
            .then(records => {
                if (records.length > 0) {
                    return sendRecordsToElasticSearch(records);
                }
                return null;
            })
            .then(() => callback(null, {}));
    }
};
