const fs = require('fs');
const bodybuilder = require('bodybuilder');
const _ = require('lodash');
const stringSimilarity = require('string-similarity');
const htmlencode = require('htmlencode');
const elasticSearch = require('elasticsearch');

const AWS = require('aws-sdk');
// eslint-disable-line import/no-extraneous-dependencies
AWS.config.update({ region: 'us-east-1' });
const s3 = new AWS.S3();
const elasticClient = new elasticSearch.Client({
    host: process.env.elasticSearchIp,
    apiVersion: '6.0'
});

const queryElasticWithPagination = (index, body, numResInResponse) => {
    console.log(`Querying elasticSearch`);
    let results = [];
    let page = 0;
    body.size = numResInResponse;

    const queryParams = {
        index,
        type: "audit",
        body
    };

    const paginate = () => {
        body.from = page * numResInResponse;
        return elasticClient.search(queryParams)
            .then(res => {
                console.log(`Got results ${page * numResInResponse + 1}-${(page + 1) * numResInResponse} from ${res.hits.total} hits`);
                results = _.concat(results, res.hits.hits);
                if (((page * numResInResponse) + numResInResponse) >= res.hits.total) {
                    return results;
                }
                page += 1;
                return paginate();
            });
    };
    return paginate();
};

const processRecords = records => {
    console.log(`Processing records`);
    _.forEach(records, record => { record._request = `${record._source.auditContext.method} : ${record._source.auditContext.pattern}`; });
    const recordsGroupedByApiCall = _.groupBy(records, "_request");
    _.forOwn(recordsGroupedByApiCall, (recordItems, apiCall) => {
        const messages = recordItems.map(record => ({
            text: record._source.data.body ? record._source.data.body.s_message : record._source.data.link,
            code: record._source.data.statusCode
        }));
        const uniqMessagesInfo = {};
        messages.forEach(message => {
            const messageKeys = Object.keys(uniqMessagesInfo);
            const matches = !_.isEmpty(messageKeys) ? stringSimilarity.findBestMatch(message.text, messageKeys) : { bestMatch: { rating: 0 } };
            if (matches.bestMatch.rating < 0.65) {
                uniqMessagesInfo[message.text] = {
                    code: message.code,
                    messages: [message]
                };
            } else {
                uniqMessagesInfo[messageKeys[matches.bestMatchIndex]].messages.push(message);
            }
        });
        recordsGroupedByApiCall[apiCall] = {
            records: recordItems,
            messages,
            uniqMessagesInfo
        };
    });
    return recordsGroupedByApiCall;
};

const generateSummaryTable = processedRecords => {
    console.log(`Generating summary tables html`);
    let html = ``;
    _.forOwn(processedRecords, (value, summaryTableHeader) => {
        html += `<div class="sum-grid-container">`;
        html += `<div class="sum-table-title">${summaryTableHeader}</div>`;
        html += `<div class="sum-table-header-item">Error Text</div>`;
        html += `<div class="sum-table-header-item">Error Code</div>`;
        html += `<div class="sum-table-header-item">Failure Records</div>`;
        const errorSummary = value.uniqMessagesInfo;
        // eslint-disable-next-line no-shadow
        _.forOwn(errorSummary, (value, text) => {
            const { code } = value;
            const numberOfFailures = value.messages.length;
            html += `<div class="data-item">${htmlencode.htmlEncode(text)}</div>`;
            html += `<div class="data-item">${code}</div>`;
            html += `<div class="data-item">${numberOfFailures}</div>`;
        });
        html += `</div>`;
    });
    return { html, processedRecords };
};

const generateDebugTables = ({ html: summaryTableHtml, processedRecords }) => {
    console.log(`Generating debug tables html`);
    let html = ``;
    _.forOwn(processedRecords, (value, summaryTableHeader) => {
        html += `<div class="debug-grid-container">`;
        html += `<div class="sum-table-title">${summaryTableHeader}</div>`;
        const errorSummary = value.uniqMessagesInfo;
        // eslint-disable-next-line no-shadow
        _.forOwn(errorSummary, (value, uniqueMessage) => {
            html += `<div class="debug-table-header">${htmlencode.htmlEncode(uniqueMessage)}.</div>`;
            const messages = _.uniqWith(value.messages, _.isEqual);
            messages.forEach(message => {
                html += `<div class="data-item">${htmlencode.htmlEncode(message.text)}</div>`;
                html += `<div class="data-item">${message.code}</div>`;
            });
        });
        html += `</div>`;
    });
    return summaryTableHtml.concat(html);
};

const generateHtml = data => new Promise((resolve, reject) => {
    console.log(`Generating html report`);
    fs.readFile(`${__dirname}/failedApiReportTemplate.html`, 'utf8', (err, fileText) => {
        if (err) {
            reject(err);
        } else {
            const html = fileText.replace(/<!--TABLE GENERATION GOES HERE-->/g, data);
            resolve(html);
        }
    });
});

const writeResultsToBucket = html => {
    console.log(`Creating object in S3 ${process.env.privateBucket}`);
    const bucketParams = {
        Bucket: process.env.privateBucket,
        Body: Buffer.from(html, 'binary'),
        Key: 'reports/api-failure/apiFailureReport.html'
    };
    return s3.putObject(bucketParams).promise();
};

const fetchFailedApi = (daysTimeRange = 7) => {
    console.log(`Fetching all failed APIs in the last ${daysTimeRange} days`);
    const body = bodybuilder()
        .query('range', 'date', { gte: `now-${daysTimeRange}d/d` })
        .query('match_phrase', 'data.status', `FAILED`)
        .build();
    const index = process.env === 'staging' ? `audit-${process.env.stage}*` : `audit-${process.env.stage}*`;
    return queryElasticWithPagination(index, body, 100)
        .then(processRecords)
        .then(generateSummaryTable)
        .then(generateDebugTables)
        .then(generateHtml)
        .then(writeResultsToBucket);
};

exports.handler = (event, context, callback) => {
    console.log("Attempting to generate API failure report");
    return fetchFailedApi(process.env.apiErrorReportTimeInterval)
        .then(() => callback(null, {}))
        .catch(err => {
            console.log("Error");
            console.log(err);
            return callback(null, {});
        });
};
