const _ = require("lodash");
const Joi = require('joi');
const axios = require('axios');
const AWS = require('aws-sdk');
const uuidv4 = require('uuid/v4');
const ASCIIFolder = require('fold-to-ascii');
const dispatch = require('./../dispatch');
const { ForbiddenError, ClientError } = require("./../errors");

const isUserPermittedOnAccount = ({ userAccounts, accountId, admin }) => {
    if (admin) {
        return !!_.find(userAccounts, { accountPublicId: accountId, userRole: "Role-1" });
    }
    return !!_.find(userAccounts, { accountPublicId: accountId });
};

const isUserPermittedOnSubscription = ({
    subscription, userAccounts, accountId, admin
}) => {
    if (!subscription) {
        return false;
    }
    if (accountId) {
        return isUserPermittedOnAccount({ userAccounts, accountId, admin }) && subscription.accounts.includes(accountId);
    }
    return _.some(subscription.accounts, accId => isUserPermittedOnAccount({ userAccounts, accId, admin }));
};

const accountsOfUser = async ({ userId, userToken }) => {
    console.log(`Fetching accounts of user ${userId}`);
    const { data: accounts } = await axios.get(`${process.env.tenancyApiUrl}/account`, {
        headers: {
            Authorization: `Bearer ${userToken}`,
            'X-User-Token': `Bearer ${userToken}`
        }
    });
    console.log(accounts);
    return accounts;
};

const createAccountForUser = async ({ accountName, userToken }) => {
    console.log(`Creating account with name ${accountName}`);
    const { data: accountResponse } = await axios.post(`${process.env.tenancyApiUrl}/account/${accountName}`, null, {
        headers: {
            Authorization: `Bearer ${userToken}`,
            'X-User-Token': `Bearer ${userToken}`
        }
    });
    console.log("Successfully created account");
    return accountResponse;
};

const mapSubscription = (subscription, userAccounts) => {
    let accounts = _.get(subscription, "accounts.values", []);
    if (userAccounts) {
        accounts = _.intersection(accounts, _.map(userAccounts, "accountPublicId"));
    }
    return { ...subscription, accounts };
};

const getSubscriptionById = async ({ subscriptionId }) => {
    console.log(`Fetching ${subscriptionId} from database`);
    const subscriptionParams = {
        TableName: process.env.subscriptionsTable,
        Key: { subscriptionId }
    };
    console.log("Successfully retrieved subscription from database");
    const documentClient = new AWS.DynamoDB.DocumentClient();

    const subscriptionResponse = await documentClient.get(subscriptionParams);
    if (!subscriptionResponse.Item) {
        return null;
    }
    return mapSubscription(subscriptionResponse.Item);
};


const getSubscriptions = async ({
    userId, accountId, serviceName, isUser, userToken
}) => {
    console.log(`Listing subscriptions for accountId: ${accountId} with user ${userId}`);
    let userAccounts = [];
    if (isUser) {
        userAccounts = await accountsOfUser({ userId, userToken });
    }
    const isPermitted = isUser ? isUserPermittedOnAccount({ userAccounts, accountId, admin: false }) : true;
    if (!isPermitted) {
        throw new ForbiddenError(`Account does not exist, or user is not permitted to get subscriptions for account: ${accountId}`);
    }

    const documentClient = new AWS.DynamoDB.DocumentClient();

    const accountParams = {
        TableName: process.env.accountsToSubscriptionsTable,
        Key: { accountId }
    };

    console.log(`Retrieving subscription ids from database`);
    const accountsMappingResponse = await documentClient.get(accountParams).promise();
    console.log(accountsMappingResponse);

    const subscriptionIds = _.get(accountsMappingResponse, "Item.subscriptions.values", []);
    if (subscriptionIds.length === 0) {
        console.log("Account has no subscriptions");
        return [];
    }

    if (subscriptionIds.length > 100) {
        console.log("Warning: account has more than 100 subscriptions. Response is partial.");
    }

    const clippedSubscriptionIds = _.take(subscriptionIds, 100);
    const batchGetParams = {
        RequestItems: {
            [process.env.subscriptionsTable]: {
                Keys: clippedSubscriptionIds.map(subscriptionId => ({ subscriptionId }))
            }
        }
    };

    console.log(`Fetching ${clippedSubscriptionIds.length} subscriptions`);
    const batchGetResponse = await documentClient.batchGet(batchGetParams).promise();
    const subscriptions = _(batchGetResponse).get(`Responses.${process.env.subscriptionsTable}`, []).map(subscription => mapSubscription(subscription, isUser ? userAccounts : undefined));
    if (serviceName) {
        console.log(`Filtering responses for service ${serviceName}`);
        return _.filter(subscriptions, { serviceName });
    }
    return subscriptions;
};

const addSubscriptionToAccountInternal = async ({ subscriptionId, accountId }) => {
    const documentClient = new AWS.DynamoDB.DocumentClient();

    const addParams = {
        TransactItems: [{
            Update: {
                TableName: process.env.subscriptionsTable,
                Key: { subscriptionId },
                UpdateExpression: "ADD accounts :a",
                ExpressionAttributeValues: {
                    ":a": documentClient.createSet([accountId])
                }
            }
        },
        {
            Update: {
                TableName: process.env.accountsToSubscriptionsTable,
                Key: { accountId },
                UpdateExpression: "ADD subscriptions :s",
                ExpressionAttributeValues: {
                    ":s": documentClient.createSet([subscriptionId])
                }
            }
        }]
    };
    console.log(`Transactionally adding subscription ${subscriptionId} to account ${accountId}`);
    await documentClient.transactWrite(addParams).promise();
    console.log("Transactional add successfully completed");
    return null;
};

const removeSubscriptionFromAccountInternal = async ({ subscriptionId, accountId }) => {
    const documentClient = new AWS.DynamoDB.DocumentClient();

    const deleteParams = {
        TransactItems: [{
            Update: {
                TableName: process.env.subscriptionsTable,
                Key: { subscriptionId },
                UpdateExpression: "DELETE accounts :a",
                ConditionExpression: "size(accounts) > :s",
                ExpressionAttributeValues: {
                    ":a": documentClient.createSet([accountId]),
                    ":s": 1
                }
            }
        },
        {
            Update: {
                TableName: process.env.accountsToSubscriptionsTable,
                Key: { accountId },
                UpdateExpression: "DELETE subscriptions :s",
                ExpressionAttributeValues: {
                    ":s": documentClient.createSet([subscriptionId])
                }
            }
        }]
    };

    console.log("Performing transactional remove from database");
    try {
        await documentClient.transactWrite(deleteParams).promise();
    } catch (e) {
        if (e.message.includes("ConditionalCheckFailed")) {
            throw new ClientError("Unable to remove subscription from account - subscription must exist in at least one account.");
        } else {
            throw e;
        }
    }
};

const addSubscriptionToAccount = async ({
    userId, accountId, subscriptionId, userToken
}) => {
    console.log(`Adding subscription ${subscriptionId} to account ${accountId} for user ${userId}`);
    const userAccounts = await accountsOfUser({ userId, userToken });
    const subscription = await getSubscriptionById({ subscriptionId });
    const isUserPermitted = isUserPermittedOnSubscription({
        userAccounts, subscription, accountId, admin: true
    });
    if (!isUserPermitted) {
        throw new ForbiddenError('User is not permitted to add subscription to account');
    }

    await addSubscriptionToAccountInternal({ subscriptionId, accountId });
    return null;
};

const getSubscriptionsByService = async ({ serviceName }) => {
    console.log(`Fetching subscriptions for ${serviceName}`);
    const getParams = {
        TableName: process.env.subscriptionsTable,
        IndexName: 'serviceNameIndex',
        KeyConditionExpression: 'serviceName = :sn',
        ExpressionAttributeValues: {
            ':sn': serviceName
        }
    };

    const documentClient = new AWS.DynamoDB.DocumentClient();
    console.log("Performing fetch from database");
    const getResponse = await documentClient.query(getParams).promise();
    return (getResponse.Items || []).map(mapSubscription);
};

const removeSubscriptionFromAccount = async ({
    userId, accountId, subscriptionId, userToken
}) => {
    console.log(`Removing subscription ${subscriptionId} from account ${accountId} for user ${userId}`);
    const userAccounts = await accountsOfUser({ userId, userToken });
    const subscription = await getSubscriptionById({ subscriptionId });
    const isUserPermitted = isUserPermittedOnSubscription({
        subscription, userAccounts, accountId, admin: true
    });
    if (!isUserPermitted) {
        throw new ForbiddenError('User is not permitted to remove subscription from account');
    }

    await removeSubscriptionFromAccountInternal({ subscriptionId, accountId });
    return null;
};

const resolveAws = async ({
    userId, registrationToken, userToken, suggestedName
}) => {
    console.log(`Attempting token resolution and subscription update - AWS - ${registrationToken} for user: ${userId}`);
    const userAccounts = await accountsOfUser({ userId, userToken });
    const adminAccounts = _.filter(userAccounts, { userRole: 'Role-1' });
    console.log(`User is admin in the following accounts: ${adminAccounts.join(', ')}`);

    let newAccountName = null;
    let newAccountCreated = false;

    if (adminAccounts.length === 0) {
        const uuid = uuidv4().split("-")[0];
        newAccountName = `${ASCIIFolder.fold(suggestedName)}-${uuid}`;
        newAccountCreated = true;
        console.log(`User is admin in no accounts, creating account for user with name: ${newAccountName}`);
        try {
            const account = await createAccountForUser({ accountName: newAccountName, userToken });
            adminAccounts.push(account);
        } catch (e) {
            console.log(`Failed to create new account for user, subscription resolution failed`);
            console.log(e);
            throw new Error('Unable to create new account to assign subscription.');
        }
    }

    const marketplacemetering = new AWS.MarketplaceMetering({
        accessKeyId: process.env.sellerAccessKey,
        secretAccessKey: process.env.sellerSecretKey
    });

    console.log("Resolving token");
    const resolvedCustomer = await marketplacemetering.resolveCustomer({ RegistrationToken: registrationToken }).promise();

    const { CustomerIdentifier: customerIdentifier, ProductCode: productCode } = resolvedCustomer;
    console.log(`Resolution successful- customerIdentifier: ${customerIdentifier}, productCode: ${productCode}`);

    console.log("Identifying product - database fetch");
    const documentClient = new AWS.DynamoDB.DocumentClient();
    const productParams = {
        TableName: process.env.productsTable,
        Key: { productCode }
    };

    const productResponse = await documentClient.get(productParams).promise();
    if (!productResponse.Item) {
        throw new ClientError(`Invalid product code: ${productCode}`);
    }

    const product = productResponse.Item;
    console.log("Product resolved:");
    console.log(product);


    const uuid = uuidv4().split("-")[0];
    const subscriptionName = `${suggestedName}-${product.serviceName}-${uuid}`;
    console.log(`If new subscription, name will be ${subscriptionName}`);

    const subscriptionId = `aws-${productCode}-${customerIdentifier}`;

    const subscriptionParams = {
        TableName: process.env.subscriptionsTable,
        Key: { subscriptionId },
        ReturnValues: "ALL_OLD",
        UpdateExpression: 'SET cloudProvider = :cp, subscriptionStatus = if_not_exists(subscriptionStatus, :s), serviceName = :sn, customerIdentifier = :ci, productCode = :pc, subscriptionName = if_not_exists(subscriptionName, :subn), createdAt = if_not_exists(createdAt, :ca)',
        ExpressionAttributeValues: {
            ':cp': "aws",
            ':s': "pending",
            ':sn': product.serviceName,
            ':ci': customerIdentifier,
            ':pc': productCode,
            ':subn': subscriptionName,
            ':ca': new Date().valueOf()
        }
    };

    console.log("Updating subscription in database");
    const updateResponse = await documentClient.update(subscriptionParams).promise();
    console.log("Subscription updated: old values");
    console.log(updateResponse);

    const newSubscription = !updateResponse.Attributes;
    const oldSubscription = mapSubscription(updateResponse.Attributes);

    let addedAccountsResults = [];

    const previousAccounts = _.get(oldSubscription, "accounts", []);
    const adminAccountPublicIds = _.map(adminAccounts, "accountPublicId");
    console.log(`User is admin on the following accounts: [${adminAccountPublicIds.join(',')}]`);
    console.log(`Subscription is already on the accounts: [${previousAccounts.join(', ')}]`);
    const accountsToAdd = _.difference(adminAccountPublicIds, previousAccounts);
    console.log(`Adding the following accounts to subscription: [${accountsToAdd.join(', ')}]`);
    addedAccountsResults = await Promise.all(accountsToAdd.map(accountId => addSubscriptionToAccountInternal({ subscriptionId, accountId }).then(() => ({ accountId, added: true }), e => {
        console.log(`Warning!! Failed adding account ${accountId}`);
        console.log(e);
        return { accountId, added: false };
    })));

    const failedToAddAccounts = addedAccountsResults.filter(result => result.added === false).map(result => result.accountId);

    return {
        subscriptionName: oldSubscription.subscriptionName || subscriptionName, newSubscription, newAccountCreated, newAccountName, accounts: _.differenceWith(adminAccounts, failedToAddAccounts, (adminAccount, failedId) => adminAccount.accountPublicId === failedId)
    };
};

/*
    Subscription name is the same across all accounts - but access to the subscription is still specified via a specific account
 */
const renameSubscription = async ({
    userId, subscriptionId, newSubscriptionName, userToken, accountId
}) => {
    console.log(`Renaming subscription ${subscriptionId} to ${newSubscriptionName} on behalf of user ${userId} in account ${accountId}`);
    const userAccounts = await accountsOfUser({ userId, userToken });
    const subscription = await getSubscriptionById({ subscriptionId });
    const isUserPermitted = isUserPermittedOnSubscription({
        userAccounts, subscription, admin: true, accountId
    });
    if (!isUserPermitted) {
        throw new ForbiddenError('User is not allowed to rename subscription');
    }

    const documentClient = new AWS.DynamoDB.DocumentClient();

    const updateParams = {
        TableName: process.env.subscriptionsTable,
        Key: { subscriptionId },
        UpdateExpression: "SET subscriptionName = :sn",
        ExpressionAttributeValues: {
            ":sn": newSubscriptionName
        }
    };

    console.log(`Updating database entry for subscription: ${subscriptionId}`);
    await documentClient.update(updateParams).promise();
    console.log("Successfully updated database entry");
    return Promise.resolve(null);
};

const routes = [
    {
        method: "GET",
        pattern: "/subscriptions/by-service-name",
        querySchema: {
            serviceName: Joi.string().required()
        },
        scopes: ["ss:internal"],
        handler: ({ queryParams }) => getSubscriptionsByService({ serviceName: queryParams.serviceName })
    },
    {
        method: "GET",
        pattern: "/subscriptions/:accountId",
        scopes: [],
        handler: ({
            principalId, urlParams, queryParams, scopes, token
        }) => {
            const isUser = !scopes.includes("ss:internal");
            const userId = isUser ? principalId : null;
            return getSubscriptions({
                userId, accountId: urlParams.accountId, isUser, serviceName: _.get(queryParams, "serviceName", null), userToken: token
            });
        }
    },
    {
        method: "DELETE",
        pattern: "/subscriptions/:accountId/:subscriptionId",
        scopes: [],
        handler: (principalId, urlParams, token) => removeSubscriptionFromAccount({
            userId: principalId, accountId: urlParams.accountId, subscriptionId: urlParams.subscriptionId, userToken: token
        })
    },
    {
        method: "POST",
        pattern: "/subscriptions/resolve-aws",
        bodySchema: Joi.object().keys({
            registrationToken: Joi.string().required(),
            suggestedName: Joi.string().required()
        }),
        handler: ({ principalId, body, token }) => resolveAws({
            userId: principalId, registrationToken: body.registrationToken, userToken: token, suggestedName: body.suggestedName
        })
    },
    {
        method: "POST",
        pattern: "/subscriptions/:accountId",
        bodySchema: Joi.object().keys({
            subscriptionId: Joi.string().required()
        }),
        scopes: [],
        handler: ({
            principalId, urlParams, body: { subscriptionId }, token
        }) => addSubscriptionToAccount({
            userId: principalId, accountId: urlParams.accountId, subscriptionId, userToken: token
        })
    },
    {
        method: "PATCH",
        pattern: "/subscriptions/:accountId/:subscriptionId",
        bodySchema: Joi.object().keys({
            subscriptionName: Joi.string().min(2).max(250).required()
        }),
        handler: ({
            principalId, body, urlParams, token
        }) => renameSubscription({
            userId: principalId, subscriptionId: urlParams.subscriptionId, accountId: urlParams.accountId, newSubscriptionName: body.subscriptionName, userToken: token
        })
    }
];

exports.handler = (event, context, callback) => {
    dispatch({
        event,
        callback,
        context,
        routes
    });
};
