const express = require('express');
const _ = require("lodash");
const Promise = require("bluebird");
const { reportMetering, resolveSubscription, saveHourlyCharge, totalPerSubscription, updateHourlyChargeStatus } = require("../services/hourlyCharge");
const database = require("../services/database");
const metering = require("../services/metering");
const { validateHourlyChargeRequest } = require("../services/validator");
const { processTimestamp } = require("../utils");

const router = express.Router();

const validateRequestBody = (req, res, next) => {
    console.log(`Validating request body: ${JSON.stringify(req.body)}`);
    validateHourlyChargeRequest(req.body)
        .then(() => {
            console.log(`Successfully validated request body`);
            next();
        })
        .catch(err => {
            next(`request body validation error: ${err}`);
        });
};

const addMergedId = (req, res, next) => {
    console.log(`Adding mergedId to charges`);
    _.forEach(req.body, charge => { charge.mergedId = `${charge.subscriptionId}-${charge.dimension}`; });
    next();
};

const validateUniqueness = (req, res, next) => {
    console.log(req.body);
    console.log(`Validating charges uniqueness by mergedId and time`);
    const comparator = (chargeVal, otherChargeVal) => ((chargeVal.mergedId === otherChargeVal.mergedId) && (processTimestamp(chargeVal.unixTimestamp, "hourly") === processTimestamp(otherChargeVal.unixTimestamp, "hourly")));
    const unique = _.uniqWith(req.body, comparator);
    if (unique.length === req.body.length) {
        console.log(`Successfully validated charges uniqueness`);
        next();
    } else {
        next(`Metering failed! Duplicate charges identified in your request.`);
    }
};

const calculateTotals = (req, res, next) => {
    console.log(req.body);
    console.log(`Validating charges uniqueness by mergedId and time`);
    _.forEach(req.body, charge => {
        const accountsTotals = totalPerSubscription(charge.accounts);
        console.log(`Totals: ${JSON.stringify(accountsTotals)}`);
        charge.accumulatedVal = accountsTotals.accumulatedVal;
        charge.accounts = accountsTotals.accounts;
    });
    next();
};

router.post("/", validateRequestBody);
router.post("/", addMergedId);
router.post("/", validateUniqueness);
router.post("/", calculateTotals);

const meteringStubFunc = enhancedHourlyCharges => {
    const c = _.map(enhancedHourlyCharges, charge => ({ ...charge, chargingStatus: _.sample([true, false]) }));
    return Promise.resolve(c);
};

// eslint-disable-next-line no-unused-vars
router.post('/', (req, res, next) => {
    Promise.map(req.body, async chargePerSubscription => {
        console.log('***************************');
        console.log(`Processing subscription charge ${JSON.stringify(chargePerSubscription)}`);
        let enhancedChargeObj = { ...chargePerSubscription };
        const resolution = await resolveSubscription(database.getSubscription)(enhancedChargeObj.subscriptionId);
        // TODO: extract resolution logic to function
        if ((resolution.status === "fulfilled" && _.isEmpty(resolution.value)) || resolution.status === "rejected") {
            console.log(`Critical error!!! Could not resolve subscription id`);
            enhancedChargeObj.error = _.isEmpty(resolution.value) ? "subscription-notFound" : resolution.error;
            return enhancedChargeObj;
        }
        if ((resolution.value.subscriptionStatus !== "subscribed") && (resolution.value.subscriptionStatus !== "unsubscribed-pending")) {
            console.log(`Critical error!!! Subscription status error`);
            enhancedChargeObj.error = `subscription- status error: status ${resolution.value.subscriptionStatus}`;
            return enhancedChargeObj;
        }
        enhancedChargeObj = { ...enhancedChargeObj, ..._.pick(resolution.value, ['cloudProvider', 'customerIdentifier', 'productCode', 'serviceName', 'subscriptionStatus']) };
        const saveToHourlyTableRes = await saveHourlyCharge(database.saveHourlyCharge)(enhancedChargeObj);
        if (saveToHourlyTableRes.status === "rejected") {
            console.log(`Critical error!!! Could not save charge to hourly table`);
            enhancedChargeObj.error = saveToHourlyTableRes.error;
            return enhancedChargeObj;
        }
        return enhancedChargeObj;
    }, { concurrency: 1 })
        .then(enhancedCharges => {
            console.log('------------------------------------------');
            const validEnhancedCharges = _.reject(enhancedCharges, 'error');
            console.log(`Reporting metering on ${validEnhancedCharges.length} charges: ${JSON.stringify(validEnhancedCharges)}`);
            const invalidEnhancedCharges = _.filter(enhancedCharges, 'error');
            console.log(`No metering done on ${invalidEnhancedCharges.length} charges: ${JSON.stringify(invalidEnhancedCharges)}`);
            if (process.env.doCharge) {
                return reportMetering(metering.reportMetering)(validEnhancedCharges);
            }
            return meteringStubFunc(validEnhancedCharges);
        })
        .then(enhancedChargesWithStatus => Promise.map(enhancedChargesWithStatus, async chargeWithStatus => {
            console.log('***************************');
            console.log(`Updating hourly charge status of: ${JSON.stringify(chargeWithStatus)}`);
            const billingUpdateResponse = await updateHourlyChargeStatus(database.updateHourlyCharge)(chargeWithStatus);
            if (billingUpdateResponse.status === "rejected") {
                console.log(`Critical error!!! Could not update charge to hourly table`);
                enhancedChargesWithStatus.error = billingUpdateResponse.error;
                return enhancedChargesWithStatus;
            }
            return enhancedChargesWithStatus;
        }, { concurrency: 1 }));
    res.end("Hello!");
});

// eslint-disable-next-line no-unused-vars
router.post("/health", (req, res, next) => {
    database.getSubscription("fake-subscription-id")
        .then(() => {
            res.status = 203;
            res.send(`Server ready`);
        })
        .catch(err => {
            res.status = 500;
            res.send(`Health check failed with error ${err}`);
        });
});

module.exports = router;
