const AWS = require('aws-sdk/index');
const Promise = require("bluebird");
const _ = require("lodash");
const { processTimestamp, fromUnixToIsoString } = require("../utils");

AWS.config.update({ region: 'us-east-1' });
AWS.config.setPromisesDependency(require('bluebird'));

const dynamoClient = new AWS.DynamoDB.DocumentClient();

const doWithRetry = (retries, asyncFunc) => asyncFunc()
    .catch(error => {
        const { code, statusCode } = error;
        const retry = (
            code === "ItemCollectionSizeLimitExceededException"
            || code === "LimitExceededException"
            || code === "ProvisionedThroughputExceededException"
            || code === "RequestLimitExceeded"
            || code === "ThrottlingException"
            || code === "UnrecognizedClientException"
            || statusCode == 500 // eslint-disable-line eqeqeq
            || statusCode == 503 // eslint-disable-line eqeqeq
        ) && (retries < process.env.retries);
        console.log(`Error: ${error} \nNumber or retries: ${retries} \nNeed to retry? ${retry}`);
        if (!retry) {
            return Promise.reject(error);
        }
        return Promise.delay(process.env.intervalMillisec * (retries + 1))
            .then(() => doWithRetry(retries + 1, asyncFunc));
    });

const getSubscription = subscriptionId => {
    console.log(`Attempting to resolve subscription ${subscriptionId}`);
    const params = {
        TableName: process.env.subscriptionsTable,
        Key: { subscriptionId }
    };
    return doWithRetry(0, () => dynamoClient.get(params).promise()
        .then(response => {
            console.log(`SubscriptionId response: ${JSON.stringify(response.Item || {})}`);
            return (response.Item || {});
        }));
};

const getAccumulatedCharge = ({ mergedId, timestamp }) => {
    console.log(`Attempting to get accumulated charge ${{ mergedId, timestamp }}`);
    const params = {
        TableName: process.env.accumulatedChargingTable,
        Key: { mergedId, timestamp: timestamp.toString() }
    };
    return doWithRetry(0, () => dynamoClient.get(params).promise()
        .then(response => {
            console.log(`Retrieved accumulated charge successfully: ${JSON.stringify(response.Item || {})}`);
            return response.Item;
        }));
};

const saveHourlyCharge = ({ mergedId, subscriptionId, customerIdentifier, dimension, timestamp, accounts, cloudProvider, productCode, serviceName, accumulatedVal, error }) => {
    console.log(`Attempting to save charge ${JSON.stringify({ mergedId, subscriptionId, customerIdentifier, dimension, timestamp, accounts, cloudProvider, productCode, serviceName, accumulatedVal, error })}`);
    const timestampByDateAndHour = processTimestamp(timestamp, "hourly");
    const params = {
        TableName: process.env.billingTable,
        Item: {
            mergedId,
            subscriptionId,
            timestamp: timestampByDateAndHour,
            isoStringTimestamp: fromUnixToIsoString(timestampByDateAndHour),
            customerIdentifier,
            dimension,
            accounts,
            cloudProvider,
            productCode,
            serviceName,
            accumulatedVal,
            error
        },
        ExpressionAttributeNames: { "#timestamp": "timestamp" },
        ConditionExpression: "attribute_not_exists(subscriptionId) AND attribute_not_exists(#timestamp)"
    };
    return doWithRetry(0, () => dynamoClient.put(params).promise()
        .then(response => {
            console.log(`Charge saved successfully`);
            return response;
        }));
};

const saveAccumulatedCharge = ({ mergedId, accountId, subscriptionId, customerIdentifier, dimension, timestamp, cloudProvider, productCode, serviceName, totalVal, resources }) => {
    console.log(`Attempting to save accumulated charge ${JSON.stringify({ mergedId, accountId, subscriptionId, dimension, timestamp, cloudProvider, productCode, serviceName, totalVal, resources })}`);
    const timestampByDate = processTimestamp(timestamp, "daily");
    const params = {
        TableName: process.env.accumulatedBillingTable,
        Item: {
            mergedId,
            timestamp: timestampByDate,
            isoStringTimestamp: fromUnixToIsoString(timestampByDate),
            subscriptionId,
            customerIdentifier,
            dimension,
            accountId,
            cloudProvider,
            productCode,
            serviceName,
            totalVal,
            resources
        }
    };
    return doWithRetry(0, () => dynamoClient.put(params).promise()
        .then(response => {
            console.log(`Accumulated charge saved successfully`);
            return response;
        }));
};

const updateHourlyCharge = ({ mergedId, timestamp, chargingStatus }) => {
    console.log(`Attempting to update hourly charge ${JSON.stringify({ mergedId, timestamp, chargingStatus })}`);
    const timestampByDateAndHour = processTimestamp(timestamp, "hourly");
    const params = {
        TableName: process.env.billingTable,
        Key: { mergedId, timestamp: timestampByDateAndHour },
        UpdateExpression: "set chargingStatus = :status",
        ExpressionAttributeValues: {
            ":status": chargingStatus
        },
        ReturnValues: "ALL_OLD"
    };
    return doWithRetry(0, () => dynamoClient.update(params).promise()
        .then(response => {
            if (_.isEmpty(response)) {
                throw new Error("Failed to update hourly charge status - item not found!");
            }
            console.log(`Updated hourly charge successfully: ${JSON.stringify(response)}`);
            return response;
        }));
};

const updateAccumulatedCharge = ({ mergedId, timestamp, resources, totalVal }) => {
    console.log(`Attempting to update accumulated charge ${JSON.stringify({ mergedId, timestamp, resources, totalVal })}`);
    const timestampByDate = processTimestamp(timestamp, "daily");
    const params = {
        TableName: process.env.accumulatedBillingTable,
        Key: { mergedId, timestamp: timestampByDate },
        UpdateExpression: "set resources = :resources, totalVal = :totalVal",
        ExpressionAttributeValues: {
            ":resources": resources,
            ":totalVal": totalVal
        },
        ReturnValues: "ALL_OLD"
    };
    return doWithRetry(0, () => dynamoClient.update(params).promise()
        .then(response => {
            if (_.isEmpty(response)) {
                throw new Error("Failed to update accumulated charge - item not found!");
            }
            console.log(`Updated accumulated charge successfully: ${JSON.stringify(response)}`);
            return response;
        }));
};

exports.doWithRetry = doWithRetry;
exports.getSubscription = getSubscription;
exports.saveHourlyCharge = saveHourlyCharge;
exports.saveAccumulatedCharge = saveAccumulatedCharge;
exports.getAccumulatedCharge = getAccumulatedCharge;
exports.updateAccumulatedCharge = updateAccumulatedCharge;
exports.updateHourlyCharge = updateHourlyCharge;
