const chai = require("chai");
const expect = require('chai').expect;
const Constants = require('../utils/Constants');
const Initializer = require('../utils/specInitializer');
const Promise = require('bluebird');
const jwtDecoder = require('../utils/general-utils').decodeToken;
const _ = require('lodash');
const chaiAsPromised = require("chai-as-promised");
chai.use(chaiAsPromised);

const { logger, clients } = Initializer('OCCM Integration Spec');

const publicIp = '1.1.1.1';
const privateIp = '2.2.2.2';
const systemId = "test-system-id";
const occmName = "test-name";
const redirectURIs = [
    `http://${publicIp}`,
    `http://${publicIp}/occmui`,
    `https://${publicIp}`,
    `https://${publicIp}/occmui`,
    `http://${privateIp}`,
    `http://${privateIp}/occmui`,
    `https://${privateIp}`,
    `https://${privateIp}/occmui`
];

const resetAppMetadataAttributes = async () => {
   await clients.auth0Client.updateUserDetails(Constants.testimioUserAuth0Sub, {"app_metadata": {}});
   await clients.auth0Client.updateUserDetails(Constants.defaultUserAuth0Sub, {"app_metadata": {}});
};

const resetUsersDynamoSubordinateUsers = async () => {
    await clients.dynamodbClient.updateSubordinateUsers(Constants.testimioUserAuth0Sub, []);
    await clients.dynamodbClient.updateSubordinateUsers(Constants.defaultUserAuth0Sub, []);
};

const deleteUserOccm = async (email, password) => {
    const govUserToken = await clients.auth0Client.getGovernanceUserToken(email, password);
    await clients.governanceClient.deleteAssociatedOccms(govUserToken);
};

const waitForDynamoToUpdate = async (millisec) => {
    await Promise.delay(millisec);
};

describe('Governance OCCM API Spec - deleteAssociatedOccms', function() {
    before(async function () {
        logger.debug('Inside after');
        await deleteUserOccm(Constants.defaultTestUserEmail, Constants.defaultTestUserPassword);
        await deleteUserOccm(Constants.testimioUserEmail, Constants.testimioUserPassword);
    });

    it('Should allow deleting OCCMs associated with user', async function() {
        const userToken = await clients.auth0Client.getCloudCentralUserToken(Constants.defaultTestUserEmail, Constants.defaultTestUserPassword);
        const decodedToken = jwtDecoder(userToken);

        let occmList = await clients.occmClient.listUserOccms(userToken);
        logger.debug(`User's OCCMs from cloud central API: ${JSON.stringify(occmList)}`);
        // noinspection BadExpressionStatementJS
        expect(occmList.occms).to.be.empty;
        let userInfoDynamo = await clients.dynamodbClient.getUserInfo(decodedToken.sub);
        logger.debug(`User info from dynamo: ${JSON.stringify(userInfoDynamo)}`);
        expect(userInfoDynamo.Item.associatedOccms).to.equal(undefined);

        const { clientId, clientSecret } = await clients.occmClient.serviceRegisterOccm(systemId, occmName, userToken);
        logger.debug(`New OCCM registered in auth0 with clientId: ${clientId} and clientSecrete: ${clientSecret}`);
        await waitForDynamoToUpdate(500);
        logger.debug(`New OCCM registered in auth0 with clientId: ${clientId} and clientSecrete: ${clientSecret}`);
        const occmToken = await clients.auth0Client.getOccmToken(clientId, clientSecret);
        await clients.occmClient.updateRedirectUris(redirectURIs, occmToken);

        occmList = await clients.occmClient.listUserOccms(userToken);
        logger.debug(`User's OCCMs from cloud central API: ${JSON.stringify(occmList)}`);
        expect(occmList.occms[0].occm).to.equal(clientId);
        userInfoDynamo = await clients.dynamodbClient.getUserInfo(decodedToken.sub);
        logger.debug(`User info from dynamo: ${JSON.stringify(userInfoDynamo)}`);
        expect(userInfoDynamo.Item.associatedOccms.values[0]).to.equal(clientId);
        let auth0Response = await clients.auth0Client.getClientById(clientId);
        logger.debug(`OCCM application info in Auth0: ${JSON.stringify(auth0Response)}`);
        // noinspection BadExpressionStatementJS
        expect(auth0Response).to.exist;

        const govToken = await clients.auth0Client.getGovernanceUserToken(Constants.defaultTestUserEmail, Constants.defaultTestUserPassword);
        await clients.governanceClient.deleteAssociatedOccms(govToken);

        occmList = await clients.occmClient.listUserOccms(userToken);
        logger.debug(`User's OCCMs from cloud central API: ${JSON.stringify(occmList)}`);
        // noinspection BadExpressionStatementJS
        expect(occmList.occms).to.be.empty;
        userInfoDynamo = await clients.dynamodbClient.getUserInfo(decodedToken.sub);
        logger.debug(`User info from dynamo: ${JSON.stringify(userInfoDynamo)}`);
        expect(userInfoDynamo.Item.associatedOccms).to.equal(undefined);
        const promise = clients.auth0Client.getClientById(clientId);
        return expect(promise)
            .to.eventually.be.rejected
            .and.be.an.instanceOf(Error)
            .then(error => {
                expect(error.response.data.message).to.equal('The client does not exist');
                expect(error.response.status).to.equal(404);
            })
    });
});

describe('Governance API Spec - connections, enableConnection and disableConnection', function() {
    const enableSupportScope = async (auth0UserId) => {
        await clients.auth0Client.updateUserDetails(auth0UserId, {"app_metadata": {"cc:support": true}});
    };

    beforeEach(async function () {
        logger.debug('Inside before');
        await resetAppMetadataAttributes();
    });

    it('Should not allow user to retrieve federation connections list if doesn\'t have cc:support scope', async function() {
        const govToken = await clients.auth0Client.getGovernanceUserToken(Constants.defaultTestUserEmail, Constants.defaultTestUserPassword);
        const promise = clients.governanceClient.getAllFederationConnections(govToken);
        return expect(promise)
            .to.eventually.be.rejected
            .and.be.an.instanceOf(Error)
            .then(async error => {
                expect(error.response.data.message).to.equal('Forbidden');
                expect(error.response.status).to.equal(403);
            })
    });

    it('Should allow user to retrieve all federation connections list if does have cc:support scope', async function() {
        await enableSupportScope(Constants.defaultUserAuth0Sub);
        const govToken = await clients.auth0Client.getGovernanceUserToken(Constants.defaultTestUserEmail, Constants.defaultTestUserPassword);
        const connections = await clients.governanceClient.getAllFederationConnections(govToken, "name");
        const existQaFederationConnection = _.find(connections, ['name', Constants.federationConnectionName]);
        logger.info(`target connection: ${JSON.stringify(existQaFederationConnection)}`);
        logger.debug(`Connections: ${JSON.stringify(connections)}`);
        // noinspection BadExpressionStatementJS
        expect(existQaFederationConnection).to.not.be.empty;
    });

    it('Should not allow user to enable federation connection if doesn\'t have cc:support scope', async function() {
        const govToken = await clients.auth0Client.getGovernanceUserToken(Constants.defaultTestUserEmail, Constants.defaultTestUserPassword);
        const promise = clients.governanceClient.enableFederationConnection(govToken, Constants.federationConnectionName, Constants.federationConnectionCompanyName, Constants.federationConnectionEmailDomains);
        return expect(promise)
            .to.eventually.be.rejected
            .and.be.an.instanceOf(Error)
            .then(async error => {
                expect(error.response.data.message).to.equal('Forbidden');
                expect(error.response.status).to.equal(403);
            })
    });

    it('Should not allow user to disable federation connection if doesn\'t have cc:support scope', async function() {
        const govToken = await clients.auth0Client.getGovernanceUserToken(Constants.defaultTestUserEmail, Constants.defaultTestUserPassword);
        const promise = clients.governanceClient.disableFederationConnection(govToken, Constants.federationConnectionName);
        return expect(promise)
            .to.eventually.be.rejected
            .and.be.an.instanceOf(Error)
            .then(async error => {
                expect(error.response.data.message).to.equal('Forbidden');
                expect(error.response.status).to.equal(403);
            })
    });

    it('Should allow user to enable/disable federation connection if does have cc:support scope and enable/disable all general clients on the connection', async function() {
        await enableSupportScope(Constants.defaultUserAuth0Sub);
        const govToken = await clients.auth0Client.getGovernanceUserToken(Constants.defaultTestUserEmail, Constants.defaultTestUserPassword);

        const enableResponse = await clients.governanceClient.enableFederationConnection(govToken, Constants.federationConnectionName, Constants.federationConnectionCompanyName, Constants.federationConnectionEmailDomains);
        logger.info(`Enable connection response: ${JSON.stringify(enableResponse)}`);
        let staticClients = await clients.dynamodbClient.getStaticClientsWithDisplayName();
        staticClients = staticClients.filter(client => !_.includes(Constants.netAppInternalClients, client.clientName));
        logger.debug(`Static clients from dynamo: ${JSON.stringify(staticClients)}`);
        expect(enableResponse.generalClientsCount).to.equal(staticClients.length);

        let connections = await clients.governanceClient.getAllFederationConnections(govToken, "enabled_clients,name");
        let targetFederationConnection = _.find(connections, ['name', Constants.federationConnectionName]);
        logger.info(`Target connection: ${JSON.stringify(targetFederationConnection)}`);
        expect(targetFederationConnection.enabled_clients.length).to.equal(staticClients.length);
        expect(_.indexOf(targetFederationConnection.enabled_clients, Constants.integrationClientId)).to.not.equal(-1);

        const disableResponse = await clients.governanceClient.disableFederationConnection(govToken, Constants.federationConnectionName);
        logger.info(`Disable connection response: ${JSON.stringify(disableResponse)}`);
        expect(disableResponse.message).to.equal(`Disabled connection \"${Constants.federationConnectionName}\" successfully.`);

        connections = await clients.governanceClient.getAllFederationConnections(govToken, "enabled_clients,name");
        targetFederationConnection = _.find(connections, ['name', Constants.federationConnectionName]);
        expect(targetFederationConnection.enabled_clients.length).to.equal(0);
    });

    it('Should apply OCCMs associated with federated users', async function() {
        //TODO: implement using refresh token
    });
});

describe('Governance API Spec - resetPassword', function() {
    const enableCourseAdminScope = async (auth0UserId) => {
        await clients.auth0Client.updateUserDetails(auth0UserId, {"app_metadata": {"cc:course-admin": true}});
    };

    beforeEach(async function () {
        logger.debug('Inside before');
        await resetUsersDynamoSubordinateUsers();
        await resetAppMetadataAttributes();
    });

    it('Should not allow user to change another user\'s password if is not cc:course-admins authorized in auth0, even if authorized to do so in dynamo', async function() {
        await clients.dynamodbClient.updateSubordinateUsers(Constants.testimioUserAuth0Sub, [Constants.defaultUserAuth0Sub]);
        const govToken = await clients.auth0Client.getGovernanceUserToken(Constants.testimioUserEmail, Constants.testimioUserPassword);
        const promise = clients.governanceClient.resetPassword(govToken, Constants.defaultTestUserEmail, "StamPassword@");
        return expect(promise)
            .to.eventually.be.rejected
            .and.be.an.instanceOf(Error)
            .then(async error => {
                expect(error.response.data.message).to.equal('Forbidden');
                expect(error.response.status).to.equal(403);
            })
    });

    it('Should not allow resetting user passwords that supervising user is not allowed to edit in dynamo', async function() {
        await enableCourseAdminScope(Constants.defaultUserAuth0Sub);
        let govToken = await clients.auth0Client.getGovernanceUserToken(Constants.defaultTestUserEmail, Constants.defaultTestUserPassword);
        let promise = clients.governanceClient.resetPassword(govToken, Constants.testimioUserEmail, "stamPassword@");
        return expect(promise)
            .to.eventually.be.rejected
            .and.be.an.instanceOf(Error)
            .then(async error => {
                expect(error.response.data.message).to.equal('Forbidden');
                expect(error.response.status).to.equal(403);
            })
    });

    it('Should allow resetting user passwords that supervising user is authorized to edit in dynamo', async function() {
        await enableCourseAdminScope(Constants.defaultUserAuth0Sub);
        let govToken = await clients.auth0Client.getGovernanceUserToken(Constants.defaultTestUserEmail, Constants.defaultTestUserPassword);
        await clients.dynamodbClient.updateSubordinateUsers(Constants.defaultUserAuth0Sub, [Constants.courseStudetAuth0Sub]);
        await waitForDynamoToUpdate(500);
        let res = await clients.governanceClient.resetPassword(govToken, Constants.courseStudetEmail, "stamPassword21@");
        expect(res.message).to.equal(`Successfully updated password for user with E-mail ${Constants.courseStudetEmail}`);
        const getGovTokenWithOldPassword = clients.auth0Client.getGovernanceUserToken(Constants.courseStudetEmail, Constants.courseStudetPassword);
        return expect(getGovTokenWithOldPassword)
            .to.eventually.be.rejected
            .and.be.an.instanceOf(Error)
            .then(async error => {
                expect(error.response.data.error_description).to.equal('Wrong email or password.');
                expect(error.response.status).to.equal(403);
                // reset back testimio old password and verify change
                await clients.governanceClient.resetPassword(govToken, Constants.courseStudetEmail, Constants.courseStudetPassword);
                govToken = await clients.auth0Client.getGovernanceUserToken(Constants.courseStudetEmail, Constants.courseStudetPassword);
                expect(jwtDecoder(govToken).sub).to.equal(Constants.courseStudetAuth0Sub);
            })
    });

    it('Should return error message if new password is not within required constraints', async function() {
        await enableCourseAdminScope(Constants.defaultUserAuth0Sub);
        let govToken = await clients.auth0Client.getGovernanceUserToken(Constants.defaultTestUserEmail, Constants.defaultTestUserPassword);
        await clients.dynamodbClient.updateSubordinateUsers(Constants.defaultUserAuth0Sub, [Constants.courseStudetAuth0Sub]);
        let promise = clients.governanceClient.resetPassword(govToken, Constants.courseStudetEmail, "blabla");
        return expect(promise)
            .to.eventually.be.rejected
            .and.be.an.instanceOf(Error)
            .then(async error => {
                expect(error.response.data.message).to.equal('PasswordStrengthError: Password is too weak');
                expect(error.response.status).to.equal(400);
            })
    });

    it('Should return error message if the user whom his password is being changed does not exist in auth0', async function() {
        await enableCourseAdminScope(Constants.defaultUserAuth0Sub);
        const notExistingUserEmail = "notexisting@netapp-test.com";
        const govToken = await clients.auth0Client.getGovernanceUserToken(Constants.defaultTestUserEmail, Constants.defaultTestUserPassword);
        const promise = clients.governanceClient.resetPassword(govToken, notExistingUserEmail, "blabla");
        return expect(promise)
            .to.eventually.be.rejected
            .and.be.an.instanceOf(Error)
            .then(async error => {
                expect(error.response.data.message).to.equal(`Could not find user with Email ${notExistingUserEmail}.`);
                expect(error.response.status).to.equal(400);
            })
    });
});


