const chai = require("chai");
const expect = require('chai').expect;
const Constants = require('../utils/Constants');
const Initializer = require('../utils/specInitializer');
const Promise = require('bluebird');
const jwtDecoder = require('../utils/general-utils').decodeToken;
const _ = require('lodash');
const chaiAsPromised = require("chai-as-promised");
chai.use(chaiAsPromised);

const { logger, clients } = Initializer('OCCM Integration Spec');

describe('Governance developer API Spec - manage client details', function() {

    beforeEach(async function () {
        logger.debug('Inside before');
        await resetAppMetadataAttributes();
    });

    const resetAppMetadataAttributes = async () => {
        await clients.auth0Client.updateUserDetails(Constants.testimioUserAuth0Sub, {"app_metadata": {}});
        await clients.auth0Client.updateUserDetails(Constants.defaultUserAuth0Sub, {"app_metadata": {}});
    };

    const enableSupportScope = async (auth0UserId) => {
        await clients.auth0Client.updateUserDetails(auth0UserId, {"app_metadata": {"cc:support": true}});
    };

    it('Should return user\'s blocks if API called by user with cc:support scope', async function() {
        await enableSupportScope(Constants.defaultUserAuth0Sub);
        const blockUser = (userEmail) =>{
            const loop = () => {
                return clients.auth0Client.getCloudCentralUserToken(userEmail, "fakePassword")
                    .catch(error => {
                        if (error.response.data.error_description.includes(`Wrong email or password.`)) {
                            return loop();
                        } else {
                            return Promise.resolve(`Successfully blocked user`);
                        }
                    });
            };
            return loop();
        };
        const response = await blockUser(Constants.courseStudetEmail);
        logger.info(JSON.stringify(response));
        const govToken = await clients.auth0Client.getGovernanceUserToken(Constants.defaultTestUserEmail, Constants.defaultTestUserPassword);
        let blocks = await clients.governanceClient.getBlocks(govToken,Constants.courseStudetAuth0Sub);
        logger.info(JSON.stringify(blocks));
        const existBlocks = _.find(blocks, {'identifier': Constants.courseStudetEmail, 'connection': 'Username-Password-Authentication'});
        logger.info(`${JSON.stringify(existBlocks)}`);
        // noinspection BadExpressionStatementJS
        expect(existBlocks).to.not.be.empty;
        const userDeleted = await clients.governanceClient.deleteUserBlock(govToken,Constants.courseStudetAuth0Sub);
        expect(userDeleted).to.equal(204);
        blocks = await clients.governanceClient.getBlocks(govToken,Constants.courseStudetAuth0Sub);
        // noinspection BadExpressionStatementJS
        expect(blocks).to.be.empty;
    });

    it('Should return Url to reset password if the user has cc:support scope', async function() {
        await enableSupportScope(Constants.defaultUserAuth0Sub);
        const userToken = await clients.auth0Client.getGovernanceUserToken(Constants.defaultTestUserEmail, Constants.defaultTestUserPassword);
        const URL = await clients.governanceClient.passwordChangeTicket(userToken, Constants.courseStudetAuth0Sub);
        logger.info(URL.resetPasswordUrl);
        // noinspection BadExpressionStatementJS
        expect(URL.resetPasswordUrl).to.not.be.empty;
    });

    it('Should return user\'s Auht0 logs if the API caller has cc:support scope', async function() {
        await enableSupportScope(Constants.defaultUserAuth0Sub);
        const userToken = await clients.auth0Client.getGovernanceUserToken(Constants.defaultTestUserEmail, Constants.defaultTestUserPassword);
        const response = await clients.governanceClient.getLogs(userToken, Constants.courseStudetAuth0Sub,0,50);
        const allLogsAreForCorrectUser = response.logs.reduce((isAllCourseStudentLogs, log) => {
            const specificLogClientId = log.user_id === Constants.courseStudetAuth0Sub;
            return isAllCourseStudentLogs && specificLogClientId;
        }, true);
        expect(allLogsAreForCorrectUser).to.equal(true);
    });

    it('Should not allow user to retrieve blocks list if doesn\'t have cc:support scope', async function() {
        const govToken = await clients.auth0Client.getGovernanceUserToken(Constants.defaultTestUserEmail, Constants.defaultTestUserPassword);
        const promise = clients.governanceClient.getBlocks(govToken,Constants.courseStudetAuth0Sub);
        return expect(promise)
            .to.eventually.be.rejected
            .and.be.an.instanceOf(Error)
            .then(async error => {
                expect(error.response.data.message).to.equal('Forbidden');
                expect(error.response.status).to.equal(403);
            })
    });

    it('Should not allow user to retrieve logs list if doesn\'t have cc:support scope', async function() {
        const govToken = await clients.auth0Client.getGovernanceUserToken(Constants.defaultTestUserEmail, Constants.defaultTestUserPassword);
        const promise = clients.governanceClient.getLogs(govToken,Constants.courseStudetAuth0Sub,0,50);
        return expect(promise)
            .to.eventually.be.rejected
            .and.be.an.instanceOf(Error)
            .then(async error => {
                expect(error.response.data.message).to.equal('Forbidden');
                expect(error.response.status).to.equal(403);
            })
    });

});