const Constants = require('../Constants');
const {authAxios, authManagementAxios} = require('./auth0Axios');
const _ = require('lodash');

class Auth0Client {

    constructor(logger) {
        this.logger = logger;
    }

    getCloudCentralUserToken(email, password) {
        this.logger.info(`Getting Cloud Central API auth0 token for ${email} from ${Constants.auth0BaseUrl}/oauth/token`);
        return authAxios().post(`/oauth/token`, {
            username: email,
            password: password,
            audience: Constants.cloudCentralAudience,
            client_id: Constants.cloudCentralClientId,
            grant_type: "password",
            scope: "openid profile cc:clients cc:course-admin cc:update-password cc:customer-stories cc:support cc:netapp-internal cc:tenancy"
        })
            .then(response => response.access_token)
    }

    getCloudSyncUserToken(email, password) {
        this.logger.info(`Getting Cloud Sync API auth0 token for ${email} from ${Constants.auth0BaseUrl}/oauth/token`);
        return authAxios().post(`/oauth/token`, {
            username: email,
            password: password,
            audience: Constants.cloudSyncAudience,
            client_id: Constants.cloudSyncClientId,
            grant_type: "password",
            scope: "openid profile"
        })
            .then(response => response.access_token)
    }

    getGovernanceUserToken(email, password) {
        this.logger.info(`Getting Cloud Central Governance auth0 token for ${email} from ${Constants.auth0BaseUrl}/oauth/token`);
        return authAxios().post(`/oauth/token`, {
            username: email,
            password: password,
            audience: Constants.cloudCentralGovernanceAudience,
            client_id: Constants.cloudCentralGovernanceClientId,
            grant_type: "password",
            scope: "openid profile cc:clients cc:course-admin cc:update-password cc:customer-stories cc:support"
        })
            .then(response => response.access_token)
    }

    getOccmRegistratorToken() {
        this.logger.info(`Getting registrator's token`);
        return authAxios().post(`/oauth/token`, {
            audience: Constants.occmRegistratorAudience,
            client_id: Constants.occmRegistratorClientId,
            client_secret: Constants.occmRegistrarClientSecret,
            grant_type: "client_credentials"
        })
            .then(response => response.access_token)
    }

    getOccmToken(clientId, clientSecret) {
        this.logger.info(`Getting OCCM auth0 token for client ${clientId} from ${Constants.auth0BaseUrl}/oauth/token`);
        return authAxios().post(`/oauth/token`, {
            audience: Constants.cloudCentralAudience,
            client_secret: clientSecret,
            client_id: clientId,
            grant_type: "client_credentials"
        })
            .then(response => response.access_token)
    }

    getUsers(returnedFields) {
        this.logger.info(`Getting fields: {${returnedFields}} of users`);
        return this.getAllEntities('users', returnedFields);
    }

    getClients(returnedFields) {
        this.logger.info(`Getting fields: {${returnedFields}} of clients`);
        return this.getAllEntities('clients', returnedFields);
    }

    getAllEntities(entity = 'users', returnedFields) {
        // example of returned fields: 'email,name' (comma separator string of user detail fields that will be returned if they exist)
        let data = [];
        let page = 0;
        const paginate = () => {
            return authManagementAxios().then(authAxiosClient => {
                return authAxiosClient.get(`${entity}?per_page=100&page=${page}&include_totals=true&fields=${returnedFields}&include_fields=true`);
            })
                .then(response => {
                    this.logger.debug(`paginated page ${page}`);
                    this.logger.debug(`returned ${entity} ${response.users ? response.users.length : response.clients.length}`);
                    this.logger.debug(`${response.start} + ${response.limit} >= ${response.total} ? ${(response.start + response.limit) >= (response.total)}`);
                    data = _.concat(data, response[entity]);
                    page++;
                    if((response.start + response.limit) >= (response.total)) {
                        return data;
                    } else {
                        this.logger.info('paginating');
                        return paginate();
                    }
                })
        };
        return paginate();
    }

    getUsersByQuery(filterExp, returnedFields) {
        this.logger.info(`Getting fields ${returnedFields} of users filtered by ${filterExp}`);
        // example of filter expression: 'email:rony.levari@netapp.com' (will return that user's details if found in Auth0 tenant)
        // example of returned fields: 'email,name' (comma separator string of user detail fields that will be returned if they exist)
        let data = [];
        let page = 0;
        const paginate = () => {
            return authManagementAxios().then(authAxiosClient => {
                return authAxiosClient.get(`users?per_page=100&page=${page}&include_totals=true&fields=${returnedFields}&include_fields=true&q=${filterExp}&search_engine=v3`);
            })
                .then(response => {
                    this.logger.debug(`paginated page ${page}`);
                    this.logger.debug(`returned users ${response.users.length}`);
                    this.logger.debug(`${response.start} + ${response.limit} >= ${response.total} ? ${(response.start + response.limit) >= (response.total)}`);
                    data = _.concat(data, response.users);
                    page++;
                    if((response.start + response.limit) >= (response.total)) {
                        return data;
                    } else {
                        this.logger.info('paginating');
                        return paginate();
                    }
                })
        };
        return paginate();
    }

    getClientById(clientId) {
        this.logger.info(`Getting applications ${clientId} info from auth0`);
        return this.getEntityById('clients', clientId);
    }

    getUserById(userId) {
        this.logger.info(`Getting user ${userId} info from auth0`);
        return this.getEntityById('users', userId);
    }

    getEntityById(entity = 'users', id) {
        return authManagementAxios().then(authAxiosClient => {
            return authAxiosClient.get(`${entity}/${id}`)
        });
    }

    deleteClient(clientId) {
        this.logger.info(`Deleting OCCM client ${clientId} from auth0`);
        return this.deleteEntityById('clients', clientId);
    }

    deleteUser(userId) {
        this.logger.info(`Deleting user ${userId} from auth0`);
        return this.deleteEntityById('users', userId);
    }

    deleteEntityById(entity = 'users', id) {
        return authManagementAxios().then(authAxiosClient => {
            return authAxiosClient.delete(`${entity}/${id}`)
        });
    }

    updateUserDetails(userId, body){
        this.logger.info(`Updating user ${userId} with details ${JSON.stringify(body)}`);
        return authManagementAxios().then(authAxioClient => {
            return authAxioClient.patch(`users/${userId}`, body);
        });
    }

    getClientGrants(clientId){
        this.logger.info(`Get all clients grants by auth0 for clientId = ${clientId} `);
        return authManagementAxios().then(authAxioClient => {
            return authAxioClient.get(`client-grants?client_id=${clientId}`);
        });
    }
}

module.exports = Auth0Client;





