const Constants = require('../Constants');
const jwtDecoder = require('../general-utils').decodeToken;
const cloudCentralAxios = require('./cloudCentralAxios').cloudCentralAxios();

class PortalOccmClient {
    constructor(logger) {
        this.logger = logger;
    }

    serviceRegisterOccm(systemId, occmName, userToken, isAgent = false) {
        this.logger.info(`Registering new OCCM ${systemId}, ${occmName} on behalf of user ${jwtDecoder(userToken).sub} from ${Constants.cloudCentralApiBaseUrl}/occm/service-register-occm`);
        return cloudCentralAxios.post(`/occm/service-register-occm`, null, {
            data: {
                systemId: systemId,
                occmName: occmName,
                setupUserId: jwtDecoder(userToken).sub,
                isAgent: isAgent
            },
            headers: {
                Authorization: `Bearer ${userToken}`
            }
        });
    }

    updateRedirectUris(uris, occmToken) {
        this.logger.info(`Updating URIs of OCCM ${jwtDecoder(occmToken).sub} in ${Constants.cloudCentralApiBaseUrl}/occm/update-uris`);
        return cloudCentralAxios.post(`/occm/update-uris`, {redirectUris: uris}, {
            headers: {
                Authorization: `Bearer ${occmToken}`
            }
        });
    }

    updateRedirectUrisManually(userToken, clientId, redirectUris) {
        this.logger.info(`User ${jwtDecoder(userToken).sub} is updating OCCM's ${clientId}'s URIs from ${Constants.cloudCentralApiBaseUrl}/occm/manual-update-uris`);
        return cloudCentralAxios.post(`/occm/manual-update-uris`, null, {
            data: {
                redirectUris: redirectUris,
                occmId: clientId
            },
            headers: {
                Authorization: `Bearer ${userToken}`
            }
        });
    }

    serviceSetupFailed(userToken, clientId) {
        this.logger.info(`User ${jwtDecoder(userToken).sub}'s OCCM ${clientId} setup failure is being reported from ${Constants.cloudCentralApiBaseUrl}/occm/service-setup-failed`);
        return cloudCentralAxios.post(`/occm/service-setup-failed`, {clientId: clientId}, {
            headers: {
                Authorization: `Bearer ${userToken}`
            }
        });
    }

    registerOccm(redirectUris, systemId, registratorToken) {
        this.logger.info(`Registering OCCM ${systemId} on behalf of client ${jwtDecoder(registratorToken).sub} from ${Constants.cloudCentralApiBaseUrl}/occm/register-occm`);
        return cloudCentralAxios.post(`/occm/register-occm`, null, {
            data: {
                systemId: systemId,
                redirectUris: redirectUris
            },
            headers: {
                Authorization: `Bearer ${registratorToken}`
            }
        });
    }

    setupComplete(systemId, occmToken) {
        this.logger.info(`Complete OCCM client ${jwtDecoder(occmToken).sub} registration in ${Constants.cloudCentralApiBaseUrl}/occm/setup-complete`);
        return cloudCentralAxios.post(`/occm/setup-complete`, null, {
            data: {
                occmName: `OccmName-${systemId}`,
                setupUserId: jwtDecoder(occmToken).sub
            },
            headers: {
                Authorization: `Bearer ${occmToken}`
            }
        });
    }

    deassociateSelf(userToken, clientId) {
        this.logger.info(`Deassociating user ${jwtDecoder(userToken).sub} from OCCM ${clientId} on behalf of himself from ${Constants.cloudCentralApiBaseUrl}/occm/deassociate-self`);
        return cloudCentralAxios.post(`/occm/deassociate-self`, {occmId: clientId}, {
            headers: {
                Authorization: `Bearer ${userToken}`
            }
        });
    }

    associateUser(occmToken, associateeEmail, associaterAuth0Id) {
        this.logger.info(`Associating user ${associateeEmail} with OCCM ${jwtDecoder(occmToken).sub} on behalf of ${associaterAuth0Id} from ${Constants.cloudCentralApiBaseUrl}/occm/associate-user`);
        return cloudCentralAxios.post(`/occm/associate-user`, null, {
            data: {
                associateeEmail: associateeEmail,
                associaterId: associaterAuth0Id
            },
            headers: {
                Authorization: `Bearer ${occmToken}`
            }
        });
    }

    associateUserWithRawData(occmToken, data) {
        this.logger.info(`Calling API Associate user with raw data object`);
        return cloudCentralAxios.post(`/occm/associate-user`, data, {
            headers: {
                Authorization: `Bearer ${occmToken}`,
                "Content-Type": "application/json"
            }
        });
    }

    deassociateUser(occmToken, deassociaterId, deassociateeId) {
        this.logger.info(`Dessociating user ${deassociateeId} with OCCM ${jwtDecoder(occmToken).sub} on behalf of ${deassociaterId} from ${Constants.cloudCentralApiBaseUrl}/occm/deassociate-user`);
        return cloudCentralAxios.post(`/occm/deassociate-user`, null, {
            data: {
                deassociaterId: deassociaterId,
                deassociateeId: deassociateeId
            },
            headers: {
                Authorization: `Bearer ${occmToken}`
            }
        });
    }

    listUris(occmToken) {
        this.logger.info(`Listing OCCM ${jwtDecoder(occmToken).sub} URIs from ${Constants.cloudCentralApiBaseUrl}/occm/list-uris`);
        return cloudCentralAxios.get(`/occm/list-uris`, {
            headers: {
                Authorization: `Bearer ${occmToken}`
            }
        });
    }

    listUserOccms(userToken) {
        this.logger.info(`Fetching user's ${jwtDecoder(userToken).sub} OCCMs from ${Constants.cloudCentralApiBaseUrl}/occm/list-occms`);
        return cloudCentralAxios.get(`/occm/list-occms`, {
            headers: {
                Authorization: `Bearer ${userToken}`
            }
        });
    }

    listOccmUsers(occmToken, usersString) {
        this.logger.info(`Getting users ${usersString} details from Auth0 on behalf of OCCM ${jwtDecoder(occmToken).sub}, from ${Constants.cloudCentralApiBaseUrl}/occm/users?usersString=${usersString}`);
        return cloudCentralAxios.get(`/occm/users?usersString=${usersString}`, {
            headers: {
                Authorization: `Bearer ${occmToken}`
            }
        });
    }
}

module.exports = PortalOccmClient;




