import React from 'react'
import querystring from 'query-string';
import './App.scss'
import './Buttons.scss'
import './components/Components.scss'
import AppHeader from 'components/AppHeader'
import HowToGetStarted from './components/getting-started/HowToGetStarted'
import RefreshTokenGenerator from 'components/misc/RefreshTokenGenerator'
import DeveloperHub from 'components/developer-hub/DeveloperHub'
import { Route, Switch, Redirect } from 'react-router-dom'
import { connect } from 'react-redux'
import ContentRedirector from './components/content/ContentRedirector';
import BlankPage from './components/BlankPage'
import links from './consts/links'
import { CloudManagersView, CloudAssessmentView, FabricView, NetAppInternalView } from './chunks.js'
import Intercom from "./components/intercom/Intercom"
import LaunchPadLoader from "./components/launch-pad/LaunchPadLoader"
import NavMenu from "./components/nav-menu/NavMenu"
import classNames from "classnames"
import SubscriptionMapping from "./components/subscriptions/SubscriptionMapping";
import MyServices from "./components/launch-pad/MyServices"
import DemoView from "./components/demo/DemoView";
import * as Dashboards from "./components/dashboards/Dashboards";

const LoginRedirectToPath = () => {
    window.auth.loginRedirect({}, { service: "portal", startOnSignup: false }, { pathname: window.location.pathname + window.location.search });
    return <BlankPage/>
};

const AuthenticatedRoute = props => props.auth.authenticated ? <Route {...props}/> : <LoginRedirectToPath/>;
const InternalRoute = props => props.auth.authenticated && props.auth.isNetAppInternal ?
    <Route {...props}/> :
    (props.auth.authenticated ? <Redirect to="/"/> : <LoginRedirectToPath/>);

const DevRedirect = () => <div>
    In dev/staging there is no redirect to marketing homepage.

    <br/><a href="https://cloud.netapp.com">Go to marketing.</a>
    <br/>(Note: If you login from marketing it will be to production.)
    <a href="/?redirectToLogin=true">Start free trial</a>
</div>;

const RootRoute = ({ auth, activeFeatures }) => {
    if (!auth.authenticated) {
        if (links.marketingPageRedirect.startsWith("/")) {
            return <DevRedirect/>
        } else {
            window.location.replace(links.marketingPageRedirect);
            return <BlankPage/>
        }
    }

    return activeFeatures.gettingStarted ? <HowToGetStarted/> : <Redirect to="/my-services"/>
};

const DashboardRoutes = () => <Switch>
    <Route path="/dashboard/cloud-volumes-ontap" render={() => <Dashboards.CloudVolumesOntapDashboard/>}/>
    <Route path="/dashboard/cloud-volumes-service" render={() => <Dashboards.CloudVolumesServiceDashboard/>}/>
    <Route path="/dashboard/netapp-kubernetes-service" render={() => <Dashboards.NetAppKubernetesServiceDashboard/>}/>
    <Route path="/dashboard/cloud-insights" render={() => <Dashboards.CloudInsightsDashboard/>}/>
    <Route path="/dashboard/cloud-sync" render={() => <Dashboards.CloudSyncDashboard/>}/>
    <Route path="/dashboard/saas-backup" render={() => <Dashboards.SaaSBackupDashboard/>}/>
    <Route path="/dashboard/cloud-secure" render={() => <Dashboards.CloudSecureDashboard/>}/>
    <Route path="/dashboard/cloud-tiering" render={() => <Dashboards.CloudTieringDashboard/>}/>
    <Route path="/dashboard/cloud-backup-service" render={() => <Dashboards.CloudBackupServiceDashboard/>}/>
    <Redirect to="/"/>
</Switch>;

const MainRoutes = ({ auth, activeFeatures, subscriptionDialogOpen }) => <Switch>
    <Route path="/" exact render={() => <RootRoute activeFeatures={activeFeatures} auth={auth}/>}/>
    <Route path="/login" render={() => <div/>}/>
    <AuthenticatedRoute auth={auth} path="/cloud-managers" exact render={() => <CloudManagersView/>}/>
    <AuthenticatedRoute auth={auth} path="/cloud-assessment" render={() => <CloudAssessmentView/>}/>
    <AuthenticatedRoute auth={auth} path="/refresh-token" exact render={() => <RefreshTokenGenerator/>}/>
    <AuthenticatedRoute auth={auth} path="/fabric-view" exact render={() => <FabricView/>}/>
    <AuthenticatedRoute auth={auth} path="/subscription-mapping" render={({location}) => {
        const search = querystring.parse(location.search);
        if(search.token || subscriptionDialogOpen) return <SubscriptionMapping {...search}/>;
        else return <Redirect to="/"/>;
    }}/>
    {/*<AuthenticatedRoute auth={auth} path="/workloads" exact render={() => <Workloads/>}/>*/}
    <AuthenticatedRoute auth={auth} path="/my-services" exact render={() => <MyServices/>}/>
    <AuthenticatedRoute auth={auth} path="/dashboard" render={() => <DashboardRoutes/>}/>
    <InternalRoute auth={auth} path="/netapp-internal" render={() => <NetAppInternalView/>}/>
    <InternalRoute auth={auth} path="/demo" render={() => <DemoView/>}/>
    <Route path="/developer-hub" exact render={() => <DeveloperHub/>}/>
    <ContentRedirector auth={auth} path="/content/:content" exact/>
    <Redirect to="/"/>
</Switch>;

const RoutesWithHeader = ({ auth, menuOpen, activeFeatures, subscriptionDialogOpen }) => (
    <div className={classNames("App", { 'nav-menu-open': menuOpen })}>
        <Intercom/>
        <AppHeader/>
        <NavMenu/>
        {auth.authenticated && <LaunchPadLoader/>}
        <MainRoutes activeFeatures={activeFeatures} auth={auth} subscriptionDialogOpen={subscriptionDialogOpen}/>
    </div>);

const mapStateToProps = state => {
    return {
        auth: state.auth,
        menuOpen: state.nav.menuOpen,
        activeFeatures: state.features.active,
        subscriptionDialogOpen: state.subscriptions.subscriptionMappingOpen
    };
};

export default connect(mapStateToProps)(RoutesWithHeader);
