import React from 'react';
import { connect } from 'react-redux';
import _ from 'lodash'

import styles from "./Accounts.module.scss"
import { ReactComponent as UserIcon } from "ui-components/src/images/monochrome-icons/ic-user.svg";
import buttons from "ui-components/src/button/button.module.scss";
import { FormFieldInput } from "ui-components"
import Loader from "../../Loader";
import FormFieldSelect from "./../widgets/FormFieldSelect"
import { ReactComponent as InlineLoader } from "ui-components/src/images/loaders/inline-loader.svg";

class AddUserOverlay extends React.Component {
    constructor(props) {
        super(props);

        this.roleSelectInput = React.createRef();

        this.handleInputName = this.handleInputName.bind(this);
        this.nextClicked = this.nextClicked.bind(this);
        this.associateUserClicked = this.associateUserClicked.bind(this);
    }

    nextClicked(event) {
        event.preventDefault();
        if (this.emailInput.checkValidity()) {
            const associatedEmail = this.state[this.emailInput.name];
            this.props.checkIfAssociatedEmailExistsInCloudCentral(associatedEmail);
        } else {
            this.props.setAssociatedEmailInputFeedback(this.emailInput.validationMessage);
        }
    }

    associateUserClicked(event) {
        event.preventDefault();

        if (!this.state.selectedRole) {
            this.setState({ roleIsRequired: true });
        } else {
            this.props.associateUser(this.state.associatedEmail, this.state.selectedRole.value, _.map(this.state.selectedWorkspaces, 'value'));
        }
    }

    handleInputName(event) {
        this.props.setAssociatedEmailInputFeedback(null);
        this.setState({
            [event.target.name]: event.target.value,
        });
    }

    handleRoleChange(selectedRole) {
        this.setState({
            selectedRole,
            selectedWorkspaces: selectedRole.value === "Role-1" ? null : this.state.selectedWorkspaces,
            roleIsRequired: null
        })
    }

    handleWorkspacesChange(selectedWorkspaces) {
        this.setState({ selectedWorkspaces });
    }

    componentDidUpdate(prevProps) {
        if (this.props.addUserOverlayState === "role" && prevProps.addUserOverlayState !== "role") {
            this.roleSelectInput.current.focus();
        }
    }

    render() {
        return (<div className={styles["add-user-overlay"]}>
            <UserIcon/>
            <div className={styles["title"]}>Associate User By Email</div>
            {this.props.addUserOverlayState === "email" && <form className={styles["email-stage"]}>
                <div className={styles['text']}>In order to add a user to your NetApp Cloud Account, that user must
                    already have signed up to Cloud Central. Sign up a new user at <a
                        href='/' target="_blank" rel="noopener noreferrer">NetApp Cloud Central</a>.
                </div>
                <FormFieldInput
                    type="email"
                    name="associatedEmail"
                    label="User's Email"
                    errorMessage={this.props.associatedEmailInputFeedback}
                    reference={input => this.emailInput = input}
                    onChange={this.handleInputName}
                    required={true}
                    autoFocus={true}
                    className={styles["associated-email-input"]}
                    pending={this.props.associatedEmailExistsInCloudCentralPending}
                    InlineLoaderIcon={InlineLoader}
                />
                <div className={styles["button-box"]}>
                    <button type="button" className={buttons.cancel}
                            onClick={() => this.props.setAddUserOverlayState({})}>Cancel
                    </button>
                    <button type="submit" className={buttons.primary} onClick={this.nextClicked}>Next</button>
                </div>
            </form>}
            {this.props.addUserOverlayState === "role" && <div className={styles["role-stage"]}>
                {this.props.addUserPending && <Loader/>}
                <div className={styles["user-email"]}>{this.state.associatedEmail}</div>
                <FormFieldSelect
                    label="Role"
                    className={styles["role-selection"]}
                    name="role-selector"
                    isSearchable={false}
                    value={this.state.selectedRole}
                    errorMessage={this.state.roleIsRequired && 'Please select'}
                    onChange={this.handleRoleChange.bind(this)}
                    isClearable={false}
                    placeholder="Select a role"
                    options={[
                        { value: 'Role-1', label: 'Account Admin' },
                        { value: 'Role-2', label: 'Workspace Admin' }
                    ]}
                    reference={this.roleSelectInput}
                />
                <FormFieldSelect
                    label=" Associate User to Workspaces"
                    name="workspaces-selector"
                    className={styles["workspaces-selection"]}
                    isSearchable={false}
                    value={this.state.selectedWorkspaces}
                    onChange={this.handleWorkspacesChange.bind(this)}
                    isClearable={true}
                    isDisabled={!this.state.selectedRole || this.state.selectedRole.value === 'Role-1'}
                    isMulti={true}
                    closeMenuOnSelect={false}
                    placeholder={this.state.selectedRole && this.state.selectedRole.value === 'Role-1' ? 'All Workspaces' : 'Select workspaces'}
                    options={this.props.accountWorkspaces.map(({ workspacePublicId, workspaceName }) => ({
                        value: workspacePublicId,
                        label: workspaceName
                    }))}
                />
                <div className={styles["button-box"]}>
                    <button type="button" className={buttons.cancel}
                            onClick={() => this.props.setAddUserOverlayState({})}>Cancel
                    </button>
                    <button type="submit" className={buttons.primary} onClick={this.associateUserClicked}>
                        Associate User
                    </button>
                </div>
                {this.props.addUserError && <div className={styles["error-message"]}>{this.props.addUserError}</div>}
            </div>}
        </div>)
    };
}

const mapStateToProps = state => {
    return {
        accountWorkspaces: state.accountsManagement.accountWorkspaces,
        addUserOverlayState: state.accountsManagement.addUserOverlayState,
        associatedEmailInputFeedback: state.accountsManagement.associatedEmailInputFeedback,
        associatedEmailExistsInCloudCentral: state.accountsManagement.associatedEmailExistsInCloudCentral,
        associatedEmailExistsInCloudCentralPending: state.accountsManagement.associatedEmailExistsInCloudCentralPending,
        addUserPending: state.accountsManagement.addUserPending,
        addUserError: state.accountsManagement.addUserError
    };
};

const mapDispatchToProps = dispatch => {
    return {
        setAddUserOverlayState: ({ newState }) => {
            dispatch({
                type: "ACCOUNTS-MANAGEMENT:SET-ADD-USER-OVERLAY-STATE",
                payload: newState
            })
        },
        checkIfAssociatedEmailExistsInCloudCentral: (userEmail) => {
            dispatch({
                type: "ACCOUNTS-MANAGEMENT:CHECK-IF-ASSOCIATED-EMAIL-EXISTS-IN-CLOUD-CENTRAL",
                payload: { userEmail }
            })
        },
        associateUser: (associatedEmail, rolePublicId, workspaces) => {
            dispatch({
                type: "ACCOUNTS-MANAGEMENT:ASSOCIATE-USER",
                payload: { associatedEmail, rolePublicId, workspaces }
            })
        },
        setAssociatedEmailInputFeedback: (newFeedback) => {
            dispatch({
                type: "ACCOUNTS-MANAGEMENT:SET-ASSOCIATED-EMAIL-INPUT-FEEDBACK",
                payload: newFeedback
            })
        }
    }
};

export default connect(mapStateToProps, mapDispatchToProps)(AddUserOverlay)
