import React from 'react';
import _ from "lodash";
import { connect } from 'react-redux';
import { AgGridReact } from "ag-grid-react";

import styles from "./Accounts.module.scss";
import WorkspacesCellRenderer from "./WorkspacesCellRenderer";
import AssociatedWorkspacesPopUp from "./AssociatedWorkspacesPopUp";
import { ManageAgentWorkspacesOverlay } from "./ManageWorkspacesOverlay";
import SearchWidget from "../widgets/SearchWidget";
import TableWidgetRow from "../widgets/TableWidgetRow";
import RowCounter from "../widgets/RowCounter";


const getColumnDefs = () => {
    const fields = [
        {
            headerName: 'Name',
            field: 'name',
            tooltipField: 'name',
            width: 600,
            cellClass: 'one-line-text',
            cellStyle: { 'line-height': `${rowHeight}px`}
        },
        { headerName: 'Workspaces', field: 'workspaces', cellRenderer: 'workspacesCellRenderer', width: 300 }
    ];

    const columnDefs = fields.map(field => _.assign(field, {
        suppressMenu: true,
        unSortIcon: true,
        sortable: true,
        lockPosition: true,
    }));
    columnDefs.push({
        headerName: "",
        field: "Name",
        cellRenderer: "addRemoveWorkspaceCellRenderer",
        colId: 'button',
        lockPosition: true,
        cellStyle: { 'justify-content': "flex-end" }
    });

    return columnDefs;
};

const AddRemoveWorkspaceCellRenderer = props => (
    <div key={props.data.name} className={styles["menu-cell"]}>
        <button className={styles["inline-grid-button"]}
                onClick={() => props.context.manageAgentWorkspacesPrompt({ agent: props.data })}>
            Add / Remove Workspaces
        </button>
    </div>
);

const rowHeight = 55; // px

class AgentsTab extends React.Component {
    constructor(props) {
        super(props);

        this.state = { gridScrollFromTop: 0, quickFilter: '' };

        this.columnDefs = getColumnDefs();
    }

    onFirstDataRendered(params) {
        this.api = params.api;
        params.api.sizeColumnsToFit();
    }

    onComponentStateChanged() {
        this.api.refreshCells({ force: true });
    }

    openAssociatedWorkspacesPopUp({ data, rowIndex }) {
        this.setState({ currentAgentWorkspacesPopUp: { agent: data, rowIndex } });
    }

    closeAssociatedWorkspacesPopUp() {
        this.setState({ currentAgentWorkspacesPopUp: null });
    }

    onGridScroll(event) {
        this.setState({ gridScrollFromTop: event.top });
    }

    onFilterChanged() {
        this.setState({ filteredCount: this.api.getModel().rootNode.childrenAfterFilter.length });
    }

    quickFilterChanged(value) {
        this.setState({ quickFilter: value });
        this.api.setQuickFilter(value);
    }

    getRowFilteredIndex(row) {
        return _.findIndex(this.api.getModel().rootNode.childrenAfterSort, { data: row.data });
    }

    render() {
        return (<div className={styles["agents-tab"]}>
            <h4>{this.props.account && `${this.props.account.accountName} Agents`}</h4>

            {this.state.currentAgentWorkspacesPopUp && <AssociatedWorkspacesPopUp
                workspacesList={this.state.currentAgentWorkspacesPopUp.agent.workspaces}
                rowIndex={this.state.currentAgentWorkspacesPopUp.rowIndex}
                closeAssociatedWorkspacesPopUp={this.closeAssociatedWorkspacesPopUp.bind(this)}
                rowHeight={rowHeight}
                baseTop={290 - this.state.gridScrollFromTop}
                left={685}
            />}
            {this.props.agents && <TableWidgetRow>
                <RowCounter filteredCount={this.state.filteredCount} totalCount={this.props.agents.length}
                            singularTitle="Agent" pluralTitle="Agents"/>
                <SearchWidget setFilter={this.quickFilterChanged.bind(this)}/>
            </TableWidgetRow>}
            <div className="ag-theme-balham portal-ag-grid">
                {this.props.agents && <AgGridReact
                    enableColResize={false}
                    columnDefs={this.columnDefs}
                    onBodyScroll={this.onGridScroll.bind(this)}
                    onFirstDataRendered={this.onFirstDataRendered.bind(this)}
                    onComponentStateChanged={this.onComponentStateChanged.bind(this)}
                    onFilterChanged={this.onFilterChanged.bind(this)}
                    suppressDragLeaveHidesColumns={true}
                    rowData={this.props.agents}
                    rowHeight={rowHeight}
                    headerHeight={rowHeight}
                    suppressColumnVirtualisation="true"
                    frameworkComponents={{
                        addRemoveWorkspaceCellRenderer: AddRemoveWorkspaceCellRenderer,
                        workspacesCellRenderer: WorkspacesCellRenderer
                    }}
                    context={{
                        openAssociatedWorkspacesPopUp: this.openAssociatedWorkspacesPopUp.bind(this),
                        manageAgentWorkspacesPrompt: this.props.manageAgentWorkspacesPrompt,
                        getRowFilteredIndex: this.getRowFilteredIndex.bind(this)
                    }}
                >
                </AgGridReact>}
            </div>
            {this.props.agentForWorkspacesManagement && <ManageAgentWorkspacesOverlay type="Agent"/>}
        </div>);
    }
}

const mapStateToProps = state => ({
    account: state.accountsManagement.account,
    agents: state.accountsManagement.agents,
    agentForWorkspacesManagement: state.accountsManagement.agentForWorkspacesManagement
});

const mapDispatchToProps = dispatch => ({
        manageAgentWorkspacesPrompt: ({ agent }) => {
            dispatch({
                type: "ACCOUNTS-MANAGEMENT:MANAGE-AGENT-WORKSPACES-PROMPT",
                payload: { agent }
            });
        }
    }
);

export default connect(mapStateToProps, mapDispatchToProps)(AgentsTab);
