import React from 'react';
import _ from "lodash";
import { connect } from 'react-redux';
import classNames from "classnames";
import { AgGridReact } from "ag-grid-react";

import buttons from "ui-components/src/button/button.module.scss";
import { ReactComponent as AccountAdminIcon } from "ui-components/src/images/monochrome-icons/ic-account-admin.svg";
import { ReactComponent as DotDotDotIcon } from "ui-components/src/images/monochrome-icons/ic-more.svg";

import styles from "./Accounts.module.scss";
import DisassociateUserOverlay from "./DisassociateUserOverlay";
import { ManageUserWorkspacesOverlay } from "./ManageWorkspacesOverlay";
import AddUserOverlay from "./AddUserOverlay";
import WorkspacesCellRenderer from "./WorkspacesCellRenderer";
import AssociatedWorkspacesPopUp from "./AssociatedWorkspacesPopUp";
import UserPopUpMenu from "./UserPopUpMenu";
import SearchWidget from "../widgets/SearchWidget";
import TableWidgetRow from "../widgets/TableWidgetRow";
import RowCounter from "../widgets/RowCounter";

const RoleCellRenderer = props => (
    <div className={styles["role-cell"]}>{props.data.userRole === 'Role-1' ?
        <span><AccountAdminIcon/>Account Admin</span> : 'Workspace Admin'}</div>
);

const MenuCellRenderer = props => (
    <div key={props.data.email} className={styles["menu-cell"]}>
        <button className={styles["inline-grid-button"]}
                onClick={() => props.context.openUserMenu({
                    user: props.data,
                    rowIndex: props.context.getRowFilteredIndex(props)
                })}>
            <DotDotDotIcon/>
        </button>
    </div>
);

const getColumnDefs = () => {
    const fields = [
        { headerName: 'Name', field: 'name' },
        { headerName: 'Email', field: 'email' },
        { headerName: 'Role', field: 'userRole', cellRenderer: 'roleCellRenderer' },
        { headerName: 'Workspaces', field: 'workspaces', cellRenderer: 'workspacesCellRenderer' }
    ];

    const columnDefs = fields.map(field => ({
        headerName: field.headerName,
        field: field.field,
        suppressMenu: true,
        unSortIcon: true,
        sortable: true,
        lockPosition: true,
        cellRenderer: field.cellRenderer
    }));
    columnDefs.push({
        headerName: "",
        field: "Name",
        cellRenderer: "menuCellRenderer",
        colId: 'button',
        lockPosition: true
    });

    return columnDefs;
};

const rowHeight = 55; // px
const baseTopForPopups = 290; //px

class UsersTab extends React.Component {
    constructor(props) {
        super(props);

        this.state = { gridScrollFromTop: 0, quickFilter: '' };

        this.columnDefs = getColumnDefs();
    }

    onFirstDataRendered(params) {
        this.api = params.api;
        params.api.sizeColumnsToFit();
    }

    onComponentStateChanged() {
        this.api.refreshCells({ force: true });
    }

    openAssociatedWorkspacesPopUp({ data, rowIndex }) {
        this.setState({ currentUserWorkspacesPopUp: { user: data, rowIndex } });
    }

    closeAssociatedWorkspacesPopUp() {
        this.setState({ currentUserWorkspacesPopUp: null });
    }

    onGridScroll(event) {
        this.setState({ gridScrollFromTop: event.top });
    }

    onFilterChanged() {
        this.setState({ filteredCount: this.api.getModel().rootNode.childrenAfterFilter.length });
    }

    quickFilterChanged(value) {
        this.setState({ quickFilter: value });
        this.api.setQuickFilter(value);
    }

    getRowFilteredIndex(row) {
        return _.findIndex(this.api.getModel().rootNode.childrenAfterSort, { data: row.data });
    }

    render() {
        return <div className={styles["users-tab"]}>
            <button className={classNames(buttons.primary, styles["associate-user-button"])}
                    onClick={() => this.props.setAddUserOverlayState({ newState: 'email' })}
                    disabled={this.props.loadingData || this.props.accountManagementError}>+ Associate User
            </button>
            <h4>{this.props.account && `${this.props.account.accountName} Users`}</h4>
            {this.props.currentUserMenu &&
            <UserPopUpMenu user={this.props.currentUserMenu.user}
                           rowIndex={this.props.currentUserMenu.rowIndex}
                           rowHeight={rowHeight}
                           isLastUserInAccount={this.props.accountUsers.length === 1}
                           closeUserMenu={this.props.closeUserMenu}
                           disassociateUserPrompt={this.props.disassociateUserPrompt}
                           manageUserWorkspacesPrompt={this.props.manageUserWorkspacesPrompt}
                           baseTop={baseTopForPopups - this.state.gridScrollFromTop}
                           loggedInUserId={this.props.loggedInUserId}/>
            }
            {this.state.currentUserWorkspacesPopUp && <AssociatedWorkspacesPopUp
                workspacesList={this.state.currentUserWorkspacesPopUp.user.workspaces}
                rowIndex={this.state.currentUserWorkspacesPopUp.rowIndex}
                closeAssociatedWorkspacesPopUp={this.closeAssociatedWorkspacesPopUp.bind(this)}
                rowHeight={rowHeight}
                baseTop={baseTopForPopups - this.state.gridScrollFromTop}
                left={755}
            />}
            {this.props.accountUsers && <TableWidgetRow>
                <RowCounter filteredCount={this.state.filteredCount} totalCount={this.props.accountUsers.length}
                            singularTitle="User" pluralTitle="Users"/>
                <SearchWidget setFilter={this.quickFilterChanged.bind(this)}/>
            </TableWidgetRow>}
            <div className={classNames("ag-theme-balham", "portal-ag-grid")}>
                {this.props.accountUsers && <AgGridReact
                    enableColResize={false}
                    columnDefs={this.columnDefs}
                    onBodyScroll={this.onGridScroll.bind(this)}
                    onFirstDataRendered={this.onFirstDataRendered.bind(this)}
                    onComponentStateChanged={this.onComponentStateChanged.bind(this)}
                    onFilterChanged={this.onFilterChanged.bind(this)}
                    suppressDragLeaveHidesColumns={true}
                    rowData={this.props.accountUsers}
                    rowHeight={rowHeight}
                    headerHeight={rowHeight}
                    suppressColumnVirtualisation="true"
                    frameworkComponents={{
                        menuCellRenderer: MenuCellRenderer,
                        roleCellRenderer: RoleCellRenderer,
                        workspacesCellRenderer: WorkspacesCellRenderer
                    }}
                    context={{
                        openUserMenu: this.props.openUserMenu,
                        openAssociatedWorkspacesPopUp: this.openAssociatedWorkspacesPopUp.bind(this),
                        getRowFilteredIndex: this.getRowFilteredIndex.bind(this)
                    }}
                >
                </AgGridReact>}
            </div>
            {this.props.userForRemoval && <DisassociateUserOverlay/>}
            {this.props.userForWorkspacesManagement && <ManageUserWorkspacesOverlay/>}
            {this.props.addUserOverlayState && <AddUserOverlay/>}
        </div>;
    }
}

const mapStateToProps = state => ({
    account: state.accountsManagement.account,
    accountUsers: state.accountsManagement.accountUsers,
    currentUserMenu: state.accountsManagement.currentUserMenu,
    currentUserWorkspacesPopUp: state.accountsManagement.currentUserWorkspacesPopUp,
    userForRemoval: state.accountsManagement.userForRemoval,
    userForWorkspacesManagement: state.accountsManagement.userForWorkspacesManagement,
    addUserOverlayState: state.accountsManagement.addUserOverlayState,
    loadingData: state.accountsManagement.loadingData,
    accountManagementError: state.accountsManagement.accountManagementError,
    loggedInUserId: state.auth.userId
});

const mapDispatchToProps = dispatch => ({
    openUserMenu: ({ user, rowIndex }) => {
        dispatch({
            type: "ACCOUNTS-MANAGEMENT:OPEN-USER-MENU",
            payload: { user, rowIndex }
        });
    },
    closeUserMenu: () => {
        dispatch({
            type: "ACCOUNTS-MANAGEMENT:CLOSE-USER-MENU"
        });
    },
    disassociateUserPrompt: ({ user }) => {
        dispatch({
            type: "ACCOUNTS-MANAGEMENT:DISASSOCIATE-USER-PROMPT",
            payload: user
        });
    },
    manageUserWorkspacesPrompt: ({ user }) => {
        dispatch({
            type: "ACCOUNTS-MANAGEMENT:MANAGE-USER-WORKSPACES-PROMPT",
            payload: user
        });
    },
    setAddUserOverlayState: ({ newState }) => {
        dispatch({
            type: "ACCOUNTS-MANAGEMENT:SET-ADD-USER-OVERLAY-STATE",
            payload: newState
        });
    }
});

export default connect(mapStateToProps, mapDispatchToProps)(UsersTab);
