import AuthenticationContext from 'adal-angular';
import axios from 'axios';
import _ from 'lodash'
import {getDiskPrice, getSnapshotPrice} from './getAzurePrices';

const baseAzureUri = `${window.location.origin}/cloud-assessment/azure`;

const config = {
    clientId: '2cf9a636-f5d0-46b5-98df-8e76daaa807a',
    redirectUri: `${baseAzureUri}${"-redirect"}`,
    postLogoutRedirectUri: `${baseAzureUri}?relogin`,
    cacheLocation: 'localStorage',
    loginResource: 'https://management.azure.com/',
    navigateToLoginRequestUrl: false
};

const authContext = new AuthenticationContext(config);
authContext._getNavigateUrl_Bak = authContext._getNavigateUrl;
authContext._getNavigateUrl = function(responseType, resource){
    if(responseType === "id_token" && !resource){
        responseType = 'id_token token';
        resource = 'https://management.azure.com/';
    }

    return authContext._getNavigateUrl_Bak(responseType, resource);
};

const getAll = (url, token) => {
    return axios.get(url, {headers: {Authorization: `Bearer ${token}`}})
        .then((headResponse) => {
            if (headResponse.data.nextLink) {
                return getAll(headResponse.data.nextLink, token)
                    .then((tailResponse) => {
                        return headResponse.data.value.concat(tailResponse);
                    });
            } else {
                return headResponse.data.value;
            }
        });
};

const getDisks = (subscriptionId, token) => {
    const url = `https://management.azure.com/subscriptions/${subscriptionId}/providers/Microsoft.Compute/disks?api-version=2018-06-01`;

    return getAll(url, token);
};

const getSnapshots = (subscriptionId, token) => {
    const url = `https://management.azure.com/subscriptions/${subscriptionId}/providers/Microsoft.Compute/snapshots?api-version=2018-06-01`;

    return getAll(url, token);
};

const getVMs = (subscriptionId, token) => {
    const url = `https://management.azure.com/subscriptions/${subscriptionId}/providers/Microsoft.Compute/virtualMachines?api-version=2018-10-01`;

    return getAll(url, token);
};

const findUnassociatedSnapshots = (disks, snapshots) => {
    const disksById = _.keyBy(disks, disk => disk.id.toLowerCase());

    return _.reject(snapshots, (snapshot) => {
        return snapshot.properties.creationData.sourceResourceId && disksById[snapshot.properties.creationData.sourceResourceId.toLowerCase()];
    });
};

const getDisksWithoutSnapshots = (disks, snapshots) => {
    const disksById = _.keyBy(disks, disk => disk.id.toLowerCase());

    _.forEach(snapshots, snapshot => {
        if (snapshot.properties.creationData.sourceResourceId && disksById[snapshot.properties.creationData.sourceResourceId.toLowerCase()]) {
            disksById[snapshot.properties.creationData.sourceResourceId.toLowerCase()].hasSnapshots = true;
        }
    });

    return _.reject(disksById, 'hasSnapshots');
};

const rpoOfVolume = ({snapshots, reportTime}) => {
    const latestSnapshotTime = _(snapshots)
        .map('properties')
        .map('timeCreated')
        .map(dateString => new Date(dateString).valueOf())
        .sort()
        .last();

    return reportTime - latestSnapshotTime;
};

const calculateAverageRpo = (associatedSnapshots, reportTime) => {
    if (associatedSnapshots.length === 0) {
        return -1;
    } else {
        return _(associatedSnapshots)
            .groupBy('properties.creationData.sourceResourceId')
            .map(snapshots => rpoOfVolume({snapshots, reportTime}))
            .mean()
    }
};

const isOntapDisk = (disk, vmsById) => {
    return disk.managedBy && _.get(vmsById[disk.managedBy.toLowerCase()], 'tags.WorkingEnvironmentId');
};

const isOntapSnapshot = (snapshot, ontapDisksById) => {
    return snapshot.properties.creationData.sourceResourceId && ontapDisksById[snapshot.properties.creationData.sourceResourceId.toLowerCase()];
};

export const login = () => {
    authContext.login();
};

export const acquireAccessToken = (azureToken) => {
    return new Promise((resolve, reject) => {
        const requestInfo = authContext.getRequestInfo(azureToken);
        if(!requestInfo.valid){
            reject(new Error("Invalid login return value."))
        } else if (requestInfo.parameters.error) {
            reject(new Error(requestInfo.parameters.error_description.split('\n')[0]));
        } else {
            resolve({expireTimeout: requestInfo.parameters.expires_in * 1000,accessToken: requestInfo.parameters.access_token, idTokenPayload: authContext._extractIdToken(requestInfo.parameters.id_token)});
        }
    })
};

export const getAzureSubscriptions = (azureAccessToken) => {
    return axios.get('https://management.azure.com/subscriptions?api-version=2016-06-01', {headers: {Authorization: `Bearer ${azureAccessToken}`}})
};

export const logout = () => {
    authContext.logOut();
};

export const assessGivenSubscription = ({subscriptionId, token}) => {
    return Promise.all([
        getDisks(subscriptionId, token),
        getSnapshots(subscriptionId, token),
        getVMs(subscriptionId, token)
    ])
    .then(([disks, snapshots, vms]) => {
        const vmsById = _.keyBy(vms, vm => vm.id.toLowerCase());

        let monthsInYear = 12; //Stupid joke about using let instead of const because it might change sometime.
        const utilizationConstant = 0.65; //Magic number. Azure charges snapshots cost based on USED size. We only know
                                          //the ALLOCATED. We arbitrarily assume that this constant is the used ratio.
        const reportTime = Date.now();

        const nonOntapDisks = _.reject(disks, disk => isOntapDisk(disk, vmsById));

        const ontapDisks = _.difference(disks, nonOntapDisks);

        const ontapDisksById = _.keyBy(ontapDisks, disk => disk.id.toLowerCase());

        const totalCapacity = _.sumBy(nonOntapDisks, 'properties.diskSizeGB');

        const unattachedDisks = _.reject(nonOntapDisks, 'managedBy');

        const unattachedDisksAnnualCost = _.sum(_.map(unattachedDisks, getDiskPrice)) * monthsInYear;

        const unassociatedSnapshots = findUnassociatedSnapshots(disks, snapshots);

        const unassociatedSnapshotsAnnualCost = _.sum(_.map(unassociatedSnapshots, getSnapshotPrice)) * monthsInYear * utilizationConstant;

        const associatedSnapshots = _.difference(snapshots, unassociatedSnapshots);

        const nonOntapAssociatedSnapshots = _.reject(associatedSnapshots, snapshot => isOntapSnapshot(snapshot, ontapDisksById));

        const averageRPO = calculateAverageRpo(nonOntapAssociatedSnapshots, reportTime);

        const disksWithoutSnapshots = getDisksWithoutSnapshots(nonOntapDisks, snapshots);

        const disksWithSnapshotsCount = nonOntapDisks.length - disksWithoutSnapshots.length;

        const averageNumberOfSnapshots = disksWithSnapshotsCount ? _.round(nonOntapAssociatedSnapshots.length / disksWithSnapshotsCount, 2) : 0;
        const endTime = Date.now();

        return {
            reportTime,
            endTime,
            duration: endTime - reportTime,
            result: {
                totalCapacity,
                totalDisksIncludingOntap: disks.length,
                disksCount: nonOntapDisks.length,
                snapshotsCount: snapshots.length,
                associatedSnapshotsCount: associatedSnapshots.length,
                nonOntapAssociatedSnapshotsCount: nonOntapAssociatedSnapshots.length,
                unattachedDisks,
                unassociatedSnapshots,
                reportTime,
                protection: {
                    averageRPO,
                    averageNumberOfSnapshots,
                    disksWithoutSnapshotsCount: disksWithoutSnapshots.length,
                    disksWithSnapshotsCount,
                },
                efficiency: {
                    unattachedDisksCount: unattachedDisks.length,
                    unattachedDisksAnnualCost,
                    unassociatedSnapshotsCount: unassociatedSnapshots.length,
                    unassociatedSnapshotsAnnualCost,
                    totalCost: unattachedDisksAnnualCost + unassociatedSnapshotsAnnualCost
                }
            }
        };

    });
};
