import _ from 'lodash'
import { assessGivenRegions } from './cloudAssessor';

let mockAccountId = 'accountId123';
let mockVolumes = [];
let mockSnapshots = [];
let mockImages = [];

jest.mock('aws-sdk/clients/sts', () => {
    return class {
        getCallerIdentity() {
            return {
                promise: () => {
                    return new Promise((resolve) => {
                        resolve({Account: mockAccountId});
                    });
                }
            };
        }
    };
});

jest.mock('aws-sdk/clients/ec2', () => {
    return class {
        describeVolumes() {
            return {
                promise: () => {
                    return new Promise((resolve) => {
                        resolve({Volumes: mockVolumes});
                    });
                }
            };
        }
        describeSnapshots() {
            return {
                promise: () => {
                    return new Promise((resolve) => {
                        resolve({Snapshots: mockSnapshots});
                    });
                }
            };
        }
        describeImages() {
            return {
                promise: () => {
                    return new Promise((resolve) => {
                        resolve({Images: mockImages});
                    });
                }
            };
        }
    };
});

describe('assessGivenRegions', () => {
    beforeEach(() => {
        mockAccountId = 'accountId123';
        mockVolumes = [];
        mockSnapshots = [];
        mockImages = [];
        jest.spyOn(Date, 'now').mockImplementation(_.constant(1529510400000)); //Wednesday, 20 June 2018 19:00:00 GMT+03:00 DST
    });

    it('Sanity: 2 volumes, no snapshots', () => {
        mockVolumes = [{
            Attachments: [],
            VolumeType: 'gp2',
            Size: 1
        }, {
            Attachments: [],
            VolumeType: 'gp2',
            Size: 1
        }];
        return assessGivenRegions({regions: ['us-east-1']})
            .then((results) => {
                const report = results.allRegions[0].result;
                expect(report).toEqual(expect.objectContaining({
                        totalCapacity: 2,
                        totalVolumes: 2,
                        protection: {
                            averageRPO: -1,
                            averageNumberOfSnapshots: 0,
                            volumesWithoutSnapshotsCount: 2,
                            volumesWithSnapshotsCount: 0
                        },
                        efficiency: {
                            snapshotsWithoutVolumesCount: 0,
                            unattachedVolumesCount: 2,
                            unusedVolumesCost: 3,
                            unusedSnapshotsCost: 0,
                            totalCost: 3
                        }
                    }));
            });
    });

    it('Should work with previous generation magnetic disk type', () => {
        mockVolumes = [{
            Attachments: [],
            VolumeType: 'standard',
            Size: 100
        }];
        return assessGivenRegions({regions: ['us-east-1']})
            .then((results) => {
                const report = results.allRegions[0].result;
                expect(report).toEqual(expect.objectContaining({
                        totalCapacity: 100,
                        totalVolumes: 1,
                        protection: {
                            averageRPO: -1,
                            averageNumberOfSnapshots: 0,
                            volumesWithoutSnapshotsCount: 1,
                            volumesWithSnapshotsCount: 0
                        },
                        efficiency: {
                            snapshotsWithoutVolumesCount: 0,
                            unattachedVolumesCount: 1,
                            unusedVolumesCost: 60,
                            unusedSnapshotsCost: 0,
                            totalCost: 60
                        }
                    }));
            });
    });

    it('0 volumes, 1 unassociated snapshot', () => {
        mockSnapshots = [{SnapshotId: 'snap1', VolumeSize: 1, VolumeId: 'vol1'}];
        return assessGivenRegions({regions: ['us-east-1']})
            .then((results) => {
                expect(results.allRegions[0].result).toEqual(expect.objectContaining({
                        totalCapacity: 0,
                        totalVolumes: 0,
                        protection: {
                            averageRPO: -1,
                            averageNumberOfSnapshots: 0,
                            volumesWithoutSnapshotsCount: 0,
                            volumesWithSnapshotsCount: 0
                        },
                        efficiency: {
                            snapshotsWithoutVolumesCount: 1,
                            unattachedVolumesCount: 0,
                            unusedVolumesCost: 0,
                            unusedSnapshotsCost: 1,
                            totalCost: 1
                        }
                    }));
            });
    });

    it('0 volumes, 1000 unassociated independent snapshots', () => {
        mockSnapshots = _.map(_.range(1000), (num) => {
            return {SnapshotId: 'snapshot' + num, VolumeSize: 1, VolumeId: 'vol'+num};
        });
        return assessGivenRegions({regions: ['us-east-1']})
            .then((results) => {
                expect(results.allRegions[0].result).toEqual(expect.objectContaining({
                        efficiency: expect.objectContaining({
                            snapshotsWithoutVolumesCount: 1000,
                            unusedSnapshotsCost: 600,
                            totalCost: 600
                        })
                    }));
            });
    });

    it('0 volumes, 1000 unassociated snapshots of the same volume', () => {
        mockSnapshots = _.map(_.range(1000), (num) => {
            return {SnapshotId: 'snapshot' + num, VolumeSize: 1, VolumeId: 'vol1'};
        });
        return assessGivenRegions({regions: ['us-east-1']})
            .then((results) => {
                expect(results.allRegions[0].result).toEqual(expect.objectContaining({
                        efficiency: expect.objectContaining({
                            snapshotsWithoutVolumesCount: 1000,
                            unusedSnapshotsCost: 31,
                            totalCost: 31
                        })
                    }));
            });
    });

    it('should ignore ONTAP volumes entirely', () => {
        //One Ontap, one attached, one unattached
        mockVolumes = [{
            Attachments: [],
            VolumeType: 'gp2',
            Size: 1,
            Tags: [{Key: 'WorkingEnvironment'}]
        }, {
            Attachments: ['Instance1'],
            VolumeType: 'gp2',
            Size: 1,
        }, {
            Attachments: [],
            VolumeType: 'gp2',
            Size: 1,
        }];
        return assessGivenRegions({regions: ['us-east-1']})
            .then((results) => {
                expect(results.allRegions[0].result).toEqual(expect.objectContaining({
                    totalCapacity: 2,
                    totalVolumes: 2,
                    totalAmiSnapshots: 0,
                    totalImages: 0,
                    totalRelevantSnapshots: 0,
                    totalSnapshots: 0,
                    totalVolumesIncludingOntap: 3,
                    protection: {
                        averageRPO: -1,
                        averageNumberOfSnapshots: 0,
                        volumesWithoutSnapshotsCount: 2,
                        volumesWithSnapshotsCount: 0
                    },
                    efficiency: {
                        snapshotsWithoutVolumesCount: 0,
                        unattachedVolumesCount: 1,
                        unusedVolumesCost: 2,
                        unusedSnapshotsCost: 0,
                        totalCost: 2
                    }
                }));
            });
    });

    it('one io1 volume with iops', () => {
        mockVolumes = [{
            Attachments: [],
            VolumeType: 'io1',
            Size: 1,
            Iops: 1000
        }];
        return assessGivenRegions({regions: ['us-east-1']})
            .then((results) => {
                expect(results.allRegions[0].result).toEqual(expect.objectContaining({
                    totalCapacity: 1,
                    totalVolumes: 1,
                    protection: {
                        averageRPO: -1,
                        averageNumberOfSnapshots: 0,
                        volumesWithoutSnapshotsCount: 1,
                        volumesWithSnapshotsCount: 0
                    },
                    efficiency: {
                        snapshotsWithoutVolumesCount: 0,
                        unattachedVolumesCount: 1,
                        unusedVolumesCost: 782,
                        unusedSnapshotsCost: 0,
                        totalCost: 782
                    }
                }));
            });
    });

    it('ONTAP snapshots should be ignored by tag "WorkingEnvironment"', () => {
        //One ontap, one not
        mockVolumes = [{
            VolumeId: 'OntapVolume',
            Attachments: ['OntapInstance'],
            VolumeType: 'gp2',
            Size: 1,
            Tags: [{Key: 'WorkingEnvironment'}]
        }, {
            VolumeId: 'vol1',
            Attachments: ['Instance1'],
            VolumeType: 'gp2',
            Size: 1,
        }];

        //One ontap, one attached, one unattached
        mockSnapshots = [
            {SnapshotId: 'snap1', VolumeSize: 1, VolumeId: 'OntapVolume', Tags: [{Key: 'WorkingEnvironment'}]},
            {SnapshotId: 'snap2', VolumeSize: 1, VolumeId: 'vol1', StartTime: new Date("Wed Jun 20 2018 18:00:00 +0300")},
            {SnapshotId: 'snap3', VolumeSize: 1, VolumeId: 'missingVol', StartTime: new Date('Wed Jun 20 2018 18:00:00 +0300')}
        ];
        return assessGivenRegions({regions: ['us-east-1']})
            .then((results) => {
                expect(results.allRegions[0].result).toEqual(expect.objectContaining({
                    totalCapacity: 1,
                    totalVolumes: 1,
                    totalAmiSnapshots: 0,
                    totalImages: 0,
                    totalRelevantSnapshots: 2,
                    totalSnapshots: 3,
                    totalVolumesIncludingOntap: 2,
                    protection: {
                        averageRPO: 60 * 60 * 1000, //1 hour
                        averageNumberOfSnapshots: 1,
                        volumesWithoutSnapshotsCount: 0,
                        volumesWithSnapshotsCount: 1
                    },
                    efficiency: {
                        snapshotsWithoutVolumesCount: 1,
                        unattachedVolumesCount: 0,
                        unusedVolumesCost: 0,
                        unusedSnapshotsCost: 1,
                        totalCost: 1
                    }
                }));
            });
    });

    it('Average RPO calculations', () => {
        mockVolumes = [{
            VolumeId: 'vol1',
            Attachments: ['Instance1'],
            VolumeType: 'gp2',
            Size: 1,
        }, {
            VolumeId: 'vol2',
            Attachments: ['Instance2'],
            VolumeType: 'gp2',
            Size: 1,
        }, {
            VolumeId: 'vol3',
            Attachments: ['Instance3'],
            VolumeType: 'gp2',
            Size: 1,
        }];

        //vol1 has hourly snapshots, vol2 has snapshots every 3 hours, vol3 has no snapshots at all.
        //Plus there is unassociated snapshot that should not be taken into account.
        //Average should be 2 hours (1+3)/2.
        mockSnapshots = [
            {SnapshotId: 'snap1-1', VolumeSize: 1, VolumeId: 'vol1', StartTime: new Date("Wed Jun 20 2018 18:00:00 +0300")},
            {SnapshotId: 'snap1-2', VolumeSize: 1, VolumeId: 'vol1', StartTime: new Date("Wed Jun 20 2018 17:00:00 +0300")},
            {SnapshotId: 'snap1-3', VolumeSize: 1, VolumeId: 'vol1', StartTime: new Date('Wed Jun 20 2018 16:00:00 +0300')},

            {SnapshotId: 'snap2-1', VolumeSize: 1, VolumeId: 'vol2', StartTime: new Date('Wed Jun 20 2018 16:00:00 +0300')},
            {SnapshotId: 'snap2-2', VolumeSize: 1, VolumeId: 'vol2', StartTime: new Date('Wed Jun 20 2018 12:00:00 +0300')},
            {SnapshotId: 'snap2-3', VolumeSize: 1, VolumeId: 'vol2', StartTime: new Date('Wed Jun 20 2018 10:00:00 +0300')},
            {SnapshotId: 'snap2-4', VolumeSize: 1, VolumeId: 'vol2', StartTime: new Date('Wed Jun 20 2018 09:59:59 +0300')},
            {SnapshotId: 'snap2-5', VolumeSize: 1, VolumeId: 'vol2', StartTime: new Date('Wed Jun 20 2018 09:59:58 +0300')},

            {SnapshotId: 'snap3-1', VolumeSize: 1, VolumeId: 'missingVolume', StartTime: new Date('Wed Jun 20 2018 01:00:00 +0300')}
        ];
        return assessGivenRegions({regions: ['us-east-1']})
            .then((results) => {
                expect(results.allRegions[0].result).toEqual(expect.objectContaining({
                    totalCapacity: 3,
                    totalVolumes: 3,
                    totalSnapshots: 9,
                    protection: {
                        averageRPO: 60 * 60 * 1000 * 2, //2 hours
                        averageNumberOfSnapshots: 4,
                        volumesWithoutSnapshotsCount: 1,
                        volumesWithSnapshotsCount: 2
                    },
                    efficiency: {
                        snapshotsWithoutVolumesCount: 1,
                        unattachedVolumesCount: 0,
                        unusedVolumesCost: 0,
                        unusedSnapshotsCost: 1,
                        totalCost: 1
                    }
                }));
            });
    });

    it('should calculate price for snapshots copied from another region or account by the full size of each snapshot, without factoring diffs', () => {
        mockSnapshots = [
            {SnapshotId: 'snap1', VolumeSize: 100, VolumeId: 'vol-ffffffff', StartTime: new Date('Wed Jun 20 2018 18:00:00 +0300')},
            {SnapshotId: 'snap2', VolumeSize: 100, VolumeId: 'vol-ffffffff', StartTime: new Date('Wed Jun 19 2018 17:00:00 +0300')},
            {SnapshotId: 'snap3', VolumeSize: 100, VolumeId: 'vol-ffffffff', StartTime: new Date('Wed Jun 18 2018 16:00:00 +0300')},
        ];
        return assessGivenRegions({regions: ['us-east-1']})
            .then((results) => {
                expect(results.allRegions[0].result).toEqual(expect.objectContaining({
                    totalCapacity: 0,
                    totalVolumes: 0,
                    totalSnapshots: 3,
                    protection: {
                        averageRPO: -1,
                        averageNumberOfSnapshots: 0,
                        volumesWithoutSnapshotsCount: 0,
                        volumesWithSnapshotsCount: 0
                    },
                    efficiency: {
                        snapshotsWithoutVolumesCount: 3,
                        unattachedVolumesCount: 0,
                        unusedVolumesCost: 0,
                        unusedSnapshotsCost: 180, //0.05 * 100 * 3 * 12 = pricePerGb * VolumeSize * num of snapshots * months in year,
                        totalCost: 180
                    }
                }));
            });
    });

    it('should not include copied snapshots in RPO calculations', () => {
        mockVolumes = [{
            VolumeId: 'vol1',
            Attachments: ['Instance1'],
            VolumeType: 'gp2',
            Size: 1,
        }];

        mockSnapshots = [
            {SnapshotId: 'snap1-1', VolumeSize: 1, VolumeId: 'vol1', StartTime: new Date('Wed Jun 20 2018 18:00:00 +0300')},

            {SnapshotId: 'snap1', VolumeSize: 100, VolumeId: 'vol-ffffffff', StartTime: new Date('Wed Jun 20 2018 18:00:00 +0300')},
            {SnapshotId: 'snap2', VolumeSize: 100, VolumeId: 'vol-ffffffff', StartTime: new Date('Wed Jun 19 2018 17:00:00 +0300')},
            {SnapshotId: 'snap3', VolumeSize: 100, VolumeId: 'vol-ffffffff', StartTime: new Date('Wed Jun 18 2018 16:00:00 +0300')},
        ];
        return assessGivenRegions({regions: ['us-east-1']})
            .then((results) => {
                expect(results.allRegions[0].result).toEqual(expect.objectContaining({
                    totalCapacity: 1,
                    totalVolumes: 1,
                    totalSnapshots: 4,
                    protection: {
                        averageRPO: 60 * 60 * 1000,
                        averageNumberOfSnapshots: 1,
                        volumesWithoutSnapshotsCount: 0,
                        volumesWithSnapshotsCount: 1
                    },
                    efficiency: {
                        snapshotsWithoutVolumesCount: 3,
                        unattachedVolumesCount: 0,
                        unusedVolumesCost: 0,
                        unusedSnapshotsCost: 180, //0.05 * 100 * 3 * 12 = pricePerGb * VolumeSize * num of snapshots * months in year,
                        totalCost: 180
                    }
                }));
            });
    });
});
