import EC2 from 'aws-sdk/clients/ec2'
import STS from 'aws-sdk/clients/sts'
import { getEbsPrices } from './getEbsPrices';
import _ from 'lodash'

const nextSnapshotSizeFactor = 0.05; //Just made up, sounds reasonable. OCCM uses that too.
const monthsInYear = 12; //Not counting Adar Alef.

const getVolumeYearlyCost = ({volume, region}) => {
    const storageCost = getEbsPrices({region, diskType: volume.VolumeType}) * volume.Size;
    const iopsCost = volume.VolumeType === 'io1' ? (volume.Iops * getEbsPrices({region, diskType: volume.VolumeType, rate: 'perPIOPSreq'})) : 0;

    return (storageCost + iopsCost) * monthsInYear;
};

const getSnapshotYearlyCost = ({snapshots, region}) => {
    const volumeIds = _(snapshots).map('VolumeId').uniq().value();
    const sizePerVolumeId = _.map(volumeIds, (volumeId) => {
        const volumeSnapshots = _.filter(snapshots, {VolumeId: volumeId});

        return volumeSnapshots[0].VolumeSize * (1 + (volumeSnapshots.length - 1) * nextSnapshotSizeFactor)
    });

    return _.sum(sizePerVolumeId) * getEbsPrices({region, diskType: 'snapshot', rate: 'perGBmoDataStored'}) * monthsInYear;
};

const determineCosts = async ({notAttachedSnapshots, unattachedVolumes, region}) => {
    const unattachedVolumesCost = _.ceil(_.reduce(unattachedVolumes, (accumulator, volume) => accumulator + getVolumeYearlyCost({volume, region}), 0));

    const [copiedSnapshots, unassociatedSnapshots] = _.partition(notAttachedSnapshots, { VolumeId: 'vol-ffffffff' });
    const copiedSnapshotsCost =  _.sumBy(copiedSnapshots, 'VolumeSize') * getEbsPrices({region, diskType: 'snapshot', rate: 'perGBmoDataStored'}) * monthsInYear;
    const unassociatedSnapshotsCost = getSnapshotYearlyCost({snapshots: unassociatedSnapshots, region});
    const notConnectedSnapshotsCost = _.ceil(copiedSnapshotsCost + unassociatedSnapshotsCost);

    return Promise.resolve({unusedVolumesCost: unattachedVolumesCost, unusedSnapshotsCost: notConnectedSnapshotsCost, totalCost: unattachedVolumesCost + notConnectedSnapshotsCost});
};

const basePermissionString = "The provided credentials do not have the required permission:";

const wrapStep = async (promise, step, region) => {
    try {
        return await promise;
    } catch (e) {
        e.step = step;
        e.isPermissions = e.statusCode === 403;
        e.region = region;

        if(e.isPermissions) {
            if(step === "regions") e.message = `${basePermissionString} ec2:DescribeRegions`;
            else if(step === "volumes") e.message = `${basePermissionString} ec2:DescribeVolumes`;
            else if(step === "snapshots") e.message = `${basePermissionString} ec2:DescribeSnapshots`;
            else if(step === "images") e.message = `${basePermissionString} ec2:DescribeImages`;
        }

        if (e.message === "The security token included in the request is invalid." ||
            e.message === "The request signature we calculated does not match the signature you provided. Check your AWS Secret Access Key and signing method. Consult the service documentation for details." ||
            e.message.includes('not a valid key=value pair ')) {
            e.message = "The provided AWS credentials are invalid."
        }

        throw e;
    }
};

const rpoOfVolume = ({snapshots, reportTime}) => {
    const latestSnapshotTime = _(snapshots)
        .map(snapshot => snapshot.StartTime.valueOf())
        .sort()
        .last();

    return reportTime - latestSnapshotTime;
};

const snapshotsUsedByAMI = ({images}) => {
    return _(images).map('BlockDeviceMappings').flatMap().map('Ebs').map('SnapshotId').uniq().value()
};

const retrieveVolumes = async ({ec2, region}) => {
    const {Volumes: allVolumes} = await wrapStep(ec2.describeVolumes().promise(), 'volumes', region);
    const nonOntapVolumes = _.reject(allVolumes, (volume) => {
        return _.find(volume.Tags, {Key: 'WorkingEnvironment'});
    });

    return {totalVolumesIncludingOntap: allVolumes.length, volumes: nonOntapVolumes};
};

const retrieveSnapshots = async ({ec2, accountId, region}) => {
    const describeImagesPromise = wrapStep(ec2.describeImages({Owners: [accountId]}).promise(), 'images', region);
    const describeSnapshotsPromise = wrapStep(ec2.describeSnapshots({OwnerIds: [accountId]}).promise(), 'snapshots', region);
    const {Images: images} = await describeImagesPromise;
    const {Snapshots: snapshots} = await describeSnapshotsPromise;

    const amiSnapshots = snapshotsUsedByAMI({images});

    return {
        totalImages: images.length,
        totalSnapshots: snapshots.length,
        totalAmiSnapshots: amiSnapshots.length,
        snapshots: _.reject(snapshots, (snapshot) => _.includes(amiSnapshots, snapshot.SnapshotId) || _.find(snapshot.Tags, {Key: 'WorkingEnvironment'}))
    };
};

const calculateAverageRpo = (associatedSnapshots, reportTime) => {
    if (associatedSnapshots.length === 0) {
        return -1;
    } else {
        return _(associatedSnapshots)
            .groupBy('VolumeId')
            .map(snapshots => rpoOfVolume({snapshots, reportTime}))
            .mean()
    }
};

/*
 returns {
 result: {}, or //success
 error: {} or //failure
 isEmpty: true //no volumes/snapshots in region
 }
 */
const assessByRegion = async ({credentials, region, accountId, reportTime}) => {
    const ec2 = new EC2({
        region,
        credentials
    });
    const retrieveVolumesPromise = retrieveVolumes({ec2, region});
    const retrieveSnapshotsPromise = retrieveSnapshots({ec2, accountId, region});

    const {totalVolumesIncludingOntap, volumes} = await retrieveVolumesPromise;
    const {
        totalImages,
        totalSnapshots,
        totalAmiSnapshots,
        snapshots
    } = await retrieveSnapshotsPromise;
    const totalRelevantSnapshots = snapshots.length;

    const totalVolumes = volumes.length;

    const snapshotsByVolumeId = _.groupBy(snapshots, "VolumeId");
    const volumesById = _.keyBy(volumes, "VolumeId");

    const totalCapacity = _(volumes).map("Size").sum();

    const [associatedSnapshots, notAttachedSnapshots] = _.partition(snapshots, snapshot => volumesById[snapshot.VolumeId]);

    const snapshotsWithoutVolumesCount = notAttachedSnapshots.length;
    const snapshotsWithVolumesCount = snapshots.length - notAttachedSnapshots.length;

    let averageRPO  = calculateAverageRpo(associatedSnapshots, reportTime);

    const volumesWithoutSnapshots = volumes.filter(volume => !snapshotsByVolumeId[volume.VolumeId]);
    const volumesWithoutSnapshotsCount = volumesWithoutSnapshots.length;

    const volumesWithSnapshotsCount = volumes.length - volumesWithoutSnapshots.length;

    const averageNumberOfSnapshots = volumesWithSnapshotsCount ? _.round(snapshotsWithVolumesCount / volumesWithSnapshotsCount, 2) : 0;

    const unattachedVolumes = volumes.filter(volume => volume.Attachments.length === 0);
    const unattachedVolumesCount = unattachedVolumes.length;

    const {unusedVolumesCost, unusedSnapshotsCost, totalCost} = await wrapStep(determineCosts({notAttachedSnapshots, unattachedVolumes, region, credentials}), "costs", region);

    return {
        region,
        result: {
            totalCapacity,
            totalVolumesIncludingOntap,
            totalVolumes,
            totalImages,
            totalSnapshots,
            totalAmiSnapshots,
            totalRelevantSnapshots,
            unattachedVolumes,
            notAttachedSnapshots,
            reportTime,
            protection: {
                averageRPO,
                averageNumberOfSnapshots,
                volumesWithoutSnapshotsCount,
                volumesWithSnapshotsCount
            },
            efficiency: {
                snapshotsWithoutVolumesCount,
                unattachedVolumesCount,
                unusedVolumesCost,
                unusedSnapshotsCost,
                totalCost
            }
        }
    }
};

export const assessGivenRegions = async ({credentials, regions}) => {
    const reportTime = Date.now();
    const sts = new STS({credentials, region: regions[0]});
    const {Account: accountId} = await wrapStep(sts.getCallerIdentity().promise(), "determine account");

    const results = await Promise.all(regions.map(region => {
        return assessByRegion({credentials, region, accountId, reportTime}).catch(error => ({
            error,
            region
        }));
    }));

    const successfulResults = results.filter(result => result.result);
    const failedResults = results.filter(result => result.error);

    if (failedResults.length !== 0) {
        const topError = failedResults[0].error;
        return Promise.reject(topError);
    } else {
        const endTime = Date.now();
        return {allRegions: successfulResults, reportTime, endTime, duration: endTime - reportTime};
    }
};
