import _ from 'lodash';
import pricing from './azurePricing.json';

const regionsMap = {
    australiaeast: 'australia-east',
    australiasoutheast: 'australia-southeast',
    brazilsouth: 'brazil-south',
    canadacentral: 'canada-central',
    canadaeast: 'canada-east',
    centralindia: 'central-india',
    centralus: 'us-central',
    eastasia: 'asia-pacific-east',
    eastus: 'us-east',
    eastus2: 'us-east-2',
    francecentral: 'france-central',
    germanycentral: 'germany-central',
    germanynortheast: 'germany-northeast',
    japaneast: 'japan-east',
    japanwest: 'japan-west',
    koreacentral: 'korea-central',
    northcentralus: 'us-north-central',
    southafricanorth: 'south-africa-north',
    southafricawest: 'south-africa-west',
    northeurope: 'europe-north',
    southcentralus: 'us-south-central',
    southeastasia: 'asia-pacific-southeast',
    uksouth: 'united-kingdom-south',
    ukwest: 'united-kingdom-west',
    usgovarizona: 'usgov-arizona',
    usgovtexas: 'usgov-texas',
    usgovvirginia: 'usgov-virginia',
    westcentralus: 'us-west-central',
    westeurope: 'europe-west',
    westus: 'us-west',
    westus2: 'us-west-2'
};

const diskSizeToDiskName = [
    { name: '4', sizeGb: '32' },
    { name: '6', sizeGb: '64' },
    { name: '10', sizeGb: '128' },
    { name: '15', sizeGb: '256' },
    { name: '20', sizeGb: '512' },
    { name: '30', sizeGb: '1024' },
    { name: '40', sizeGb: '2048' },
    { name: '50', sizeGb: '4096' },
    { name: '60', sizeGb: '8192' },
    { name: '70', sizeGb: '16384' },
    { name: '80', sizeGb: '32768' }
];

const getDiskGroup = (diskSize, tier) => {
    let diskGroup = _.find(diskSizeToDiskName, group => group.sizeGb >= diskSize).name;

    // Standard_SSD disks start from group '10' (128 GiB):
    if (tier === 'standardssd-e' && (diskGroup === '4' || diskGroup === '6')) diskGroup = '10';

    return diskGroup;
};

const getDiskPrice = disk => {
    const pricesObject = pricing;

    if (!disk.sku) {
        console.error('Error, disk with no SKU. Calculating as 0: ', disk);
        return 0;
    }
    let tier;
    switch (disk.sku.name) {
        case 'Premium_LRS':
            tier = 'premiumssd-p';
            break;
        case 'Standard_LRS':
            tier = 'standardhdd-s';
            break;
        case 'StandardSSD_LRS':
            tier = 'standardssd-e';
            break;
        default:
            console.error('Error, unknown disk type. Calculating as 0: ', disk);
            return 0;
    }

    const region = regionsMap[disk.location];
    if (!region) {
        console.log('Unknown region: ', disk.location);
        return 0;
    }
    const diskSize = disk.properties.diskSizeGB;
    const diskGroup = getDiskGroup(diskSize, tier);
    const offer = pricesObject.offers[tier + diskGroup];

    if (!offer) {
        console.log(`Unknown offer. tier: ${tier}, diskGroup: ${diskGroup}, disk: `, disk);
        return 0;
    }

    return offer.prices[region].value;
};

const getSnapshotPrice = snapshot => {
    const pricesObject = pricing;

    if (!snapshot.sku) {
        console.error('Error, snapshot with no SKU. Calculating as 0: ', snapshot);
        return 0;
    }
    let tier;
    switch (snapshot.sku.name) {
        case 'Premium_LRS':
            tier = 'premiumssd-snapshot';
            break;
        case 'Standard_LRS':
            tier = 'standardhdd-snapshot-lrs';
            break;
        case 'StandardSSD_LRS':
            tier = 'standardssd-snapshot';
            break;
        default:
            console.error('Error, unknown snapshot type. Calculating as 0: ', snapshot);
            return 0;
    }

    const region = regionsMap[snapshot.location];
    if (!region) {
        console.log('Unknown region: ', snapshot.location);
        return 0;
    }
    const snapshotSize = snapshot.properties.diskSizeGB;

    const offer = pricesObject.offers[tier];

    if (!offer) {
        console.log(`Unknown offer. tier: ${tier}, snapshot: `, snapshot);
        return 0;
    }

    return offer.prices[region].value * snapshotSize;
};

export { getDiskPrice, getSnapshotPrice };
