import React from 'react';
import connect from "react-redux/es/connect/connect";
import _ from "lodash";
import { STATUS_SUCCESS } from "../../actions";
import Loader from "../../Loader";
import { AgGridReact } from 'ag-grid-react';

import 'ag-grid-community/dist/styles/ag-grid.css';
import 'ag-grid-community/dist/styles/ag-theme-balham.css';
import styles from "./UserManagement.module.scss";
import layout from "ui-components/src/layout/layout.module.scss";
import "./UserManagement.scss";
import buttons from "ui-components/src/button/button.module.scss";

import classNames from "classnames";
import { ReactComponent as DotDotDotIcon } from "ui-components/src/images/monochrome-icons/ic-more.svg";
import ClickAway from "./../functional/ClickAway";
import { parentHasClass } from "../../utils/utils";
import { FormFieldInput } from "ui-components"
import { ReactComponent as InlineLoader } from "ui-components/src/images/loaders/inline-loader.svg";

const getColumnDefs = () => {
    const fields = ["User Email", "User Name", "User ID", "Has Blocks"];
    const fieldsName = {
        'User Email': "email",
        'User Name': "name",
        'User ID': "user_id",
        'Has Blocks': "hasBlocks"
    };
    const initialColumnsWidth = {
        'User Email': 280,
        'User Name': 280,
        'User ID': 280,
        'Has Blocks': 150
    };
    const columnDefs = fields.map(field => {
        return ({
            headerName: field,
            field: fieldsName[field],
            width: initialColumnsWidth[field],
            suppressMenu: true,
            unSortIcon: true,
            sortable: true
        })
    });
    columnDefs.push({
        headerName: "",
        field: "menu",
        cellRenderer: "menuCellRenderer",
        colId: 'button',
        width: 150
    });
    return columnDefs;
};

const getLogsColumnDefs = () => {
    const fields = ["Event", "When", "App", "Connection", "IP Address"];
    const fieldsName = {
        'Event': "displayType",
        'When': "date",
        'App': "client_name",
        'Connection': "session_connection",
        'IP Address': "ip"
    };
    return fields.map(field => {
        return ({
            headerName: field,
            field: fieldsName[field],
            suppressMenu: true,
            unSortIcon: true,
            sortable: true
        })
    });
};

const rowHeight = 55; // px

const MenuCellRenderer = props => {
    return (
        <div key={props.data.email} className="menu-cell">
            <button className="open-menu-button"
                    onClick={() => {
                        props.context.openUserMenu({ user: props.data, rowIndex: props.rowIndex })
                    }}>
                <DotDotDotIcon/>
            </button>
        </div>
    );
};

const UserPopUpMenuMenu = ({user, rowIndex, closeUserMenu, unblockUser, generatePasswordChangeTicket, getUserLogs, resetUserMfa}) => {
    return (
        <div className={classNames(styles["user-popup-menu"], styles["popup"])}
             style={{top: 10 + (rowHeight * (rowIndex + 1))}}>
            <ClickAway listener={event => {
                if (!parentHasClass(event.target, styles["popup"])) {
                    closeUserMenu()
                }
            }}/>
            <div className={styles["user-popup-menu-item"]}>
                <button title={!user.hasBlocks && "No blocks detected for the user"} disabled={!user.hasBlocks} onClick={() => unblockUser(user.user_id, user.email)}>Unblock
                </button>
            </div>
            <div className={styles["user-popup-menu-item"]}>
                <button onClick={() => generatePasswordChangeTicket(user.user_id)}>Generate Reset Password Link</button>
            </div>
            <div className={styles["user-popup-menu-item"]}>
                <button onClick={() => getUserLogs(user.user_id, 0)}>get User Logs</button>
            </div>
            <div className={styles["user-popup-menu-item"]}>
                <button onClick={() => resetUserMfa(user.user_id)}>Reset MFA</button>
            </div>
        </div>
    );
};

const mapStateToProps = state => {
    return {
        managementAxiosStatus: state.management.managementAxiosStatus,
        searchInputPending: state.management.searchInputPending,
        loading: state.management.loading,
        users: state.management.users,
        userLogs: state.management.userLogs,
        successMessage: state.management.successMessage,
        errorMessage: state.management.errorMessage,
        searchBarErrorMessage: state.management.searchBarErrorMessage,
        currentUserMenu: state.management.currentUserMenu
    };
};

const mapDispatchToProps = dispatch => {
    return {
        fetchManagementAxios: _.once(() => {
            dispatch({
                type: "MANAGEMENT:FETCH-MANAGEMENT-TOKEN"
            })
        }),
        searchUserByEmail: (userEmail) => {
            dispatch({
                type: "MANAGEMENT:SEARCH-USERS-BY-EMAIL",
                payload: {userEmail}
            })
        },
        unblockUser: (userId, userEmail) => {
            dispatch({
                type: "MANAGEMENT:UNBLOCK-USER",
                payload: {userId, userEmail}
            })
        },
        generatePasswordChangeTicket: (userId) => {
            dispatch({
                type: "MANAGEMENT:GENERATE-PASSWORD-CHANGE-TICKET",
                payload: {userId}
            })
        },
        getUserLogs: (userId, page) => {
            dispatch({
                type: "MANAGEMENT:GET-USER-LOGS",
                payload: {userId, page}
            })
        },
        resetUserMfa: (userId) => {
            dispatch({
                type: "MANAGEMENT:RESET-USER-MFA",
                payload: { userId }
            })
        },
        openUserMenu: ({ user, rowIndex }) => {
            dispatch({
                type: "MANAGEMENT:OPEN-USER-MENU",
                payload: { user, rowIndex }
            })
        },
        closeUserMenu: () => {
            dispatch({
                type: "MANAGEMENT:CLOSE-USER-MENU"
            })
        },
    }
};

class UserManagement extends React.Component {
    constructor(props) {
        super(props);

        this.state = {userEmail: ''};
        props.fetchManagementAxios();
        this.columnDefs = getColumnDefs();
        this.logsColumnDefs = getLogsColumnDefs();
        this.handleEmailChange = this.handleEmailChange.bind(this);
    }

    handleEmailChange(event) {
        this.setState({[event.target.name]: event.target.value});
    }

    onFirstDataRendered(params) {
        this.api = params.api;
        params.api.sizeColumnsToFit();
    }

    onComponentStateChanged() {
        this.api.refreshCells({force: true});
    }

    render() {
        const { managementAxiosStatus, loading, users, userLogs, successMessage, searchBarErrorMessage, errorMessage, currentUserMenu } = this.props;

        return (
            <div className={classNames(styles["main-container"], layout["main-content-container"])}>

                {(managementAxiosStatus !== STATUS_SUCCESS|| loading) && <Loader/>}

                <h3>User management</h3>
                {managementAxiosStatus === STATUS_SUCCESS &&
                <form className="search-user-by-email-form">
                    <div style={{display: "flex", alignItems: "center"}}>
                        <FormFieldInput type="email" name="userEmail" label="User Email" value={this.state.userEmail}
                                        className={styles["input-container"]}
                                        onChange={this.handleEmailChange} placeholder="user_name@gmail.com"
                                        autoFocus={true} errorMessage={searchBarErrorMessage}
                                        pending={this.props.searchInputPending}
                                        InlineLoaderIcon={InlineLoader}/>
                        <button className={styles.primary} onClick={event => {
                            this.state.userEmail && this.props.searchUserByEmail(this.state.userEmail);
                            event.preventDefault();
                        }}>Search
                        </button>
                    </div>
                </form>}
                {(users && (users.length || "")) &&
                <div className={styles["table-wrapper"]}>
                    {currentUserMenu && !loading &&
                    <UserPopUpMenuMenu user={currentUserMenu.user}
                                       rowIndex={currentUserMenu.rowIndex}
                                       closeUserMenu={this.props.closeUserMenu}
                                       unblockUser={this.props.unblockUser}
                                       generatePasswordChangeTicket={this.props.generatePasswordChangeTicket}
                                       getUserLogs={this.props.getUserLogs}
                                       resetUserMfa={this.props.resetUserMfa}/>
                    }
                    <div className={classNames("portal-ag-grid", "ag-theme-balham")}>
                        <AgGridReact
                            columnDefs={this.columnDefs}
                            rowData={users}
                            enableColResize={false}
                            onFirstDataRendered={this.onFirstDataRendered.bind(this)}
                            onComponentStateChanged={this.onComponentStateChanged.bind(this)}
                            suppressDragLeaveHidesColumns={true}
                            rowHeight={rowHeight}
                            headerHeight={rowHeight}
                            suppressColumnVirtualisation="true"
                            frameworkComponents={{
                                menuCellRenderer: MenuCellRenderer
                            }}
                            context={{
                                unblockUser: this.props.unblockUser,
                                generatePasswordChangeTicket: this.props.generatePasswordChangeTicket,
                                getUserLogs: this.props.getUserLogs,
                                openUserMenu: this.props.openUserMenu
                            }}
                        >
                        </AgGridReact>
                    </div>
                </div>}
                {errorMessage && <div className={styles["action-error"]}>{errorMessage}</div>}
                {successMessage && <div className={styles["action-success"]}>{successMessage}</div>}
                {userLogs &&
                <div>
                    <h4>User logs</h4>
                    <div className={styles["logs-table-navbar"]}>
                        {userLogs.start > 0 && <button className={buttons.primary}
                            onClick={() => this.props.getUserLogs(userLogs.userId, userLogs.page - 1)}
                            style={{display: "inline", marginRight: "14px", color: 'black' }}>Previous page</button>}
                        {userLogs.start + userLogs.length < userLogs.total && <button className={buttons.primary}
                            onClick={() => this.props.getUserLogs(userLogs.userId, userLogs.page + 1)}
                            style={{display: "inline", marginRight: "14px", color: 'black' }}>Next page</button>}
                        <span>Displaying {userLogs.length} log entries ({userLogs.start + 1} to {userLogs.start + userLogs.length}) out of {userLogs.total} for user {userLogs.userId}</span>
                    </div>
                    <div className={classNames("logs-ag-grid", "ag-theme-balham")}>
                        <AgGridReact
                            style={{ height: "600px" }}
                            columnDefs={this.logsColumnDefs}
                            rowData={userLogs.logs}
                            enableColResize={false}
                            onFirstDataRendered={this.onFirstDataRendered.bind(this)}
                            onComponentStateChanged={this.onComponentStateChanged.bind(this)}
                            suppressDragLeaveHidesColumns={true}
                            rowHeight={rowHeight}
                            headerHeight={rowHeight}
                            suppressColumnVirtualisation="true"
                        >
                        </AgGridReact>
                    </div>
                </div>}
            </div>
        );
    }
}

export default connect(mapStateToProps, mapDispatchToProps)(UserManagement);
