import React from 'react'
import "./CloudManagersView.scss"
import { ReactComponent as CloseIcon } from "ui-components/src/images/monochrome-icons/x-circle-icon.svg"
import { ReactComponent as EditIcon } from "ui-components/src/images/monochrome-icons/edit-pencil-icon.svg"
import { ReactComponent as GotoIcon } from "ui-components/src/images/monochrome-icons/arrow-circle-icon.svg"
import className from "classnames";
import { Dialog } from "./../dialog/Dialog"
import moment from 'moment';
import { WarningIcon } from 'components/svg-icons/Icons';
import { connect } from 'react-redux'
import * as actions from "./../../actions";
import RemoveOccmDialog from "./RemoveOccmDialog";
import UpdateOccmIps from "./UpdateOccmIps";
import { ReactComponent as ONTAPCloudLogo } from "ui-components/src/images/netapp-services/ic-cvo.svg"
import ExternalLink from './../widgets/ExternalLink'
import { ReactComponent as InlineLoader } from 'ui-components/src/images/loaders/inline-loader.svg'
import loaders from 'ui-components/src/loaders/loaders.module.scss';
import Popover from "../widgets/Popover";
import popoverStyle from "ui-components/src/Popover/popover.module.scss";

const isOlderThanThirtyMinutes = (date) => moment().subtract(30, 'minutes').valueOf() > date;

const serviceTracker = ({action}) =>
    (service) => window.ga('send', {
        hitType: 'event',
        eventCategory: 'LaunchPad',
        eventAction: action,
        eventLabel: service
    });

const track = serviceTracker({action: "ClickService"});

const CloudManagersView = props => {
    const editUrisRef = React.createRef();
    const removeOccmRef = React.createRef();

    return (<div className="cloud-managers-view">
        <div className="component-container">
            <div className="component-header">
                <h3 className="left">Cloud Manager Systems</h3>
                <ExternalLink linkKey="launchPad.services.occmSetupService" className="button large primary right">Create a New Cloud Manager</ExternalLink>
            </div>
            {props.occms.status === actions.STATUS_SUCCESS && props.occms.data.occms.length > 0 && <div className="component-contents">
                {props.occms.data.occms.map(occm => (
                    <div key={occm.occm} className={className("occm-row", occm.status, {solitary: props.occms.data.occms.length === 1})}>
                        <ExternalLink linkKey="launchPad.services.cloudManager" href={occm.primaryCallbackUri} className="occm-info" onClick={() => {track("cloud-manager")}}>
                            <div className="logo">
                                <ONTAPCloudLogo/>
                            </div>
                            <div className="info">
                                <div className="title text-ellipsis">Name:</div>
                                <div className="text text-ellipsis" title={occm.occmName}>{occm.occmName}</div>
                            </div>
                            {occm.status === "ready" &&
                                <div className="info" style={{marginLeft: "20px", marginRight: "5px"}}>
                                    <div className="title text-ellipsis">Address:</div>
                                    <div className="text text-ellipsis">{occm.primaryCallbackUri}</div>
                                </div>
                            }
                            {occm.status === "ready" &&
                                <div className="goto">
                                    <GotoIcon style={{position: "relative", top: 7}}/>
                                    Go to Cloud Manager
                                </div>
                            }
                            {
                                occm.status === "initializing" &&
                                (isOlderThanThirtyMinutes(occm.createDate) ?
                                        <Popover trigger="mouseover" containerStyle={{cursor: "pointer", display: "block", width: "212px", float: "right", transform: "translateY(31px)"}} className={popoverStyle['info']} component={
                                            <div className="error-text warning">
                                                <WarningIcon style={{position: "relative", width: 15, color: "#ffa100", marginRight: "8px"}} />
                                                <span style={{color: "#ffa100"}}>Launch is hanging...</span>
                                            </div>
                                        }>
                                            There may be a problem deploying this Cloud Manager system. Go to your cloud provider’s console to find the IP address for the system and then log in to complete the setup process. If you need help, contact us by clicking the chat icon in the lower right of your web browser.
                                        </Popover>
                                        :
                                        <div className="launching-text">
                                            <InlineLoader className={loaders["inline-loader"]} style={{position: "relative", top: 3, marginRight: 0}}/>
                                            <span>Launching Cloud Manager</span>
                                        </div>
                                )
                            }
                            {
                                occm.status === "failed" &&
                                <div className="failed-text">
                                    Cloud Manager failed to launch
                                </div>
                            }
                        </ExternalLink>
                        <div className="occm-menu">
                            <div className="dots">
                                <div className="d1"/>
                                <div className="d2"/>
                                <div className="d3"/>
                            </div>
                            <div className="menu-children">
                                <button disabled={occm.status !== "ready"} className="menu-child" ref={editUrisRef} onClick={() => props.openUrisDialog(occm.occm)}>
                                    <EditIcon style={{width: 13.5, height: 13.5, top: 2}}/> Edit URIs
                                </button>
                                <button className="menu-child" ref={removeOccmRef} onClick={() => props.openRemoveDialog(occm.occm)}>
                                    <CloseIcon style={{width: 15, height: 15, top: 3}}/> Remove Listing
                                </button>
                                <Dialog target={() => editUrisRef.current} open={props.occmDialogState === `EDIT-URIS-${occm.occm}`} closeTrigger={props.closeDialogs}>
                                    <UpdateOccmIps occm={occm} className="update-occm-uris-dialog"/>
                                </Dialog>
                                <Dialog target={() => removeOccmRef.current} open={props.occmDialogState === `REMOVE-OCCM-${occm.occm}`} closeTrigger={props.closeDialogs}>
                                    <RemoveOccmDialog occm={occm} className="remove-occm-dialog"/>
                                </Dialog>
                            </div>
                        </div>
                    </div>
                ))}
            </div>}
        </div>
    </div>
)};

const mapStateToProps = state => {
    return {
        occms: state.services.cloudVolumesOntap,
        occmDialogState: state.services.occmDialogState
    };
};

const mapDispatchToProps = dispatch => {
    return {
        openUrisDialog: (occmId) => {
            dispatch({
                type: "LAUNCHPAD:SET-OCCM-DIALOG-STATE",
                payload: `EDIT-URIS-${occmId}`
            })
        },
        openRemoveDialog: (occmId) => {
            dispatch({
                type: "LAUNCHPAD:SET-OCCM-DIALOG-STATE",
                payload: `REMOVE-OCCM-${occmId}`
            })
        },
        closeDialogs: () => {
            dispatch({
                type: "LAUNCHPAD:SET-OCCM-DIALOG-STATE",
                payload: null
            })
        }
    }
};

export default connect(mapStateToProps, mapDispatchToProps)(CloudManagersView)