import axios from "axios";
import _ from "lodash";
import { history } from "../../store";
import moment from "moment"
import eventAcronymMapping from "./auth0LogEventAcronymMapping"
import { reflect } from './../../utils/utils'

let managementAxios;

export default ({ getState, dispatch, next, action }) => {
    const { type, payload } = action;
    if (type === "MANAGEMENT:FETCH-MANAGEMENT-TOKEN") {
        if (!managementAxios) {
            window.auth.refreshSso({ audience: "https://governance.cloud.netapp.com", scope: "cc:support" })
                .then(({ accessToken, scope }) => {
                    if (!_.includes(scope, 'cc:support')) {
                        history.replace('/');
                        dispatch({ type: "MANAGEMENT:FETCH-MANAGEMENT-TOKEN-UNAUTHORIZED" });
                    } else {
                        managementAxios = axios.create({
                            baseURL: window.__react_app_portal_backend_base_url,
                            headers: { Authorization: `Bearer ${accessToken}` }
                        });
                    }
                    dispatch({ type: "MANAGEMENT:FETCH-MANAGEMENT-TOKEN-SUCCESS" });
                }, (error) => {
                    dispatch({ type: "MANAGEMENT:FETCH-MANAGEMENT-TOKEN-FAILED", payload: error.response.data.message })
                })
        } else {
            dispatch({ type: "MANAGEMENT:FETCH-MANAGEMENT-TOKEN-SUCCESS" });
        }
    } else if (type === "MANAGEMENT:FETCH-CONNECTIONS") {
        managementAxios.get('/governance/connections?fields=name,id,metadata,options')
            .then(({ data: connections }) => {
                dispatch({ type: "MANAGEMENT:FETCH-CONNECTIONS-SUCCESS", payload: { connections } });
            }, (error) => {
                dispatch({ type: "MANAGEMENT:FETCH-CONNECTIONS-FAILED", payload: error.response.data.message })
            })
    } else if (type === "MANAGEMENT:ENABLE-CONNECTION") {
        const { connectionName, companyName, emailDomains } = payload;

        managementAxios.post('/governance/enable-connection', { connectionName, companyName, emailDomains })
            .then(({ data }) => {
                    dispatch({ type: "MANAGEMENT:ENABLE-CONNECTION-SUCCESS", payload: data });
                    dispatch({ type: "MANAGEMENT:FETCH-CONNECTIONS" });
                },
                (error) => {
                    dispatch({ type: "MANAGEMENT:ENABLE-CONNECTION-ERROR", payload: error.response.data.message });
                });
    } else if (type === "MANAGEMENT:DISABLE-CONNECTION") {
        const { connectionName } = payload;

        managementAxios.post('/governance/disable-connection', { connectionName })
            .then(({ data }) => {
                    dispatch({ type: "MANAGEMENT:DISABLE-CONNECTION-SUCCESS", payload: data });
                    dispatch({ type: "MANAGEMENT:FETCH-CONNECTIONS" });
                },
                (error) => {
                    dispatch({ type: "MANAGEMENT:DISABLE-CONNECTION-ERROR", payload: error.response.data.message });
                });
    } else if (type === "MANAGEMENT:SEARCH-USERS-BY-EMAIL") {
        const { userEmail } = payload;
        dispatch({ type: "MANAGEMENT:SEARCH-USERS-BY-EMAIL-PENDING" });

        managementAxios.get('/governance/support/users', {
            params: {
                fields: 'name,email,user_id',
                q: `email:${userEmail}`
            }
        })
            .then(({ data: users }) => {
                    return Promise.all(users.map(user => reflect(managementAxios.get(`/governance/support/users/${user.user_id}/blocks`))))
                        .then(blocksResponses => {
                            let firstGetBlocksError = null;

                            _.forEach(users, (user, index) => {
                                if (blocksResponses[index].status === 'fulfilled') {
                                    user.hasBlocks = !_.isEmpty(blocksResponses[index].value.data)
                                } else {
                                    user.hasBlocks = 'Error';
                                    firstGetBlocksError = firstGetBlocksError || `Could not retrieve blocks info for user ${user.user_id}: ${blocksResponses[index].error.response.data.message}`;
                                }
                            });

                            dispatch({
                                type: "MANAGEMENT:SEARCH-USERS-BY-EMAIL-SUCCESS",
                                payload: { users, userEmail, firstGetBlocksError }
                            });
                        })
                }, error => {
                    dispatch({
                        type: "MANAGEMENT:SEARCH-USERS-BY-EMAIL-ERROR",
                        payload: { message: error.response.data.message }
                    });
                }
            );
    } else if (type === "MANAGEMENT:UNBLOCK-USER") {
        const { userId, userEmail } = payload;
        dispatch({ type: "MANAGEMENT:UNBLOCK-USER-PENDING" });
        managementAxios.delete(`/governance/support/users/${userId}/blocks`)
            .then(() => {
                    dispatch({ type: "MANAGEMENT:SEARCH-USERS-BY-EMAIL", payload: { userEmail } });
                },
                error => {
                    dispatch({
                        type: "MANAGEMENT:UNBLOCK-USER-ERROR",
                        payload: { message: error.response.data.message }
                    });
                });
    } else if (type === "MANAGEMENT:GENERATE-PASSWORD-CHANGE-TICKET") {
        const { userId } = payload;
        dispatch({ type: "MANAGEMENT:GENERATE-PASSWORD-CHANGE-TICKET-PENDING" });
        managementAxios.post(`/governance/support/users/${userId}/password-change-ticket`)
            .then(({ data }) => {
                    dispatch({
                        type: "MANAGEMENT:GENERATE-PASSWORD-CHANGE-TICKET-SUCCESS",
                        payload: { resetPasswordUrl: data.resetPasswordUrl }
                    });
                },
                error => {
                    dispatch({
                        type: "MANAGEMENT:GENERATE-PASSWORD-CHANGE-TICKET-ERROR",
                        payload: { message: error.response.data.message }
                    });
                });
    } else if (type === "MANAGEMENT:GET-USER-LOGS") {
        const { userId, page } = payload;
        dispatch({ type: "MANAGEMENT:GET-USER-LOGS-PENDING" });
        managementAxios.get(`/governance/support/users/${userId}/logs`, { params: { page, perPage: 100 } })
            .then(({ data }) => {
                    const logsForDisplay = {
                        ...data,
                        page,
                        userId,
                        logs: _.map(data.logs, logEntry => ({
                            ...logEntry,
                            displayType: eventAcronymMapping[logEntry.type],
                            relativeDate: moment(logEntry.date).fromNow()
                        }))
                    };
                    dispatch({ type: "MANAGEMENT:GET-USER-LOGS-SUCCESS", payload: { logs: logsForDisplay } });
                },
                error => {
                    dispatch({
                        type: "MANAGEMENT:GET-USER-LOGS-ERROR",
                        payload: { message: error.response.data.message }
                    });
                });
    } else if (type === "MANAGEMENT:RESET-USER-MFA") {
        const { userId } = payload;
        dispatch({ type: "MANAGEMENT:RESET-USER-MFA-PENDING" });
        managementAxios.post(`/governance/support/users/${userId}/reset-mfa`)
            .then(() => {
                    dispatch({ type: "MANAGEMENT:RESET-USER-MFA-SUCCESS" });
                },
                error => {
                    dispatch({
                        type: "MANAGEMENT:RESET-USER-MFA-ERROR",
                        payload: { message: error.response.data.message }
                    });
                });
    }

    return next(action);
};
